Require Export Machine.


Set Implicit Arguments.

Section CAP.

(* block specification *)
Variable bspec : Type.

(* code heap specification *)
Definition chspec := nat -> optionT bspec.

(* code heap and state predicate *)
Definition chstpred := chspec -> state -> Prop.

Variable interp : bspec -> chstpred.

Definition btPred (b:btype) (x y:nat) : Prop :=
  match b with
    | bteq => x = y
    | btne => x <> y
  end.

Definition btNeg (b:btype) :=
  match b with
    | bteq => btne
    | btne => bteq
  end.

Inductive iseqOk : chstpred -> iseq -> Prop :=
  | jOk : forall (a:chstpred) f,
    (forall P s, a P s -> optPredT (P f) (fun sp => interp sp P s)) ->
    iseqOk a (j f)
  | jalOk : forall (a:chstpred) f fret,
    (forall P s, a P s -> optPredT (P f) (fun sp => interp sp P (rset s ra fret))) ->
    iseqOk a (jal f fret)
  | jrOk : forall (a:chstpred) r,
    (forall P s, a P s -> optPredT (P (rget s r)) (fun sp => interp sp P s)) ->
    iseqOk a (jr r)
  | brOk : forall (a a':chstpred) bt rs rt f i,
    iseqOk a' i ->
    (forall P s, a P s ->
      (btPred (btNeg bt) (rget s rs) (rget s rt) -> a' P s) /\
      (btPred bt (rget s rs) (rget s rt) -> optPredT (P f) (fun sp => interp sp P s))) ->
    iseqOk a (br bt rs rt f i)
  | seqOk : forall (a a':chstpred) c i,
    iseqOk a' i ->
    (forall P s, a P s -> optPred (cstep c s) (a' P)) ->
    iseqOk a (c; i)
.


Definition lift (a:chstpred) (P:chspec) : chstpred :=
  fun P' s =>
    (forall f sp, P f = SomeT sp -> P' f = SomeT sp) /\
    (a P' s).

Inductive cdHeapOk : chspec -> cdHeap -> chspec -> Prop :=
  | cdhpOk : forall P C P',
    (forall f sp, P' f = SomeT sp -> 
      optPred (C f) (iseqOk (lift (interp sp) P))) ->
    cdHeapOk P C P'
  | link : forall P1 P2 C1 C2 P1' P2',
    cdHeapOk P1 C1 P1' ->
    cdHeapOk P2 C2 P2' ->
    (~exists f, C1 f <> None /\ C2 f <> None) ->
    (forall f, P1 f <> NoneT _ -> P2 f <> NoneT _ -> P1 f = P2 f) ->
    cdHeapOk
      (fun f => match P1 f with NoneT => P2 f | x => x end)
      (fun f => match C1 f with None => C2 f | x => x end)
      (fun f => match P1' f with NoneT => P2' f | x => x end)
.


Inductive progOk : prog -> Prop :=
  | progOki : forall P C S I (a:chstpred),
    cdHeapOk P C P ->
    a P S ->
    iseqOk a I ->
    progOk (progi C S I).



Inductive cdHeapOk' : chspec -> cdHeap -> chspec -> Prop :=
  | cdhpOk' : forall P C P',
    (forall f sp, P' f = SomeT sp -> 
      optPred (C f) (iseqOk (lift (interp sp) P))) ->
    cdHeapOk' P C P'.


Lemma iseqLift1 : forall a sp P1 P2 ii,
  a = (lift sp P1) ->
  iseqOk a ii ->
  iseqOk (lift sp (fun f => match P1 f with | SomeT _ => P1 f | NoneT => P2 f end)) ii.
Proof.
  induction 2; subst a; try (constructor; rename H0 into H).

  intros P s [L1 L2].
  apply H; clear H.
  constructor.
  intros f' sp' Eq'; apply L1.
  destruct (P1 f').
  assumption.
  discriminate Eq'.
  assumption.

  intros P s [L1 L2].
  apply H; clear H.
  constructor.
  intros f' sp' Eq'; apply L1.
  destruct (P1 f').
  assumption.
  discriminate Eq'.
  assumption.

  intros P s [L1 L2].
  apply H; clear H.
  constructor.
  intros f' sp' Eq'; apply L1.
  destruct (P1 f').
  assumption.
  discriminate Eq'.
  assumption.

  econstructor.
  apply H0.
  intros P s [L1 L2].
  apply H1.
  constructor.
  intros f' sp' Eq'; apply L1.
  destruct (P1 f').
  assumption.
  discriminate Eq'.
  assumption.

  econstructor.
  apply H0.
  intros P s [L1 L2].
  apply H1.
  constructor.
  intros f' sp' Eq'; apply L1.
  destruct (P1 f').
  assumption.
  discriminate Eq'.
  assumption.

Qed.


Lemma iseqLift2 : forall a sp P1 P2 ii,
  a = (lift sp P2) ->
  (forall f, P1 f <> NoneT _ -> P2 f <> NoneT _ -> P1 f = P2 f) ->
  iseqOk a ii ->
  iseqOk (lift sp (fun f => match P1 f with | SomeT _ => P1 f | NoneT => P2 f end)) ii.
Proof.
  intros a sp P1 P2 ii Aeq Peqv.
  induction 1; subst a; econstructor.

  intros P s [L1 L2].
  apply H; clear H.
  constructor.
  intros f' sp' Eq'; apply L1.
  assert (Peqvf := Peqv f'); clear Peqv.
  destruct (P1 f').
  rewrite Eq' in Peqvf.
  apply Peqvf; discriminate.
  assumption.
  assumption.

  intros P s [L1 L2].
  apply H; clear H.
  constructor.
  intros f' sp' Eq'; apply L1.
  assert (Peqvf := Peqv f'); clear Peqv.
  destruct (P1 f').
  rewrite Eq' in Peqvf.
  apply Peqvf; discriminate.
  assumption.
  assumption.

  intros P s [L1 L2].
  apply H; clear H.
  constructor.
  intros f' sp' Eq'; apply L1.
  assert (Peqvf := Peqv f'); clear Peqv.
  destruct (P1 f').
  rewrite Eq' in Peqvf.
  apply Peqvf; discriminate.
  assumption.
  assumption.

  eauto.
  clear IHiseqOk H.
  intros P s [L1 L2].
  apply H0; clear H0.
  constructor.
  intros f' sp' Eq'; apply L1.
  assert (Peqvf := Peqv f'); clear Peqv.
  destruct (P1 f').
  rewrite Eq' in Peqvf.
  apply Peqvf; discriminate.
  assumption.
  assumption.

  eauto.
  clear IHiseqOk H.
  intros P s [L1 L2].
  apply H0; clear H0.
  constructor.
  intros f' sp' Eq'; apply L1.
  assert (Peqvf := Peqv f'); clear Peqv.
  destruct (P1 f').
  rewrite Eq' in Peqvf.
  apply Peqvf; discriminate.
  assumption.
  assumption.

Qed.


Lemma cdHeapOk'Link : forall P1 P2 C1 C2 P1' P2',
  cdHeapOk' P1 C1 P1' ->
  cdHeapOk' P2 C2 P2' ->
  (~exists f, C1 f <> None /\ C2 f <> None) ->
  (forall f, P1 f <> NoneT _ -> P2 f <> NoneT _ -> P1 f = P2 f) ->
  cdHeapOk'
  (fun f => match P1 f with NoneT => P2 f | x => x end)
  (fun f => match C1 f with None => C2 f | x => x end)
  (fun f => match P1' f with NoneT => P2' f | x => x end).
Proof.
  intros P1 P2 C1 C2 P1' P2' Cok1 Cok2 Cdisj Peqv.
  constructor.
  inversion_clear Cok1; rename H into Cok1.
  inversion_clear Cok2; rename H into Cok2.
  intros f sp P1'eq.
  assert (Cok1' := Cok1 f); clear Cok1.

  assert (H : (exists v, C1 f = Some v) \/ C1 f = None).
  destruct (C1 f); [left | right].
  econstructor; reflexivity.
  reflexivity.

  destruct H as [[ii C1eq] | C1eq].

  rewrite C1eq; rewrite C1eq in Cok1'.
  assert (C2eq : ~ (exists ii, C2 f = Some ii)).
  intros [ii' C2eq].
  apply Cdisj.
  exists f; rewrite C1eq; rewrite C2eq; constructor; discriminate.
  clear Cdisj.
  assert (Cok2' := Cok2 f); clear Cok2.
  destruct (C2 f).
  elimtype False; apply C2eq; econstructor; reflexivity.
  clear C2eq.
  assert (P2eq : P2' f = NoneT _).
  destruct (P2' f).
  elimtype False; eauto.
  reflexivity.
  clear Cok2'; rewrite P2eq in P1'eq.
  assert (P1eq : P1' f = SomeT sp).
  destruct (P1' f).
  assumption.
  discriminate P1'eq.

  clear P1'eq.
  assert (Cok1 := Cok1' _ P1eq); clear Cok1'.
  eapply iseqLift1.
  reflexivity.
  assumption.

  rewrite C1eq; rewrite C1eq in Cok1'.

  assert (P1eq : P1' f = NoneT _).
  destruct (P1' f).
  elimtype False; eauto.
  reflexivity.

  clear Cok1'.
  rewrite P1eq in P1'eq.
  assert (Cok2' := Cok2 _ _ P1'eq); clear Cok2.

  assert (H : (exists v, C2 f = Some v) \/ C2 f = None).
  destruct (C2 f); [left | right].
  econstructor; reflexivity.
  reflexivity.

  destruct H as [[ii C2eq] | C2eq]; rewrite C2eq; rewrite C2eq in Cok2'.
  eapply iseqLift2.
  reflexivity.
  assumption.
  assumption.
  
  assumption.

Qed.


Lemma linkAdmissible : forall P C P',
  cdHeapOk P C P' ->
  cdHeapOk' P C P'.
Proof.
  intros P C P' CHO.
  induction CHO.

  constructor; auto.
  
  apply cdHeapOk'Link; assumption.
  
Qed.


Theorem cdHeapOkEqv : forall P C P',
  cdHeapOk P C P' <-> cdHeapOk' P C P'.
Proof.
  intros P C P'.
  constructor; intro H.
  apply linkAdmissible; assumption.
  destruct H.
  constructor 1.
  assumption.
Qed.


Lemma doBtestTrue : forall bt x y,
  CAP.btPred bt x y ->
  doBtest bt x y = true.
Proof.
  destruct bt; simpl.
  intros; subst x; rewrite beq_true; reflexivity.
  intros x y NE; rewrite (beq_false NE); reflexivity.
Qed.


Lemma doBtestFalse : forall bt x y,
  CAP.btPred (CAP.btNeg bt) x y ->
  doBtest bt x y = false.
Proof.
  destruct bt; simpl.
  intros x y NE; rewrite (beq_false NE); reflexivity.
  intros x y E; subst y; rewrite beq_true; reflexivity.
Qed.


End CAP.
