Require Import POtype.
Require Import List.
Require ListSet.
Require Omega.


Set Implicit Arguments.

(** Set implemented using lists. *)

Module M (Adef : POtype).
  Ltac elf := elimtype False.

  Import Adef.

  Hint Resolve eqNotLT : comp.
  Hint Resolve LTNotEq : comp.
  Hint Resolve LTNotGT : comp.

  Ltac eintuit := intuition eauto with * .

  Ltac leafSolve := eintuit; elimtype False; eintuit.


  (** *** Basic definitions *)

  Definition set := list A.
  
  Definition empty_set : set := nil.
  
  Definition set_add : A -> set -> set := List.cons (A:=A).

  Definition set_mem : A -> set -> bool := ListSet.set_mem eqDec.

  (** ListSet.set_remove only works if set has no duplicates *)
  Fixpoint set_remove (a:A) (x:set) {struct x} : set:=
    match x with
      | nil => empty_set
      | a1::x1 => match (eqDec a a1) with
                    | left _ => set_remove a x1
                    | right _ => a1 :: (set_remove a x1)
                  end
    end.

  Definition set_union (x y : set) : set := x ++ y.

(*  Fixpoint set_unions (s:list set) : set :=
    match s with
      | nil => empty_set
      | x::nil => x
      | x::s' => set_union x (set_unions s')
    end.*)

  (** Copied from ListSet, but is different, because
     it uses a different set_add *)
  Fixpoint set_diff (x y:set) {struct x} : set :=
    match x with
    | nil => nil
    | a1 :: x1 =>
        if set_mem a1 y then set_diff x1 y else set_add a1 (set_diff x1 y)
    end.

  Fixpoint set_inter (x y:set) {struct x} : set :=
    match x with
      | nil => nil
      | a::x' =>
	if set_mem a y then
          set_add a (set_inter x' y)
        else
          set_inter x' y
    end.

  Definition set_disj (s s':set) : Prop :=
    set_diff s s' = s.

  Definition set_fold (B:Set) (f:B -> A -> B) (s:set) (b:B) :=
    fold_left f s b.


  (** **** Canonizing *)

  (** If s is sorted, perform a sorted insertion of a into s,
     removing duplicates *)
  Fixpoint set_insert (a:A) (s:set) {struct s} : set :=
    match s with
      | nil => (a::nil)
      | a'::s' =>
	match compareDec a a' with
	  | inleft H =>
	    match H with
	      | left _ => a::s
	      | right _ => s
	    end
	  | inright _ => a'::(set_insert a s')
	end
    end.


  (** Return the canonical representation of s, by performing 
    a duplicate-removing insertion sort of s *)
  Fixpoint set_canon (s:set) : set := 
    match s with
      | nil => nil
      | x::s' => set_insert x (set_canon s')
    end.


  (** set cardinality *)
  Definition set_card (s:set) := length (set_canon s).

  (** **** Set predicates *)

  Definition set_In : A -> set -> Prop := In (A := A).

  Definition set_eqv (s1 s2:set) : Prop :=
    (forall a, (set_In a s1 <-> set_In a s2)).

  Definition set_subset (s1 s2:set) : Prop :=
    forall a, set_In a s1 -> set_In a s2.

  (** holds on a set if it contains no duplicates *)
  Definition set_noDupP s := set_card s = length s.

  (** iso_set p s s' holds iff p is an isomorphism from s to s' *)
  Definition iso_set (p:A -> option A) (s s':set) : Prop :=
    (forall x, set_In x s <-> p x <> None) /\ (* respects domain *)
    (forall x y, p x = Some y -> set_In y s') /\ (* respects range *)
    (forall x x' y, p x = Some y -> p x' = Some y -> x = x') /\ (* injective *)
    (forall y, set_In y s' -> exists x, p x = Some y).  (* surjective *)


  (** *** Basic properties *)

  Lemma set_diffUnit : forall s,
    set_diff s empty_set = s.
  Proof.
    induction s.
    reflexivity.
    simpl.
    rewrite IHs.
    reflexivity.
  Qed.


  Lemma removeIsDiffStep : forall a s1 s2,
    set_remove a (set_diff s1 s2) = set_diff s1 (set_add a s2).
  Proof.
    induction s1.
    
    reflexivity.
    
    simpl.
    intro.
    destruct (set_mem a0 s2).
    destruct (eqDec a0 a); auto.
    simpl.
    destruct (eqDec a0 a).
    subst; destruct (eqDec a a); intuition.
    destruct (eqDec a a0).
    elimtype False; auto.
    rewrite IHs1; reflexivity.
  Qed.
  

  (** **** set_In properties *)

  (** same property as in ListSet *)
  Lemma set_In_dec : forall a s,
    {set_In a s} + {~ set_In a s}.
  Proof.
    induction s; simpl; auto.
    destruct (eqDec a a0) as [E | NE]; auto.
    destruct IHs as [I | NI]; auto.
    right; intros [E | I']; auto.
  Qed.

  Lemma set_memSound : forall a s,
    if set_mem a s then
      set_In a s
    else
      ~set_In a s.
  Proof.
    induction s.
    simpl; auto.
    simpl.
    destruct (eqDec a a0).
    auto.
    destruct (set_mem a s); intuition.
  Qed.


  Lemma set_mem_true : forall a s,
    set_In a s ->
    set_mem a s = true.
  Proof.
    intros a s I.
    assert (SMS := set_memSound a s).
    destruct (set_mem a s).
    reflexivity.
    elimtype False; exact (SMS I).
  Qed.
  Implicit Arguments set_mem_true [a s].


  Lemma set_mem_false : forall a s,
    ~set_In a s ->
    set_mem a s = false.
  Proof.
    intros a s I.
    assert (SMS := set_memSound a s).
    destruct (set_mem a s).
    elimtype False; exact (I SMS).
    reflexivity.
  Qed.
  Implicit Arguments set_mem_false [a s].
    

  (** intro and elim properties for set_In *)

  Lemma empty_set_elim : forall x,
    set_In x empty_set -> False.
  Proof.
    unfold empty_set.
    intros x H; simpl; auto.
  Qed.


  (* The set_add_* lemmas are from ListSet, but the proofs are different. *)

  Lemma set_add_intro1 : forall a b s,
    a = b -> set_In a (set_add b s).
  Proof.
    simpl; auto.
  Qed.

  Lemma set_add_intro2 : forall a b s,
    set_In a s -> set_In a (set_add b s).
  Proof.
    simpl; auto.
  Qed.

  Lemma set_add_intro : forall a b s,
    a = b \/ set_In a s -> set_In a (set_add b s).
  Proof.
    simpl; intuition.
  Qed.

  Lemma set_add_elim : forall a b s,
    set_In a (set_add b s) -> a = b \/ set_In a s.
  Proof.
    simpl; intuition.
  Qed.

  Lemma set_add_elim2 : forall a b s,
    set_In a (set_add b s) -> a <> b -> set_In a s.
  Proof.
    simpl; intuition.
    subst; intuition.
  Qed.


  (** The set_remove_* lemmas are not in ListSet,
     and do not hold on ListSet.set_remove *)

  Lemma set_remove_intro : forall a b s,
    set_In a s -> a <> b ->
    set_In a (set_remove b s).
  Proof.
    induction s; simpl.
    auto.
    intros Q Neq.
    destruct Q.
    destruct (eqDec b a0).
    subst.
    intuition.
    subst.
    left; reflexivity.
    destruct (eqDec b a0).
    auto.
    simpl.
    right.
    apply IHs; auto.
  Qed.


  Lemma set_remove_elim1 : forall a b s,
    set_In a (set_remove b s) ->
    set_In a s.
  Proof.
    induction s; simpl.
    auto.
    destruct (eqDec b a0).
    subst.
    intuition.
    simpl.
    destruct 1.
    left; assumption.
    right; apply IHs; assumption.
  Qed.


  Lemma set_remove_elim2 : forall a b s,
    set_In a (set_remove b s) ->
    a <> b.
  Proof.
    induction s.
    simpl; auto.
    simpl.
    destruct (eqDec b a0).
    subst.
    assumption.
    simpl.
    destruct 1.
    subst; auto.
    apply IHs; auto.
  Qed.


  Lemma set_remove_elim : forall a b s,
    set_In a (set_remove b s) ->
    a <> b /\ set_In a s.
  Proof.
    intros a b s H.
    constructor.
    eapply set_remove_elim2; eauto.
    eapply set_remove_elim1; eauto.
  Qed.


  (** The set_union_* lemmas are from ListSet, 
    but the proofs are different. *)

  Lemma set_union_intro1 : forall a s1 s2,
    set_In a s1 -> set_In a (set_union s1 s2).
  Proof.
    induction s1; simpl.
    intuition.
    destruct 1; auto.
  Qed.

  Lemma set_union_intro2 : forall a s1 s2,
    set_In a s2 -> set_In a (set_union s1 s2).
  Proof.
    induction s1; simpl; intuition.
  Qed.

  Lemma set_union_intro : forall a s1 s2,
    set_In a s1 \/ set_In a s2 -> set_In a (set_union s1 s2).
  Proof.
    induction s1; simpl; intuition.
  Qed.

  Lemma set_union_elim : forall a s1 s2,
    set_In a (set_union s1 s2) -> set_In a s1 \/ set_In a s2.
  Proof.
    induction s1; simpl; intuition.
    destruct (IHs1 _ H0); intuition.
  Qed.
  Implicit Arguments set_union_elim [a s1 s2].

  (** The set_union_* lemmas are (more or less) 
     from ListSet, but the proofs are different. *)

  Lemma set_diff_intro : forall a s1 s2,
    set_In a s1 -> ~set_In a s2 ->
    set_In a (set_diff s1 s2).
  Proof.
    induction s1.

    simpl; auto.

    simpl.
    intros s2 Z NI.
    assert (MS := set_memSound a0 s2).
    destruct Z.
    
    subst.
    destruct (set_mem a s2).  
    intuition.
    
    left; reflexivity.
    destruct (set_mem a0 s2).
    auto.
    right; auto.
  Qed.


  Lemma set_diff_elim : forall a s1 s2,
    set_In a (set_diff s1 s2) ->
    set_In a s1 /\ ~set_In a s2.
  Proof.
    induction s1.
    intuition.
    simpl.
    intros s2 I.
    assert (MS := set_memSound a0 s2).
    destruct (set_mem a0 s2).
    assert (IH := IHs1 _ I); intuition.
    destruct I.
    subst; intuition.
    assert (IH := IHs1 _ H); intuition.
  Qed.
  Implicit Arguments set_diff_elim [a s1 s2].


  Lemma set_inter_intro : forall a s s',
    set_In a s -> set_In a s' ->
    set_In a (set_inter s s').
  Proof.
    induction s.
    simpl; auto.
    simpl.
    intros s' I I'.
    assert (MS := set_memSound a0 s').
    destruct I as [E | I]; [subst a | idtac]; destruct (set_mem a0 s').
    apply set_add_intro1; reflexivity.
    elimtype False; auto.
    apply set_add_intro2; eauto.
    eauto.
  Qed.


  Lemma set_inter_elim : forall a s s',
    set_In a (set_inter s s') ->
    set_In a s /\ set_In a s'.
  Proof.
    intros a s s'.
    induction s; simpl.
    intro; elimtype False; assumption.

    assert (MS := set_memSound a0 s').
    intro I.
    destruct (set_mem a0 s').
    destruct (set_add_elim I) as [E | I']; clear I; intuition.
    subst a0; intuition.
    intuition.
  Qed.
  Implicit Arguments set_inter_elim [a s s'].

  Lemma set_inter_elimL : forall a s s',
    set_In a (set_inter s s') ->
    set_In a s.
  Proof.
    intros a s s' I.
    destruct (set_inter_elim I).
    assumption.
  Qed.
  Implicit Arguments set_inter_elimL [a s s'].


  Lemma set_inter_elimR : forall a s s',
    set_In a (set_inter s s') ->
    set_In a s'.
  Proof.
    intros a s s' I.
    destruct (set_inter_elim I).
    assumption.
  Qed.
  Implicit Arguments set_inter_elimR [a s s'].


  Lemma set_insert_intro1 : forall a s,
    set_In a (set_insert a s).
  Proof.
    induction s; simpl.
    auto.
    destruct (compareDec a a0).
    destruct s0.
    simpl; auto.
    subst; simpl; auto.
    simpl; right; assumption.
  Qed.
  

  Lemma set_insert_intro2 : forall a b s,
    set_In a s ->
    set_In a (set_insert b s).
  Proof.
    induction s; simpl.
    auto.
    destruct (compareDec b a0).
    destruct s0.
    destruct 1.
    subst; simpl; auto.
    simpl; auto.
    auto.
    simpl.
    destruct 1.
    auto.
    auto.
  Qed.


  Lemma set_insert_elim : forall a b s,
    set_In a (set_insert b s) ->
    a = b \/ set_In a s.
  Proof.
    induction s; simpl.
    intuition.
    destruct (compareDec b a0).
    destruct s0.
    simpl; intuition.
    simpl; intuition.
    simpl; intuition.
  Qed.


  Lemma set_canon_intro : forall a s,
    set_In a s ->
    set_In a (set_canon s).
  Proof.
    induction s.
    simpl; auto.
    simpl.
    intro Z; destruct Z.
    subst; apply set_insert_intro1.
    apply set_insert_intro2.
    apply IHs; assumption.
  Qed.


  Lemma set_canon_elim : forall a s,
    set_In a (set_canon s) ->
    set_In a s.
  Proof.
    induction s.
    simpl; auto.
    
    simpl.
    intro I.
    destruct (set_insert_elim  _ _ _ I).
    left; auto.
    right; apply IHs; assumption.
  Qed.


  (** **** Properties of ordered insertion *)

  (** These are used to show soundness of canonizing. *)

  (** Insertion removes duplicates *)
  Lemma setInsertInsertEq : forall a s,
    set_insert a (set_insert a s) = (set_insert a s).
  Proof.
    induction s; simpl.
    
    destruct (compareDec a a).
    destruct s; leafSolve.
    leafSolve.
    destruct (compareDec a a0).
    destruct s0.
    simpl.
    destruct (compareDec a a).
    destruct s0; leafSolve.
    destruct (compareDec a a0).
    destruct s0; leafSolve.
    leafSolve.
    subst.
    simpl.
    destruct (compareDec a0 a0).
    destruct s0; leafSolve.
    leafSolve.
    simpl.
    destruct (compareDec a a0).
    destruct s0; leafSolve.
    rewrite IHs; auto.
  Qed.


  (** We can reverse a comparison by switching the order of branches *)
  Lemma matchSwap : forall a b (s1 s2 s3:set),
    match compareDec b a with
      | inleft H => if H then s1 else s2
      | inright _ => s3
    end = 
    match compareDec a b with
      | inleft H => if H then s3 else s2
      | inright _ => s1
    end.
  Proof.
    intros.
    destruct (compareDec b a); destruct (compareDec a b).
    destruct s; destruct s0; leafSolve.
    destruct s; leafSolve.
    destruct s; leafSolve.
    leafSolve.
  Qed.


  (** Insert can be pushed into another insert's body *)
  Lemma set_insertSink : forall a b c s1 s2 s3,
    set_insert a
    (match compareDec b c with
       | inleft H => if H then s1 else s2
       | inright _ => s3
     end) =
    (match compareDec b c with
       | inleft H => if H then set_insert a s1 else set_insert a s2
       | inright _ => set_insert a s3
     end).
  Proof.
    intros.
    destruct (compareDec b c).
    destruct s.
    reflexivity.
    reflexivity.
    reflexivity.
  Qed.


  (** Insertion doesn't care about the order elements are added *)
  Lemma setInsertSwap : forall a b s,
    set_insert a (set_insert b s) = set_insert b (set_insert a s).
  Proof.
    induction s.
    simpl; rewrite matchSwap.
    destruct (compareDec b a).
    destruct s; subst; reflexivity.
    reflexivity.
    
    simpl.
    rewrite set_insertSink.
    rewrite set_insertSink.
    simpl.
    destruct (compareDec b a0).
    rewrite matchSwap.
    destruct (compareDec a a0).
    destruct (compareDec b a).
    destruct s0; destruct s1.
    destruct s2; subst; reflexivity.
    subst; destruct s2; leafSolve.
    subst; destruct s2; leafSolve.
    subst; reflexivity.
    destruct s0.
    destruct s1; subst; leafSolve.
    reflexivity.
    destruct s0.
    destruct (compareDec b a).
    destruct s0; subst; leafSolve.
    assert (Q := LTtrans l0 l); leafSolve.
    leafSolve.

    destruct (compareDec a a0).
    destruct s0.
    destruct (compareDec b a).
    destruct s0.
    assert (Q := LTtrans l0 l); leafSolve.
    subst; leafSolve.
    reflexivity.
    reflexivity.
    rewrite IHs.
    reflexivity.
  Qed.


  (** **** Properties of canonizing *)

  (** canonical set predicate *)
  Inductive set_canonP : set -> Prop :=
    | empty_set_canonP : set_canonP empty_set
    | add_set_canonP : forall a s,
      set_canonP s -> (forall a', set_In a' s -> LT a a') ->
      set_canonP (set_add a s).


  Lemma insertPreservesCanonP : forall a s,
    set_canonP s ->
    set_canonP (set_insert a s).
  Proof.
    induction s.
    
    simpl.
    constructor.
    assumption.
    simpl.
    contradiction.

    intro C.
    simpl.
    destruct (compareDec a a0).
    destruct s0.

    (** a < a0 *)
    constructor.
    assumption.
    simpl; intros a' I.
    destruct I.
    
    subst; assumption.
    
    assert (LT a0 a').
    inversion C; subst.
    apply H3.
    assumption.
    eapply LTtrans; eauto.

    (** a = a0 *)
    assumption.

    (** a0 < a *)
    inversion C; subst.
    constructor.
    apply IHs.
    assumption.
    intros.
    destruct (set_insert_elim _ _ _ H).
    subst; assumption.
    apply H2.
    assumption.

  Qed.


  Lemma set_canonCanonP : forall s,
    set_canonP (set_canon s).
  Proof.
    induction s.
    constructor.
    simpl.
    apply insertPreservesCanonP.
    assumption.
  Qed.


  Lemma canonP_set_add_elim1' : forall s,
    set_canonP s ->
    forall a s', s = set_add a s' ->
      ~set_In a s'.
  Proof.
    induction 1.
    intros a s' B; discriminate B.
    
    intros a' s' Eq.
    unfold set_add in Eq.
    injection Eq; clear Eq; intros; subst.
    destruct s'.
    auto.
    
    simpl; intro Z.
    assert (LT a' a').
    auto.
    eapply eqNotLT; eauto.
  
  Qed.
  

  Lemma canonP_set_add_elim1 : forall a s',
    set_canonP (set_add a s') ->
    ~set_In a s'.
  Proof.
    intros; eapply canonP_set_add_elim1'; eauto.
  Qed.

  
  Lemma canonP_set_union_In_elim2' : forall s,
    set_canonP s ->
    forall a s1 s2, s = (set_union s1 s2) -> set_In a s2 ->
      ~set_In a s1.
  Proof.
    unfold set_union.
    induction 1.
  
    unfold empty_set.
    destruct s1.
    auto.
    simpl; intros; discriminate.
  
    intros a' s1 s2 Eq I.
    destruct s1.
    auto.

    simpl in Eq.
    injection Eq; clear Eq; intros Eqs Eqa.
    destruct Eqa.
    assert (IH := IHset_canonP a' _ _ Eqs I); clear IHset_canonP.
    simpl; intro Z.
    destruct Z.
    destruct H1.
    assert (LT a a).
    apply H0.
    rewrite Eqs.
    apply set_union_intro.
    right; assumption.
    eapply eqNotLT; eauto.
    auto.

  Qed.


  Lemma canonP_set_union_In_elim2 : forall a s1 s2,
    set_canonP (set_union s1 s2) -> set_In a s2 ->
    ~set_In a s1.
  Proof.
    intros; eapply canonP_set_union_In_elim2'; eauto.
  Qed.


  Lemma canonP_set_union_elim2' : forall s,
    set_canonP s ->
    forall s1 s2, s = set_union s1 s2 -> 
      set_canonP s2.
  Proof.
    induction 1.
    destruct s1.
    simpl.
    intros; subst; constructor.
    unfold empty_set; simpl; intros; discriminate.
    
    destruct s1; simpl.
    destruct s2; simpl.
    intro; discriminate.
    intro Z; injection Z; clear Z; intros.
    constructor.
    apply (IHset_canonP nil).
    assumption.
    subst; assumption.
    intros s2 Z; injection Z; clear Z; intros.
    apply (IHset_canonP s1 s2 H1).
  Qed.


  Lemma canonP_set_union_elim2 : forall s1 s2,
    set_canonP (set_union s1 s2) ->
    set_canonP s2.
  Proof.
    intros; eapply canonP_set_union_elim2'; eauto.
  Qed.

  Lemma canonP_set_add_elim : forall a s,
    set_canonP (set_add a s) ->
    set_canonP s.
  Proof.
    inversion_clear 1.
    assumption.
  Qed.


  Lemma canonInsertRemoveEq : forall a s,
    set_canonP s -> set_In a s ->
    s = set_insert a (set_remove a s).
  Proof.
    induction 1.
    simpl; intuition.
    
    simpl; intro Z; destruct Z.
    subst.
    destruct (eqDec a a).
    clear e IHset_canonP.
    clear H.
    assert (set_remove a s = s).
    induction s.
    reflexivity.

    simpl.
    simpl in H0.
    destruct (eqDec a a0).
    elimtype False.
    assert (LT a a0).
    apply H0; left; reflexivity.
    eapply eqNotLT; eauto.
    rewrite IHs.
    reflexivity.
    intros; apply H0.
    right; assumption.
    rewrite H.
    destruct s.
    reflexivity.
    simpl.
    destruct (compareDec a a0).
    destruct s0.
    reflexivity.
    assert (LT a a0).
    apply H0; left; reflexivity.
    elimtype False; eapply eqNotLT; eauto.
    assert (LT a a0).
    apply H0; left; reflexivity.
    elimtype False; eapply LTNotGT; eauto.
    elimtype False; apply n; reflexivity.

    assert (LT a0 a).
    apply H0.
    assumption.
    assert (IH := IHset_canonP H1).
    clear H0 H1 IHset_canonP.
    
    destruct (eqDec a a0).
    elimtype False; eapply eqNotLT.
    symmetry; apply e.
    assumption.

    simpl.
    destruct (compareDec a a0).
    destruct s0.
    elimtype False; eapply LTNotGT; eauto.
    elimtype False; auto.
    rewrite <- IH.
    reflexivity.

  Qed.


  Lemma removeInsertEq : forall a s,
    set_canonP s ->
    set_remove a (set_insert a s) = set_remove a s.
  Proof.
    induction 1; simpl.
    destruct (eqDec a a); intuition.

    destruct (compareDec a a0).
    destruct (eqDec a a0).
    subst.
    destruct s0.
    elimtype False; eapply eqNotLT; eauto.
    simpl; destruct (eqDec a0 a0); intuition.
    destruct s0.
    simpl.
    destruct (eqDec a a0); destruct (eqDec a a); intuition.
    intuition.
    destruct (eqDec a a0).
    subst.
    simpl.
    destruct (eqDec a0 a0); intuition.
    simpl.
    destruct (eqDec a a0); intuition.
    rewrite IHset_canonP; reflexivity.
  Qed.


  Lemma removeInsertEqN : forall a b s,
    a <> b -> set_canonP s ->
    set_remove a (set_insert b s) = set_insert b (set_remove a s).
  Proof.
    induction 2.
    simpl.
    destruct (eqDec a b); intuition.

    simpl.
    destruct (eqDec a a0).
    subst.
    destruct (compareDec b a0).
    destruct s0.
    simpl.
    destruct (eqDec a0 b).
    intuition.
    destruct (eqDec a0 a0); intuition.
    clear e.
    rewrite <- IHset_canonP.
    destruct s.
    simpl.
    destruct (eqDec a0 b); intuition.
    simpl.
    destruct (eqDec a0 a).
    subst.
    absurd (LT a a).
    eintuit.
    apply H1.
    left; reflexivity.
    destruct (compareDec b a).
    destruct s0.
    simpl.
    destruct (eqDec a0 a).
    intuition.
    destruct (eqDec a0 b); intuition.
    elimtype False.
    eapply LTNotGT.
    apply l.
    apply H1.
    simpl; auto.
    elimtype False.
    eapply LTNotGT.
    eapply H1.
    left; reflexivity.
    eapply LTtrans; eauto.
    subst; intuition.
    simpl.
    destruct (eqDec a0 a0).
    assumption.
    intuition.

    simpl.
    destruct (compareDec b a0).
    destruct s0.
    simpl.
    destruct (eqDec a a0); intuition.
    destruct (eqDec a b); intuition.
    subst.
    simpl.
    destruct (eqDec a a0); intuition.
    simpl.
    destruct (eqDec a a0); intuition.
    rewrite IHset_canonP.
    reflexivity.
  Qed.


  Lemma set_canonRemoveSwap : forall a s,
    set_canon (set_remove a s) = set_remove a (set_canon s).
  Proof.
    induction s.
    auto.

    simpl.
    assert (Z := set_canonCanonP s).
    destruct (eqDec a a0).
    subst.
    rewrite IHs; clear IHs.
    symmetry; apply removeInsertEq.
    assumption.

    simpl.
    rewrite IHs; clear IHs.
    symmetry; apply removeInsertEqN; auto.

  Qed.


  Lemma canonDupRemove : forall a s,
    set_In a s ->
    set_insert a (set_canon s) = set_canon s.
  Proof.
    induction s.
    
    simpl; intuition.
    
    simpl; intro Z; destruct Z.
    subst.
    rewrite setInsertInsertEq; reflexivity.
    
    rewrite setInsertSwap.
    rewrite (IHs H).
    reflexivity.
  Qed.


  Lemma set_canonInsertSwap : forall x s,
    set_canon (set_insert x s) = set_insert x (set_canon s).
  Proof.
    intros  x s.
    induction s.
    reflexivity.
    simpl.
    destruct (compareDec x a).
    destruct s0.
    reflexivity.
    subst.
    rewrite setInsertInsertEq.
    reflexivity.
    simpl.
    rewrite IHs.
    rewrite setInsertSwap.
    reflexivity.
  Qed.


  Lemma set_canonCanonPEq : forall s,
    set_canonP s ->
    set_canon s = s.
    induction 1.
    reflexivity.
    simpl.
    rewrite IHset_canonP.
    destruct s.
    reflexivity.
    simpl.
    assert (LT a a0).
    apply H0.
    left; auto.
    destruct (compareDec a a0).
    destruct s0.
    reflexivity.
    elimtype False; eapply eqNotLT; eauto.
    elimtype False; eapply LTNotGT; eauto.
  Qed.


  Lemma removePreservesCanon : forall a s,
    set_canonP s ->
    set_canonP (set_remove a s).
  Proof.
    induction 1.
    constructor.

    simpl.
    destruct (eqDec a a0).
    assumption.
    constructor.
    assumption.
    intros.
    apply H0.
    eapply set_remove_elim1.
    eauto.
  Qed.


  (** If two sets are equivalent, their canonical forms
     are equal. *)
  Theorem set_canonEqvEq : forall s1 s2,
    set_eqv s1 s2 ->
    set_canon s1 = set_canon s2.
  Proof.
    induction s1.
    
    simpl.
    destruct s2.
    reflexivity.
    intros.
    elimtype False; destruct (H a) as [_ M].
    apply M; simpl; intuition.
    
    simpl.
    destruct (set_In_dec a s1).
    rewrite (canonDupRemove _ _ s).
    intros.
    apply IHs1; clear IHs1.
    intro a'.
    assert (Z := H a'); clear H.
    destruct Z.
    constructor.
    simpl in H; auto.
    simpl in H0.
    intro Z.
    destruct (H0 Z).
    subst; assumption.
    assumption.
    
    intros s2 Eq.
    assert (set_In a s2).
    destruct (Eq a) as [Z _].
    apply Z; simpl.
    left; reflexivity.
    rewrite (canonInsertRemoveEq a
      (set_canonCanonP _) (set_canon_intro _ _ H)).
    assert (R := IHs1 (set_remove a (set_canon s2))).
    rewrite R.
    rewrite set_canonCanonPEq.
    reflexivity.
    apply removePreservesCanon.
    apply set_canonCanonP.

    intro a'.
    destruct (Eq a') as [R1 R2].
    constructor.
    simpl in R1.
    intro.

    apply set_remove_intro.
    apply set_canon_intro.
    apply R1.
    right; assumption.
    intro Rz; rewrite <- Rz in n; elimtype False; apply n; assumption.

    simpl in R2.
    intro.
    destruct R2.
    apply set_canon_elim.
    eapply set_remove_elim1; eauto.
    subst.
    clear R1.
    absurd (a' <> a').
    intro; auto.
    eapply set_remove_elim2.
    apply H0.
    assumption.
  Qed.


  (** **** Properties of set_eqv *)

  Theorem set_eqvRefl : forall s, set_eqv s s.
  Proof.
    unfold set_eqv; intuition.
  Qed.

  Theorem set_eqvSymm : forall s1 s2,
    set_eqv s1 s2 -> set_eqv s2 s1.
  Proof.
    unfold set_eqv.
    intros s1 s2 E a.
    destruct (E a) as [R T].
    intuition.
  Qed.


  Theorem set_eqvTrans : forall s1 s2 s3,
    set_eqv s1 s2 -> set_eqv s2 s3 ->
    set_eqv s1 s3.
  Proof.
    unfold set_eqv.
    intros s1 s2 s3 E1 E2 a.
    destruct (E1 a) as [E11 E12].
    destruct (E2 a) as [E22 E23].
    intuition.
  Qed.


  Theorem set_canonEqv : forall s,
    set_eqv s (set_canon s).
  Proof.
    unfold set_eqv.
    intros s a.
    constructor.
    intro Z; apply set_canon_intro; assumption.
    intro Z; apply set_canon_elim; assumption.
  Qed.


  Lemma set_eqv_elim1 : forall s1 s2,
    set_eqv s1 s2 -> 
    forall a, set_In a s1 -> 
      set_In a s2.
  Proof.
    unfold set_eqv.
    intros s1 s2 F a.
    destruct (F a); assumption.
  Qed.


  Lemma set_eqv_elim2 : forall s1 s2,
    set_eqv s1 s2 -> 
    forall a, set_In a s2 -> 
      set_In a s1.
  Proof.
    unfold set_eqv.
    intros s1 s2 F a.
    destruct (F a); assumption.
  Qed.


  Lemma set_eqvUnionSymm : forall s1 s2,
    set_eqv (set_union s1 s2) (set_union s2 s1).
  Proof.
    intros.
    intro a.
    constructor;
      intros;
	apply set_union_intro;
	  destruct (set_union_elim H); auto.
  Qed.


  Lemma set_eqvUnionEmptyUnit : forall s,
    set_eqv s (set_union empty_set s).
  Proof.
    intros.
    intro a.
    constructor; intro H.
    apply set_union_intro2; assumption.
    destruct (set_union_elim H).
    elimtype False; simpl in H0; assumption.
    assumption.
  Qed.

  Lemma set_disj_intro : forall s1 s2,
    (forall x, set_In x s1 -> set_In x s2 -> False) ->
    set_disj s1 s2.
  Proof.
    unfold set_disj.
    induction s1.
    reflexivity.
    simpl.
    intros s2 H.
    assert (X := set_memSound a s2).
    destruct (set_mem a s2).
    elimtype False; apply (H a).
    left; reflexivity.
    assumption.
    rewrite IHs1.
    reflexivity.
    intros x I1.
    apply H.
    right; assumption.
  Qed.


  Lemma set_disj_elim : forall s s',
    set_disj s s' ->
    forall x, set_In x s -> set_In x s' -> 
      False.
  Proof.
    unfold set_disj.
    intros s s' H x I.
    rewrite <- H in I.
    destruct (set_diff_elim I).
    assumption.
  Qed.


  (** **** Setoids *)
  Require Import Setoid.

  Lemma SetSetoid : Setoid_Theory set set_eqv.
  Proof.
    constructor.
    apply set_eqvRefl.
    apply set_eqvSymm.
    apply set_eqvTrans.
  Qed.

  Add Setoid set set_eqv SetSetoid as setSetoid.


  Add Morphism set_add : set_addMorph.
  Proof.
    intros a s s' H x; constructor; intro Hx.
    destruct (set_add_elim Hx).
    apply set_add_intro1; assumption.
    apply set_add_intro2.
    eapply set_eqv_elim1; eauto.
    destruct (set_add_elim Hx).
    apply set_add_intro1; assumption.
    apply set_add_intro2.
    eapply set_eqv_elim2; eauto.
  Qed.


  Add Morphism set_mem : set_memMorph.
  Proof.
    intros a s s' E.
    assert (X := set_memSound a s).
    assert (X' := set_memSound a s').
    destruct (set_mem a s); destruct (set_mem a s').
    reflexivity.
    elimtype False; apply X'; eapply set_eqv_elim1; eauto.
    elimtype False; apply X; eapply set_eqv_elim2; eauto.
    reflexivity.
  Qed.


  Add Morphism set_remove : set_removeMorph.
  Proof.
    intros a s s' E x; constructor; intro H.
    destruct (set_remove_elim _ _ _ H).
    apply set_remove_intro.
    eapply set_eqv_elim1; eauto.
    assumption.
    destruct (set_remove_elim _ _ _ H).
    apply set_remove_intro.
    eapply set_eqv_elim2; eauto.
    assumption.
  Qed.


  Add Morphism set_union : set_unionMorph.
  Proof.
    intros s1 s1' s2 s2' H1 H2.
    constructor; intro H; destruct (set_union_elim H) as [HL | HR].
    apply set_union_intro1; eapply set_eqv_elim1; eauto.
    apply set_union_intro2; eapply set_eqv_elim1; eauto. 
    apply set_union_intro1; eapply set_eqv_elim2; eauto.
    apply set_union_intro2; eapply set_eqv_elim2; eauto. 
  Qed.


  Add Morphism set_diff : set_diffMorph.
  Proof.
    intros s1 s1' s2 s2' H1 H2 x.
    constructor; intro Hx.
    destruct (set_diff_elim Hx);
      apply set_diff_intro;
	[eapply set_eqv_elim1; eauto |
	  intro Z; apply H0; eapply set_eqv_elim2; eauto].
    destruct (set_diff_elim Hx);
      apply set_diff_intro;
	[eapply set_eqv_elim2; eauto |
	  intro Z; apply H0; eapply set_eqv_elim1; eauto].
  Qed.


  Add Morphism set_In with signature eq ==> set_eqv ==> iff as set_InMorph.
  Proof.
    intros a s s' E; constructor; intros H.
    eapply set_eqv_elim1; eauto.
    eapply set_eqv_elim2; eauto.
  Qed.


  Add Morphism set_inter : set_interMorph.
  Proof.
    intros s1 s1' E1 s2 s2' E2.
    constructor; intro I.

    destruct (set_inter_elim I) as [I1 I2].
    apply set_inter_intro.
    setoid_rewrite <- E1; assumption.
    setoid_rewrite <- E2; assumption.

    destruct (set_inter_elim I) as [I1 I2].
    apply set_inter_intro.
    setoid_rewrite E1; assumption.
    setoid_rewrite E2; assumption.

  Qed.


  Add Morphism set_insert : set_insertMorph.
  Proof.
    intros a s s' E x; constructor; intro H.
    destruct (set_insert_elim _ _ _ H).
    subst; apply set_insert_intro1.
    apply set_insert_intro2.
    eapply set_eqv_elim1; eauto.
    destruct (set_insert_elim _ _ _ H).
    subst; apply set_insert_intro1.
    apply set_insert_intro2.
    eapply set_eqv_elim2; eauto.
  Qed.
  

  Add Morphism set_canon : set_canonMorph.
  Proof.
    intros s s' E x; constructor; intro H.
    apply set_canon_intro.
    eapply set_eqv_elim1.
    apply E.
    apply set_canon_elim; assumption.
    apply set_canon_intro.
    eapply set_eqv_elim2.
    apply E.
    apply set_canon_elim; assumption.
  Qed.


  Add Morphism set_card : set_cardMorph.
  Proof.
    intros s s' E.
    unfold set_card.
    rewrite (set_canonEqvEq E).
    reflexivity.
  Qed.


  Add Morphism set_subset with signature set_eqv ==> set_eqv ==> iff 
    as set_subsetMorph.
  Proof.
    intros s1 s1' E1 s2 s2' E2; constructor; intros H x I.
    rewrite <- E2; apply H; rewrite E1; auto.
    rewrite E2; apply H; rewrite <- E1; auto.
  Qed.


  Add Morphism set_disj with signature set_eqv ==> set_eqv ==> iff as set_disjMorph.
  Proof.
    intros s1 s1' E1 s2 s2' E2; constructor; intros H; 
      apply set_disj_intro; intros x I1 I2; 
        eapply set_disj_elim with (x := x); eauto.
    setoid_rewrite E1; apply I1.
    setoid_rewrite E2; apply I2.
    setoid_rewrite <- E1; apply I1.
    setoid_rewrite <- E2; apply I2.
  Qed.


  (** **** Properties of set_card *)

  Lemma set_insertCard1 : forall s x,
    set_In x s ->
    set_card (set_insert x s) = (set_card s).
  Proof.
    intros.
    unfold set_card.
    rewrite set_canonInsertSwap.
    rewrite canonDupRemove; auto.
  Qed.

  
  Lemma set_insertCard2 : forall s x,
    ~set_In x s ->
    set_card (set_insert x s) = S (set_card s).
  Proof.
    intros s x I.
    unfold set_card.
    rewrite set_canonInsertSwap.
    assert (~set_In x (set_canon s)).
    intro.
    apply I.
    eapply set_eqv_elim2.
    apply set_canonEqv.
    assumption.
    clear I.
    generalize dependent (set_canon s); clear s.
    induction s; simpl.
    reflexivity.
    intro H.
    assert (Ne: a <> x).
    auto.
    assert (NI: ~set_In x s).
    auto.
    clear H.
    assert (IH := IHs NI); clear IHs NI.
    destruct (compareDec x a).
    destruct s0.
    reflexivity.
    elimtype False; auto.
    simpl; rewrite IH; reflexivity.
  Qed.


  Lemma empty_setCard : set_card empty_set = 0.
  Proof.
    reflexivity.
  Qed.


  Lemma set_cardInsertAddEq : forall x s,
    set_card (set_insert x s) = set_card (set_add x s).
  Proof.
    intros x s.
    unfold set_card.
    simpl.
    rewrite set_canonInsertSwap.
    reflexivity.
  Qed.


  Lemma cardZeroIsEmpty_set : forall s,
    set_card s = 0 -> 
    set_eqv s empty_set.
  Proof.
    unfold set_card.
    intros s H.
    assert (set_eqv (set_canon s) empty_set).
    generalize dependent (set_canon s); clear s.
    induction s.
    intros; eapply set_eqvRefl.
    simpl.
    intro H; elimtype False; discriminate H.
    eapply set_eqvTrans.
    apply set_canonEqv.
    assumption.
  Qed.

  Lemma set_addCard1 : forall s x,
    set_In x s ->
    set_card (set_add x s) = (set_card s).
  Proof.
    intros s x I.
    rewrite <- (set_cardInsertAddEq x s).
    apply set_insertCard1.
    assumption.
  Qed.

  
  Lemma set_addCard2 : forall s x,
    ~set_In x s ->
    set_card (set_add x s) = S (set_card s).
  Proof.
    intros s x NI.
    rewrite <- (set_cardInsertAddEq x s).
    apply set_insertCard2.
    assumption.
  Qed.


  (* move elsewhere in this file? *)
  Lemma set_removeEq : forall a s,
    ~set_In a s ->
    set_remove a s = s.
  Proof.
    induction s.
    intros; reflexivity.
    simpl.
    intro.
    destruct (eqDec a a0).
    subst; elimtype False; auto.
    rewrite IHs.
    reflexivity.
    intro; apply H; right; assumption.
  Qed.
  Implicit Arguments set_removeEq [a s].


  Lemma set_card_remove : forall a s,
    set_In a s ->
    S (set_card (set_remove a s)) = set_card s.
  Proof.
    intros.
    unfold set_card.
    rewrite (set_canonRemoveSwap a s).
    assert (set_In a (set_canon s)).
    eapply set_eqv_elim1.
    apply set_canonEqv.
    assumption.
    clear H.
    assert (set_canonP (set_canon s)).
    apply set_canonCanonP.
    move H0 after H.
    generalize dependent (set_canon s).
    
    induction 1.
    simpl; intros; elimtype False; assumption.
  
    simpl.
    intro Q.
    destruct (eqDec a a0).
    clear Q; subst.
    rename a0 into a.
    assert (~set_In a s0).
    intro.
    assert (Q := H0 _ H1).
    assert (LT a a).
    assumption.
    eintuit.
    rewrite set_removeEq.
    reflexivity.
    assumption.

    destruct Q.
    elimtype False; auto.
    assert (I := IHset_canonP H1); clear IHset_canonP H1.
    simpl.
    rewrite I; reflexivity.
  Qed.


  (* **** Properties of set_disj *)

  Lemma set_disjEmptyUnit : forall s,
    set_disj s empty_set.
  Proof.
    intros s.
    apply set_disj_intro; intros x I1 I2.
    simpl in I2; assumption.
  Qed.


  Lemma unionDisjEqv : forall s1 s1' s2,
    set_eqv (set_union s1 s2) (set_union s1' s2) ->
    set_disj s1 s2 -> set_disj s1' s2 ->
    set_eqv s1 s1'.
  Proof.
    intros s1 s1' s2 E D1 D2 x.
    constructor; intro H.
    assert (set_In x (set_union s1' s2)).
    setoid_rewrite <- E.
    apply set_union_intro1; assumption.
    destruct (set_union_elim H0).
    assumption.
    elimtype False; exact (set_disj_elim D1 _ H H1).
    
    assert (set_In x (set_union s1 s2)).
    setoid_rewrite E.
    apply set_union_intro1; assumption.
    destruct (set_union_elim H0).
    assumption.
    elimtype False; exact (set_disj_elim D2 _ H H1).
  
  Qed.


  (** **** Properties of set_subset *)

  Lemma set_subsetUnionIntroL : forall x y,
    set_subset x y -> forall z,
      set_subset x (set_union y z).
  Proof.
    intros x y S z a H.
    apply set_union_intro1.
    auto.
  Qed.


  Lemma set_subsetRefl : forall x,
    set_subset x x.
  Proof.
    intros x y H; assumption.
  Qed.

  Lemma set_subsetTrans : forall s s',
    set_subset s s' -> forall s'',
      set_subset s' s'' ->
      set_subset s s''.
  Proof.
    intros s s' S s'' S' x I.
    eauto.
  Qed.

  
  Lemma set_interSymm : forall X Y,
    set_eqv (set_inter X Y) (set_inter Y X).
  Proof.
    intros X Y; constructor; intro I;
      destruct (set_inter_elim I) as [I1 I2]; apply set_inter_intro; assumption.
  Qed.


  Lemma set_disjSymm : forall s1 s2,
    set_disj s1 s2 ->
    set_disj s2 s1.
  Proof.
    intros s1 s2 H.
    apply set_disj_intro.
    intros x I1 I2.
    apply (set_disj_elim H x I2 I1).
  Qed.

  
  Lemma addRemoveEqv : forall x s, 
    set_In x s ->
    set_eqv (set_add x (set_remove x s)) s.
  Proof.
    intros x s H y; destruct (eqDec y x) as [R | R].
    rewrite R; constructor; intro X.
    assumption.
    apply set_add_intro1; reflexivity.
    constructor; intro X.
    destruct (set_add_elim X).
    elimtype False; auto.
    eapply set_remove_elim1; eauto.
    apply set_add_intro2.
    apply set_remove_intro; assumption.
  Qed.
  Implicit Arguments addRemoveEqv [x s].


  Lemma set_subsetSplit : forall sa s,
    set_subset sa s -> 
    exists sb,
      set_eqv (set_union sa sb) s /\ set_disj sa sb.
  Proof.

    (* we actually want to induct on the canonical form *)
    assert (F : forall sa, set_canonP sa -> forall s, set_subset sa s ->
      exists sb, set_eqv (set_union sa sb) s /\ set_disj sa sb).
    induction 1; intros s' S.
    
    exists s'; constructor.
    apply set_eqvRefl.
    apply set_disjSymm; apply set_disjEmptyUnit.
    
    destruct (IHset_canonP (set_remove a s')) as [sb [Eqv Disj]].
    intros x I.
    apply set_remove_intro.
    apply S.
    apply set_add_intro2; assumption.
    assert (LT a x).
    apply H0; assumption.
    eintuit.

    clear IHset_canonP.
    exists sb.
    constructor.
    setoid_replace (set_union (set_add a s) sb) with
      (set_add a (set_union s sb)).
    setoid_rewrite Eqv.
    assert (set_In a s').
    apply S.
    apply set_add_intro1; reflexivity.
    setoid_rewrite (addRemoveEqv H1).
    apply set_eqvRefl.
    apply set_eqvRefl.

    apply set_disj_intro.
    intros k I1 I2.
    apply (set_disj_elim Disj k).
    destruct (set_add_elim I1); clear I1.
    subst k.
    elimtype False.
    assert (I3 : set_In a (set_remove a s')).
    setoid_rewrite <- Eqv.
    apply set_union_intro2; assumption.
    destruct (set_remove_elim _ _ _ I3).
    auto.
    assumption.
    assumption.
    
    intros sa s S.
    assert (R := set_canonEqv sa).
    destruct (F (set_canon sa) (set_canonCanonP sa) s) as [sb [HA HB]].
    setoid_rewrite <- R; assumption.
    exists sb; constructor; setoid_rewrite R; assumption.
  
  Qed.




  (** **** Misc properties *)

  Lemma set_canonLengthLe : forall s, 
    length (set_canon s) <= length s.
  Proof.
    intros s.
    assert (X : exists s', s' = set_canon s).
    econstructor; reflexivity.
    destruct X as [s' Hs].
    generalize dependent s.
    induction s';intros s Hs;rewrite <- Hs;simpl.
    destruct (length s); intuition.
    assert (s' = set_canon (set_remove a s)).
    rewrite set_canonRemoveSwap.
    clear IHs'.
    rewrite <- Hs;simpl.
    destruct (eqDec a a).
    2: elimtype False;auto.
    assert (set_canonP (a::s')).
    rewrite Hs;apply set_canonCanonP.
    inversion_clear H.
    clear Hs s e.
    induction s'; auto.
    simpl.
    destruct (eqDec a a0).
    assert (LT a a0).
    apply H1;simpl;auto.
    subst a0.
    elimtype False; eapply eqNotLT; eauto.
    assert (s' = set_remove a s').
    apply IHs';auto.
    inversion_clear H0;auto.
    intros;apply H1.
    simpl;right;auto.
    rewrite <-H;auto.
    rewrite H.
    assert (S (length (set_remove a s)) <= length s).
    assert (set_In a s).
    apply set_canon_elim.
    rewrite <-Hs;simpl;left;reflexivity.
    clear IHs' Hs H s'.
    induction s.
    elimtype False;auto.
    simpl.
    destruct (eqDec a a0).
    subst a0.
    destruct (set_In_dec a s).
    assert (H' := IHs s0).
    auto.
    rewrite (set_removeEq n);auto.
    simpl.
    assert(S (length (set_remove a s)) <= (length s)).
    apply IHs.
    simpl in H0.
    destruct H0;auto.
    elimtype False;auto.
    intuition.
    assert (Hrms := IHs' (set_remove a s) H).
    omega.
  Qed.


  Lemma not_In_removeEq : forall x s,
    ~set_In x s -> set_remove x s = s.
  Proof.
    induction s; simpl.
    reflexivity.
    intro H.
    destruct (eqDec x a).
    subst.
    elimtype False.
    apply H; left; reflexivity.
    rewrite IHs.
    reflexivity.
    intro; apply H; right; assumption.
  Qed.
  

  (* **** Properties of set_noDupP *)

  (* bad name? *)
  Lemma set_noDupRemove : forall s a, 
    set_noDupP (set_add a s) ->
    set_noDupP s.
  Proof.
    unfold set_noDupP.
    intros s a Hs.
    assert (~ set_In a s).
    intro H.
    unfold set_card in Hs.
    simpl in Hs.
    rewrite (canonDupRemove _ _ H) in Hs.
    assert (Hs' := set_canonLengthLe s).
    omega.
    rewrite set_addCard2 in Hs; auto.
  Qed.
  

  Lemma set_noDupAdd : forall s a, 
    set_noDupP s ->
    ~ set_In a s ->
    set_noDupP (set_add a s).
  Proof.
    unfold set_noDupP.
    intros.
    rewrite set_addCard2; auto.
    simpl.
    auto.
  Qed.

  Lemma set_noDupElim : forall s x, 
    set_noDupP (set_add x s) -> 
    ~ set_In x s.
  Proof.
    unfold set_noDupP; unfold set_card.
    simpl.
    intros s x Hn Hin.
    rewrite canonDupRemove in Hn;auto.
    assert (H := set_canonLengthLe s).
    omega.
  Qed.
  

  Lemma set_removeNoDup : forall s x, 
    set_noDupP s -> 
    set_noDupP (set_remove x s).
  Proof.
    unfold set_noDupP.
    induction s.
    intros x _.
    simpl;reflexivity.
    intros x.
    intros Has.
    simpl.
    destruct (eqDec x a).
    apply (IHs x (set_noDupRemove Has)).
    apply set_noDupAdd.
    apply (IHs x (set_noDupRemove Has)).
    assert (~ set_In a s).
    apply set_noDupElim;auto.
    intro X;apply H.
    apply (set_remove_elim1 _ _ _ X).
  Qed.


  Lemma set_diffAddSub : forall X Y,
    set_subset (set_union (set_diff X Y) Y) (set_union X (set_diff Y X)).
  Proof.
    intros X Y a I.
    destruct (set_union_elim I) as [I' | I']; clear I.
    destruct (set_diff_elim I').
    apply set_union_intro1; assumption.
    destruct (set_In_dec a X) as [I | I].
    apply set_union_intro1; assumption.
    apply set_union_intro2; apply set_diff_intro; assumption.
  Qed.

  Lemma set_diffAdd : forall X Y,
    set_eqv (set_union (set_diff X Y) Y) (set_union X (set_diff Y X)).
  Proof.
    intros X Y; constructor; intro I.
    apply set_diffAddSub; assumption.
    rewrite (set_eqvUnionSymm (set_diff X Y) Y);
      apply set_diffAddSub; 
	rewrite (set_eqvUnionSymm (set_diff Y X) X); assumption.
  Qed.


  Lemma set_diffUnionDistrib : forall X Y Z,
    set_eqv (set_diff (set_union X Y) Z) (set_union (set_diff X Z) (set_diff Y Z)).
  Proof.
    intros X Y Z; constructor; intro I.

    destruct (set_diff_elim I) as [I1 I2]; clear I;
      destruct (set_union_elim I1) as [I | I]; 
	[apply set_union_intro1 | apply set_union_intro2]; clear I1;
	  apply set_diff_intro; assumption.

    destruct (set_union_elim I) as [I' | I']; clear I.
    destruct (set_diff_elim I') as [I1 I2]; clear I';
      apply set_diff_intro; [apply set_union_intro; auto | assumption].
    destruct (set_diff_elim I') as [I1 I2]; clear I';
      apply set_diff_intro; [apply set_union_intro; auto | assumption].
  Qed.


  Lemma setUnionDiffDiff : forall X Y Z,
    set_eqv (set_diff X (set_union Y Z)) (set_diff (set_diff X Y) Z).
  Proof.
    intros X Y Z; constructor; intro I.

    destruct (set_diff_elim I) as [I1 I2]; clear I.
    apply set_diff_intro.
    apply set_diff_intro.
    assumption.
    intro; apply I2; apply set_union_intro1; assumption.
    intro; apply I2; apply set_union_intro2; assumption.
    
    destruct (set_diff_elim I) as [I1 I2]; clear I.
    destruct (set_diff_elim I1) as [I I3]; clear I1.
    apply set_diff_intro.
    assumption.
    intro I'; destruct (set_union_elim I') as [I'' | I'']; auto.

  Qed.


  Lemma set_diffUnionEq : forall X Y,
    set_subset Y X ->
    set_eqv (set_union (set_diff X Y) Y) X.
  Proof.
    intros X Y Sub x; constructor; intros Ix.
    destruct (set_union_elim Ix) as [Ix' | Ix']; clear Ix.
    destruct (set_diff_elim Ix'); auto.
    apply Sub; auto.
    destruct (set_In_dec x Y) as [Ix' | NI].
    apply set_union_intro2; auto.
    apply set_union_intro1; apply set_diff_intro; auto.
  Qed.


  Lemma set_diffDisj : forall X Y,
    set_disj X Y ->
    set_diff X Y = X.
  Proof.
    unfold set_disj; auto.
  Qed.

  
  Lemma set_diffSuperSet : forall s s', 
    set_subset s s' -> 
    set_diff s s' = empty_set.
  Proof.
    induction s; intros.
    reflexivity.
    assert (H' :=H a (or_introl _ (refl_equal a))).
    simpl.
    generalize (set_memSound a s').
    case (set_mem a s').
    intros.
    apply IHs.
    intros b Hb.
    apply (H b (or_intror _ Hb)).
    intros; elimtype False; apply (H0 H').
  Qed.


  Lemma set_diffSelf : forall s,
    set_diff s s = empty_set.
  Proof.
    intros s; apply set_diffSuperSet; apply set_subsetRefl.
  Qed.


  (** *** Mapping sets *)
  Section CanonListMap.

    (** The Basic definition is same as ListSet, 
       except we restrict B = A for simplicity, 
       because that is all we need *)

    Definition set_map : (A -> A) -> set -> set := map (A:=A) (B:=A).


    (** Useful properties of existing set functions *)
    Lemma set_map_elim : forall a f s,
      set_In a (set_map f s) ->
      exists b, a = f b /\ set_In b s.
    Proof.
      unfold set_In.
      induction s; simpl.
      intuition.
      destruct 1.
      econstructor; intuition; intuition.
      destruct (IHs H) as [b [H1 H2]].
      exists b.
      constructor; auto.
    Qed.
    Implicit Arguments set_map_elim [a f s].

    
    Lemma set_map_intro : forall a b f s,
      a = f b -> set_In b s ->
      set_In a (set_map f s).
    Proof.
      unfold set_In.
      induction s; simpl.
      auto.
      destruct 2; subst; auto.
    Qed.

    Lemma set_map_add_elim : forall a x f s,
      set_In a (set_map f (set_add x s)) ->
      a = f x \/ set_In a (set_map f s).
    Proof.
      intros.
      destruct (set_map_elim H) as [b [FB SI]].
      destruct (set_add_elim SI).
      subst; auto.
      right.
      eapply set_map_intro; eauto.
    Qed.

    Lemma set_map_eqv : forall f s s',
      set_eqv s s' ->
      set_eqv (set_map f s) (set_map f s').
    Proof.
      intros f s s' E x.
      constructor; intro I;
	destruct (set_map_elim I) as [b [Feq Ib]];
	  subst;
	    eapply set_map_intro; eauto;
	      destruct (E b); auto.
    Qed.  
     
    Lemma set_mapUnionDistrib : forall f X Y,
      set_eqv (set_map f (set_union X Y)) (set_union (set_map f X) (set_map f Y)).
    Proof.
      intros f X Y; constructor; intro I.
      destruct (set_map_elim I) as [b [Aeq Ib]].
      destruct (set_union_elim Ib) as [Ib' | Ib']; clear Ib; 
	[apply set_union_intro1 | apply set_union_intro2];
	apply set_map_intro with b; assumption.
      destruct (set_union_elim I) as [I' | I']; clear I;
	destruct (set_map_elim I') as [b [Aeq Ib]];
	  apply set_map_intro with b; auto; 
	    [apply set_union_intro1 | apply set_union_intro2]; assumption.
    Qed.


  End CanonListMap.


  (* map set elements to entire sets, combining the result *)

  Section MultiMap.

    Variable f  : A -> set.

    Definition set_mmap (s:set) : set :=
      fold_right (fun x s' => set_union (f x) s') empty_set s.

    Lemma set_mmap_intro : forall a b s,
      set_In b s -> set_In a (f b) ->
      set_In a (set_mmap s).
    Proof.
      intros a b s Ib Ia.
      induction s; simpl.

      assumption.
      
      simpl in Ib; destruct Ib.
      
      subst a0.
      apply set_union_intro1.
      assumption.
      
      apply set_union_intro2.
      auto.
    Qed.
    
    Lemma set_mmap_elim : forall a s,
      set_In a (set_mmap s) ->
      exists b, set_In b s /\ set_In a (f b).
    Proof.
      intros a s I.
      induction s; simpl.
      
      elimtype False; assumption.
      
      simpl in I; destruct (set_union_elim I) as [I' | I']; clear I.
      exists a0; intuition.
      destruct (IHs I') as [b H]; exists b; intuition.
    Qed.


    Lemma set_mmap_eqv : forall s s',
      set_eqv s s' ->
      set_eqv (set_mmap s) (set_mmap s').
    Proof.
      intros s s' E x.
      constructor; intro I.
      
      destruct (set_mmap_elim _ _ I) as [y [Iy Ia]]; clear I.
      eapply set_mmap_intro; eauto.
      rewrite <- E; assumption.
      
      destruct (set_mmap_elim _ _ I) as [y [Iy Ia]]; clear I.
      eapply set_mmap_intro; eauto.
      rewrite E; assumption.
    
    Qed.

  
  End MultiMap.


  Section Filter.

    Variable f  : A -> bool.

    Definition set_filter (s:set) : set :=
      fold_right (fun x s' => if f x then set_add x s' else s') empty_set s.

    Lemma set_filter_intro : forall a s,
      f a = true ->
      set_In a s ->
      set_In a (set_filter s).
    Proof.
      intros a s Feq I.
      induction s; simpl.

      assumption.
      
      simpl in I; destruct I.
      
      subst a0; rewrite Feq; apply set_add_intro1; reflexivity.

      assert (IH := IHs H).
      destruct (f a0); [apply set_add_intro2 | idtac]; exact IH.

    Qed.
    
    Lemma set_filter_elim : forall a s,
      set_In a (set_filter s) ->
      f a = true /\ set_In a s.
    Proof.
      intros a s I.
      induction s; simpl.
      
      elimtype False; assumption.
      
      assert (D : f a0 = true \/ f a0 = false).
      destruct (f a0); [left | right]; reflexivity.

      simpl in I; destruct D as [Feq | Feq]; rewrite Feq in I.
      simpl in I; destruct I as [I' | I'].
      subst a0; constructor; auto.
      destruct (IHs I'); auto.
      destruct (IHs I); auto.

    Qed.


    Add Morphism set_filter : set_filterMorph.
    Proof.
      intros s s' E x.
      constructor; intro I.
      
      destruct (set_filter_elim _ _ I) as [Eq I']; clear I.
      apply set_filter_intro.
      assumption.
      rewrite <- E; assumption.
      
      destruct (set_filter_elim _ _ I) as [Eq I']; clear I.
      apply set_filter_intro.
      assumption.
      rewrite E; assumption.
    
    Qed.

  
  End Filter.


  (** unsorted *)

  Lemma unionRemoveAddEqv : forall x s1 s2, 
    set_In x s1 ->
    set_eqv (set_union s1 s2) (set_union (set_remove x s1) (set_add x s2)).
  Proof.
    intros.
    intro a; constructor; intro Q.
    destruct (set_union_elim Q).
    destruct (eqDec a x) as [R | Ne].

    rewrite R.
    apply set_union_intro2.
    apply set_add_intro1.
    reflexivity.

    apply set_union_intro1.
    apply set_remove_intro; assumption.
    apply set_union_intro2.
    apply set_add_intro2; assumption.
    destruct (set_union_elim Q).
    apply set_union_intro1.
    apply (set_remove_elim1 _ _ _ H0).
    destruct (set_add_elim H0).
    rewrite H1; apply set_union_intro1; assumption.
    apply set_union_intro2; assumption.
  Qed.

  Lemma set_eqvUnionAssoc : forall s1 s2 s3,
    set_eqv (set_union (set_union s1 s2) s3) (set_union s1 (set_union s2 s3)).
  Proof.
    intros s1 s2 s3 x; constructor; intro H;
      destruct (set_union_elim H).
    destruct (set_union_elim H0).
    apply set_union_intro1; assumption.
    apply set_union_intro2; apply set_union_intro1; assumption.
    apply set_union_intro2; apply set_union_intro2; assumption.
    apply set_union_intro1; apply set_union_intro1; assumption.
    destruct (set_union_elim H0).
    apply set_union_intro1; apply set_union_intro2; assumption.
    apply set_union_intro2; assumption.
  Qed.


  Lemma set_unionDisjElimL : forall s1 s2 s3, 
    set_disj (set_union s1 s2) s3 -> set_disj s1 s3.
  Proof.
    intros s1 s2 s3 H.
    apply set_disj_intro.
    intros x I1 I3.
    eapply (set_disj_elim H).
    apply set_union_intro1; eauto.
    assumption.
  Qed.


  Lemma set_unionDisjElimR : forall s1 s2 s3, 
    set_disj (set_union s1 s2) s3 -> set_disj s2 s3.
  Proof.
    intros s1 s2 s3 H.
    apply set_disj_intro.
    intros x I1 I3.
    eapply (set_disj_elim H).
    apply set_union_intro2; eauto.
    assumption.
  Qed.

  
  Lemma set_removeUnionDistrib : forall a s s',
    set_eqv (set_remove a (set_union s s')) 
    (set_union (set_remove a s) (set_remove a s')).
  Proof.
    intros a s s' x.
    constructor; intro H.
    destruct (set_remove_elim _ _ _ H) as [Ne I]; clear H.
    destruct (set_union_elim I); clear I.
    apply set_union_intro1; apply set_remove_intro; assumption.
    apply set_union_intro2; apply set_remove_intro; assumption.
    destruct (set_union_elim H); clear H.
    destruct (set_remove_elim _ _ _ H0) as [Ne I]; clear H0.
    apply set_remove_intro.
    apply set_union_intro1; assumption.
    assumption.
    destruct (set_remove_elim _ _ _ H0) as [Ne I]; clear H0.
    apply set_remove_intro.
    apply set_union_intro2; assumption.
    assumption.
  Qed.


  Lemma set_unionDisjIntro : forall s1 s3,
    set_disj s1 s3 -> 
    forall s2, set_disj s2 s3 ->
      set_disj (set_union s1 s2) s3.
  Proof.
    intros s1 s3 D s2 D'.
    apply set_disj_intro.
    intros x I I'.
    destruct (set_union_elim I); clear I.
    exact (set_disj_elim D _ H I').
    exact (set_disj_elim D' _ H I').
  Qed.


  Lemma set_disjSub : forall s1 s2,
    set_disj s1 s2 ->
    forall s1', set_subset s1' s1 ->
      set_disj s1' s2.
  Proof.
    intros s1 s2 D s1' S.
    apply set_disj_intro; intros x I I'.
    refine (set_disj_elim D _ _ I').
    apply (S _ I).
  Qed.


  Lemma phi_injection : forall phi x y,
    phi x = phi y -> forall s1 s2, 
      iso_set phi s1 s2 -> set_In x s1 ->
      x = y.
  Proof.
    intros phi x y E s1 s2 iso I.
    destruct iso as [I1 [I2 [I3 _]]].
    assert (I1x := I1 x); assert (I1y := I1 y); clear I1.
    assert (I2x := I2 x); assert (I2y := I2 y); clear I2.
    assert (I3' := I3 x y).
    destruct (phi x); destruct (phi y); eauto.
    intuition.
  Qed.


  Lemma reflInterEq : forall X,
    set_eqv (set_inter X X) X.
  Proof.
    intros X; constructor; intro I.
    destruct (set_inter_elim I); assumption.
    apply set_inter_intro; assumption.
  Qed.


  Lemma set_disjInterEmpty : forall X Y,
    set_disj X Y ->
    set_eqv (set_inter X Y) empty_set.
  Proof.
    intros X Y D; constructor; intro I.
    simpl; destruct (set_inter_elim I); apply (set_disj_elim D a); assumption.
    elf; assumption.
  Qed.
  Implicit Arguments set_disjInterEmpty [X Y].
  
  
  Lemma addRemoveInter : forall x X Y,
    set_eqv (set_add x (set_inter X Y)) 
    (set_add x (set_inter X (set_remove x Y))).
  Proof.
    intros x X Y; constructor; intro I;
      destruct (eqDec a x).
    
    subst a; left; reflexivity.
    simpl in I; destruct I as [I | I].
    elf; auto.
    destruct (set_inter_elim I).
    right; apply set_inter_intro.
    assumption.
    apply set_remove_intro.
    assumption.
    assumption.

    subst a; left; reflexivity.
    simpl in I; destruct I as [I | I].
    elf; auto.
    destruct (set_inter_elim I); clear I.
    destruct (set_remove_elim _ _ _ H0).
    right; apply set_inter_intro; assumption.

  Qed.



  (* iso_set lemmas *)

  Require Import FnMap.

  Lemma iso_union : forall phi s s2 phi' s' s2',
    iso_set phi s s2 -> iso_set phi' s' s2' -> 
    set_disj s s' -> set_disj s2 s2' ->
    iso_set (map_union phi phi') (set_union s s') (set_union s2 s2').
  Proof.
    intros phi s s2 phi' s' s2' [I1 [I2 [I3 I4]]] [I1' [I2' [I3' I4']]] D D'.

    constructor.
    clear I2 I3 I4 I2' I3' I4'.
    intros a; destruct (I1 a) as [I1a I1b]; 
      destruct (I1' a) as [I1a' I1b']; clear I1 I1'.
    constructor; intro H.
    unfold map_union.
    destruct (set_union_elim H); clear H.
    assert (Q := I1a H0); clear I1a I1b I1a' I1b'.
    destruct (phi a).
    assumption.
    elimtype False; auto.
    assert (Q := I1a' H0); clear I1a I1b I1a' I1b'.
    destruct (phi a).
    intro U; discriminate U.
    assumption.
    unfold map_union in H.
    destruct (phi a).
    apply set_union_intro1; apply I1b; assumption.
    apply set_union_intro2; apply I1b'; assumption.
  
    constructor.
    clear I3 I4 I3' I4'.
    intros a b H.
    assert (X := I2 a b); assert (X' := I2' a b); clear I2 I2'.
    unfold map_union in H.
    destruct (phi a).
    apply set_union_intro1; apply X; assumption.
    apply set_union_intro2; apply X'; assumption.

    constructor.
    unfold map_union.
    intros a a' b H1 H2.
    clear I4 I4'.
    assert (I1a := I1 a); assert (I2a := I2 a); assert (I3a := I3 a).
    destruct (phi a).
    destruct I1a as [_ I1a].
    assert (set_In a s).
    apply I1a; intro X; discriminate X.
    clear I1a.
    assert (set_In b s2).
    eapply I2a; auto.
    clear I2a.
    assert (phi a' = Some b -> a = a').
    intros; eauto.
    clear I3a.
    injection H1; clear H1; intro R; subst.
    assert (I1a := I1 a'); assert (I2a := I2 a'); assert (I3a := I3 a').
    destruct (phi a').
    auto.
    clear H3 I2a I3a.
    elimtype False; apply (set_disj_elim D' b).
    assumption.
    eapply I2'; eauto.

    destruct I1a as [X _].
    clear I2a I3a.
    assert (I1a := I1 a'); assert (I2a := I2 a'); assert (I3a := I3 a').
    destruct (phi a').
    injection H2; clear H2; intro; subst.
    assert (set_In b s2).
    eauto.
    assert (forall a'', phi a'' = Some b -> a' = a'').
    eauto.
    clear I1a I2a I3a.
    assert (set_In b s2').
    eauto.
    elimtype False; apply (set_disj_elim D' b); auto.
    
    exact (I3' _ _ _ H1 H2).
    
    intros b I.
    unfold map_union.
    destruct (set_union_elim I); clear I.
    destruct (I4 b H) as [a E].
    exists a; rewrite E; reflexivity.

    destruct (I4' b H) as [a E].
    exists a; rewrite E.
    assert (R : phi a = None).

    assert (set_In a s').
    destruct (I1' a) as [_ X]; clear I1'.
    rewrite E in X; apply X; intro U; discriminate U.
    assert (~set_In a s).
    intro I'; apply (set_disj_elim D _ I' H0).
    destruct (I1 a) as [_ Y]; clear I1 I4'.
    destruct (phi a).
    elimtype False; apply H1; apply Y; intro I; discriminate I.
    reflexivity.
    
    rewrite R.
    reflexivity.
  
  Qed.


  Lemma empty_mapIso : iso_set (empty_map A A) empty_set empty_set.
  Proof.
    unfold iso_set.
    intuition.
    discriminate H.
    discriminate H.
    elimtype False; exact H.
  Qed.


  Require Import EqNat.

  Definition eqDecB x y :=
    match eqDec x y with
      | left _ => true
      | right _ => false
    end.

  Lemma iso_setRemove : forall phi x y,
      phi x = Some y -> forall objs objs',
      iso_set phi objs objs' ->
      iso_set (map_remove _ _ eqDecB phi x) (set_remove x objs) (set_remove y objs').
  Proof.
    unfold iso_set; unfold map_remove; unfold eqDecB.
    intros phi x y Peq objs objs' [I1 [I2 [I3 I4]]].
    constructor.

    (* first part *)
    intros z; constructor; intro I.
    destruct (set_remove_elim _ _ _ I) as [NE I']; clear I.
    destruct (eqDec x z).
    elimtype False; auto.
    destruct (I1 z) as [Q _]; apply Q; assumption.
    
    destruct (eqDec x z) as [NE | NE].
    elimtype False; auto.
    apply set_remove_intro.
    destruct (I1 z) as [_ Q]; apply Q; assumption.
    auto.
    
    (* second part *)
    constructor.
    intros a b E.
    destruct (eqDec x a) as [NE | NE].
    discriminate E.
    apply set_remove_intro.
    exact (I2 _ _ E).
    intro Eb; subst b; apply NE.
    eapply I3; eauto.

    (* third part *)
    constructor.
    intros a a' b E1 E2.
    destruct (eqDec x a) as [NE | NE].
    discriminate E1.
    destruct (eqDec x a') as [NE' | NE'].
    discriminate E2.
    eapply I3; eauto.

    (* fourth part *)
    intros b I.
    destruct (set_remove_elim _ _ _ I) as [NE I']; clear I.
    destruct (I4 _ I') as [a Peq'].
    exists a.
    destruct (eqDec x a).
    subst a.
    rewrite Peq in Peq'; injection Peq'; intro; elimtype False; auto.
    assumption.
  Qed.


  Lemma iso_setMorphHalf : forall x x1 x2,
    set_eqv x1 x2 -> forall x3 x4, 
      set_eqv x3 x4 -> 
      iso_set x x1 x3 -> iso_set x x2 x4.
  Proof.
    unfold iso_set.
    intros pp s1 s1' E1 s2 s2' E2 [H1 [H2 [H3 H4]]].
    
    constructor; [clear H2 H3 H4 | clear H1].
    intros a; destruct (H1 a) as [H H'].
    constructor.
    intro U; apply H; setoid_rewrite E1; assumption.
    intro U; setoid_rewrite <- E1; apply H'; assumption.

    constructor; [clear H3 H4 | clear H2].
    intros a b H; setoid_rewrite <- E2; exact (H2 a b H).
    
    constructor; [clear H4 | clear H3].
    assumption.

    intros b H; apply (H4 b).
    setoid_rewrite E2; assumption.

  Qed.


  Add Morphism iso_set with signature eq ==> set_eqv ==> set_eqv ==> iff
    as iso_setMorph.
  Proof.
    intros pp s1 s1' E1 s2 s2' E2; constructor; intro H;
      eapply iso_setMorphHalf; eauto; apply set_eqvSymm; auto.
  Qed.


  Ltac hypReplace H y :=
    let HT := fresh in
      (assert (HT := y); clear H; rename HT into H).


  Lemma iso_split : forall X Y phi,
    iso_set phi X Y -> forall X',
    set_subset X' X ->
    exists Y', exists phi',
      iso_set phi' X' Y' /\
      set_subset Y' Y /\
      mapSub phi' phi.
  Proof.
    (* we want to induct on the canonical form of objs1 *)
    intros objs0 objs1' phi1 Iso0 objs2 Sub0.
    assert (Ceq := set_canonEqv objs0).
    assert (Iso : iso_set phi1 (set_canon objs0) objs1').
    rewrite <- Ceq; assumption.
    assert (Sub : set_subset objs2 (set_canon objs0)).
    rewrite <- Ceq; assumption.
    clear Iso0 Sub0 Ceq.
    assert (C := set_canonCanonP objs0).
    set (objs1 := set_canon objs0) in * .
    generalize dependent objs2.
    generalize dependent phi1.
    generalize dependent objs1'.
    generalize dependent objs1.
    clear objs0.

    induction 1; intros objs1' phi1 Iso1 objs2 Sub.

    (* base case *)
    exists empty_set; exists (fun x : A => None (A := A)).
    constructor.
    constructor.
    intros x; constructor; intro I; elimtype False.
    apply (Sub x); assumption.
    auto.
    constructor.
    discriminate 2.
    constructor.
    discriminate 1.
    simpl; intros; elimtype False; assumption.
    constructor.
    simpl; intros x I; elimtype False; assumption.
    intros x y I; discriminate I.
    
    (* inductive case *)
    rename s into objs1; clear C.
    assert (X : exists b, phi1 a = Some b).
    destruct Iso1 as [X _]; destruct (X a) as [Q _].
    destruct (phi1 a).
    econstructor; reflexivity.
    elimtype False; apply Q.
    left; reflexivity.
    reflexivity.
    destruct X as [b Peq].

    assert (Iso1' : iso_set (map_remove _ _ eqDecB phi1 a) objs1 (set_remove b objs1')).
    replace objs1 with (set_remove a (set_add a objs1)).
    apply iso_setRemove; assumption.
    simpl; destruct (eqDec a a).
    apply set_removeEq.
    intro I; hypReplace H (H _ I); elimtype False; eapply eqNotLT; eauto.
    elimtype False; apply n; reflexivity.
    destruct (IHC _ _ Iso1' (set_remove a objs2)) as [objs2' [phi2 [Iso2 [Sub2 Psub2]]]];
      clear Iso1' IHC.
    intros x I; destruct (set_remove_elim _ _ _ I) as [NE I']; clear I;
      hypReplace Sub (Sub _ I').
    simpl in Sub; destruct Sub; [elimtype False; auto | assumption].
    

    (* show that the IH implies the full thing *)
    unfold map_remove in Psub2; unfold eqDecB in Psub2.
    destruct (set_In_dec a objs2) as [Ia | NIa].

    (* case: a in objs2 *)
    exists (set_add b objs2'); exists (map_add eqDecB phi2 a b).
    constructor.
    destruct Iso2 as [I1 [I2 [I3 I4]]].
    unfold map_add.
    constructor.
    intros x; hypReplace I1 (I1 x); destruct I1 as [I1a I1b]; constructor; intro I.
    unfold eqDecB; destruct (eqDec a x) as [NE | NE].
    subst x; discriminate.
    apply I1a; apply set_remove_intro; auto.
    unfold eqDecB in I; destruct (eqDec a x) as [NE | NE].
    subst x; clear I.
    assumption.
    hypReplace I1b (I1b I); clear I.
    destruct (set_remove_elim _ _ _ I1b); assumption.
    constructor.
    intros x y E.
    unfold eqDecB in E; destruct (eqDec a x) as [NE | NE].
    subst x.
    injection E; intro; subst y; left; reflexivity.
    right; eapply I2; eauto.
    constructor.
    intros x x' y E1 E2.
    unfold eqDecB in E1, E2; destruct (eqDec a x) as [NE | NE].
    subst x.
    injection E1; clear E1; intro; subst y.
    destruct (eqDec a x') as [NE | NE].
    assumption.
    elimtype False.
    hypReplace Psub2 (Psub2 _ _ E2); simpl in Psub2.
    destruct (eqDec a x').
    discriminate Psub2.
    apply NE; clear NE.
    destruct Iso1 as [_ [_ [X _]]]; eapply X; eauto.
    destruct (eqDec a x') as [NE2 | NE2].
    subst x'; injection E2; clear E2; intro; subst y.
    hypReplace Psub2 (Psub2 _ _ E1); simpl in Psub2.
    destruct (eqDec a x).
    auto.
    destruct Iso1 as [_ [_ [X _]]]; eapply X; eauto.
    eapply I3; eauto.
    unfold eqDecB; intros y I; simpl in I; destruct I.
    subst y; exists a; destruct (eqDec a a); [reflexivity | elimtype False; auto].
    destruct (I4 _ H0) as [x Peqx].
    exists x; destruct (eqDec a x).
    subst x; hypReplace Psub2 (Psub2 _ _ Peqx).
    unfold map_remove in Psub2; unfold eqDecB.
    destruct (eqDec a a).
    discriminate Psub2.
    elimtype False; apply n; reflexivity.
    assumption.
    (* done showing isomorphism *)

    constructor.
    intros x I; simpl in I; destruct I.
    subst x; destruct Iso1 as [_ [I _]]; eauto.
    destruct (set_remove_elim _ _ _ (Sub2 _ H0)); assumption.
    
    (* show phi sub *)
    unfold eqDecB; unfold map_add; intros x y E.
    destruct (eqDec a x) as [NE | NE].
    subst x; injection E; clear E; intro; subst y; assumption.
    hypReplace Psub2 (Psub2 _ _ E).
    unfold map_remove in Psub2; destruct (eqDec a x).
    discriminate Psub2.
    assumption.

    (* case: a is not in objs2.  apply IH more directly *)
    exists objs2'; exists phi2.
    constructor.
    rewrite <- (set_removeEq NIa (a := a)); assumption.
    constructor.
    intros x I; destruct (set_remove_elim _ _ _ (Sub2 _ I)) as [NE I']; assumption.
    intros x y E; hypReplace Psub2 (Psub2 x y E); simpl in Psub2.
    unfold map_remove in Psub2; destruct (eqDec a x).
    discriminate Psub2.
    assumption.
  Qed.


  Lemma iso_pres_disj : forall x y phi x' y',
    iso_set phi (set_union x y) (set_union x' y') -> forall phix,
    iso_set phix x x' -> forall phiy,
    iso_set phiy y y' ->
    set_disj x y ->
    mapSub phix phi ->
    mapSub phiy phi ->
    set_disj x' y'.
  Proof.
    intros x y phi x' y' Iso phix Isox phiy Isoy Disj Subx Suby.
    apply set_disj_intro; intros b I1 I2.
    destruct Isox as [Ix [_ [_ Isox]]]; destruct (Isox _ I1) as [ax Ex]; clear Isox I1.
    destruct Isoy as [Iy [_ [_ Isoy]]]; destruct (Isoy _ I2) as [ay Ey]; clear Isoy I2.
    
    assert (ax = ay).
    hypReplace Subx (Subx _ _ Ex); clear Ex.
    hypReplace Suby (Suby _ _ Ey); clear Ey.
    destruct Iso as [_ [_ [Iso _]]]; apply (Iso ax ay b); assumption.

    subst ay; rename ax into a.
    apply (set_disj_elim Disj a).
    destruct (Ix a) as [_ Q]; apply Q; rewrite Ex; discriminate.
    destruct (Iy a) as [_ Q]; apply Q; rewrite Ey; discriminate.

  Qed.



  Definition iso_comp (phi phi' : A -> option A) :=
    fun x =>
      match phi' x with
	| Some y => phi y
	| None => None
      end.


  (* this is probably unused, but it is totally sw33t. *)
  Lemma iso_trans : forall phi phi' X Y Z,
    iso_set phi X Y ->
    iso_set phi' Y Z ->
    iso_set (iso_comp phi' phi) X Z.
  Proof.
    unfold iso_comp; unfold iso_set.
    intros phi phi' S1 S2 S3 I1 I2.
    destruct I1 as [W [X [Y Z]]].
    destruct I2 as [W' [X' [Y' Z']]].
    
    constructor.
    intros x.
    hypReplace W (W x); destruct W as [W1 W2].
    hypReplace X (X x).
    constructor.
    intro IX; hypReplace W1 (W1 IX).
    destruct (phi x).
    destruct (W' a) as [W'1 W'2].
    eauto.
    assumption.
    destruct (phi x).
    intro; apply W2; discriminate.
    intro Q; elimtype False; auto.
    
    constructor.
    intros x y.
    hypReplace W (W x); destruct W as [W1 W2].
    hypReplace X (X x).
    hypReplace Z (Z y).
    destruct (phi x).
    intro I; eauto.
    discriminate 1.
    
    constructor.
    intros x x' y.
    hypReplace Y (Y x x').
    destruct (phi x).
    destruct (phi x').
    intros.
    assert (a = a0).
    intros; apply Y' with y; assumption.
    subst a0; eauto.
    discriminate 2.
    discriminate 1.

    intros y IC.
    destruct (Z' y IC) as [x Peq].
    assert (IB : set_In x S2).
    destruct (W' x) as [_ Q].
    apply Q; rewrite Peq; discriminate.
    destruct (Z x IB) as [x0 Peq0].
    exists x0; rewrite Peq0.
    assumption.
  
  Qed.






End M.



