Require Import CAP.
Require Import LinLog.
Require Import Heap.
Require Export Coq.Arith.Even.
Require Import NatListSet.
Require Import FnMap.
Require Import List.
Require Import StateTheory.


(* generic infrastructure for GC *)

Set Implicit Arguments.

Notation "s |||- A" := (heapOf s ||- A) (at level 80).

Notation map := (mapnn).
Notation map_add := (map_addnn).


(*Require Import Coq.Arith.Div2.*)

Definition buffer (x y : nat) : hpred :=
  foralls z in addrRange x y, z |->_ .


(* Basic object heap definitions
   - the heap is a set of objects
   - each object is a pair of fields
   - each field is either odd or an object pointer *)

Definition atomP (x:nat) := odd x.
Definition okFieldVal (objs:set) (x:nat) : Prop := atomP(x) \/ set_In x objs.
Definition okFieldAt (objs:set) (x:nat) : hpred :=
  lexists y, ! okFieldVal objs y ** x |-> y.
Definition okObjAt (objs:set) (x:nat) : hpred :=
  (okFieldAt objs x) ** (okFieldAt objs (4 + x)).
Definition objHp (objs' objs:set) : hpred := 
  foralls x in objs, okObjAt objs' x.


Definition stateTy rroot s objs :=
  okFieldVal objs (rget s rroot) /\
  s |||- objHp objs objs.

(* Our definition of prune is looser than that of Birkedal et al.  They
  say that "prune(s,h) = s, g, where g <= h is the subheap of h restricted
  to those pointers reachable in s,h.", which implies that g contains EXACTLY
  the reachable pointers, whereas we only require that g contains
  AT LEAST the reachable pointers.  I think this is fine for safety.
  You just want to know that every thing you can reach from the root
  is still there, and mapped isomorphically.  Any extra unreachable junk 
  will still be unreachable in the resulting state.
*)
Definition prunesTo rroot s s' :=
  (exists objs, stateTy rroot s' objs) /\
  subHeap (heapOf s') (heapOf s) /\
  rfileOf s = rfileOf s'.


Definition fieldsEqv P rroot s1 s2 :=
  exists s1', exists s2',
    prunesTo rroot s1 s1' /\
    prunesTo rroot s2 s2' /\
    P rroot s1' s2'.


Definition eqState rroot s s' := 
  rget s rroot = rget s' rroot /\ heapOf s = heapOf s'.


Definition mapX (p:map) (x:nat) : option nat :=
  if isOdd x then Some x else p x.


Definition mapHp (S:set) (phi:map) : hpred :=
  !(forall x y, phi x = Some y -> set_In x S) **
  (foralls x in S, lexists a,
    !(phi x = Some a) ** x |-> a ** 4+x |->_).


(** objHp lemmas *)


Lemma okFieldValAssumEqv : forall s s',
  set_eqv s s' -> forall x,
    okFieldVal s x ->
    okFieldVal s' x.
Proof.
  unfold okFieldVal.
  intros.
  destruct H0.
  left; assumption.
  right; eapply set_eqv_elim1; eauto.
Qed.


Lemma okFieldAtAssumEqv : forall s s',
  set_eqv s s' -> forall x h,
  h ||- okFieldAt s x ->
  h ||- okFieldAt s' x.
Proof.
  unfold okFieldAt.
  intros.
  linLogElim H0.
  apply lexI with (x := x0).
  linLogIntro.
  eapply okFieldValAssumEqv; eauto.
  assumption.
Qed.


Lemma okObjAtAssumEqv : forall s s',
  set_eqv s s' -> forall x h,
  h ||- okObjAt s x ->
  h ||- okObjAt s' x.
Proof.
  unfold okObjAt.
  intros.
  eapply sconjImpl.
  apply H0.
  intros; eapply okFieldAtAssumEqv; eauto.
  intros; eapply okFieldAtAssumEqv; eauto.
Qed.


Lemma objHpAssumEqv : forall h s1 s1' s2,
  h ||- objHp s1 s2 -> 
    set_eqv s1 s1' ->
    h ||- objHp s1' s2.
Proof.
  unfold objHp.
  intros.
  apply (fasImpl H).
  intros; eapply okObjAtAssumEqv; eauto.
Qed.


Lemma mapHpIsObjs : forall h M phi,
  h ||- mapHp M phi ->
  h ||- foralls x in M, x |->__.
Proof.
  unfold mapHp.
  intros h M phi Hp.
  linLogElim Hp.
  apply (fasImpl Hp).
  cbv beta; intros x hh HpH; linLogElim HpH; linLogIntro; eauto.
Qed.


Lemma objHpIsObjs : forall h objs' objs,
  h ||- objHp objs' objs ->
  h ||- foralls x in objs, x |->__.
Proof.
  unfold objHp.
  intros h objs' objs Hp.
  apply (fasImpl Hp).
  unfold okObjAt; unfold okFieldAt.
  cbv beta; intros x hh HH; linLogElim HH; linLogIntro; eauto.
Qed.
Implicit Arguments objHpIsObjs [h objs' objs].



(* hmmm... *)
Lemma fasImplB : forall P Q,
  (forall x h, h ||- P x -> h ||- Q x) ->
  forall s h,
    fas P s h ->
    fas Q s h.
Proof.
  intros P Q F.
  unfold fas.
  intro s.
  induction (set_canon s).
  auto.

  simpl.
  intros.
  destruct H as [ha [hb [J [HA HB]]]].
  exists ha; exists hb; intuition.
  unfold hpredTrue in HB |- * .
  apply IHs0; assumption.
Qed.


Add Morphism objHp with signature set_eqv ==> set_eqv ==> eq ==> iff as objHpMorph.
Proof.
  unfold objHp.
  intros s1 s1' E1 s2 s2' E2 h; constructor; intro H.
  setoid_rewrite <- E2.
  eapply fasImplB.
  2:apply H.
  intros.
  eapply okObjAtAssumEqv; eauto.
  rewrite E2.
  eapply fasImplB.
  2:apply H.
  cbv beta; intros.
  eapply okObjAtAssumEqv; eauto.
  apply set_eqvSymm; auto.
Qed.


Lemma objHpElim : forall x s2,
  set_In x s2 ->
  forall s1 h, h ||- objHp s1 s2 -> 
    h ||- okObjAt s1 x ** objHp s1 (set_remove x s2).
Proof.
  unfold objHp.
  intros x s2 I s1 h H.
  assert (Q := fasElim I H).
  assumption.
Qed.
Implicit Arguments objHpElim [x s2 s1 h].


Lemma objHpElimField : forall h s1 s2 A x,
  h ||- objHp s1 s2 ** A -> set_In x s2 ->
    h ||- okFieldAt s1 x ** top.
Proof.
  intros h s1 s2 A x H1 H2.
  assert (h ||- (okFieldAt s1 x ** top) ** top).
  eapply sconjImpl.
  apply H1.
  clear H1; intros h1 H1.
  assert (Q := objHpElim H2 H1).
  unfold okObjAt in Q.
  linLogElim Q.
  eapply sconjImpl.
  apply Q.
  auto.
  intros; linLogIntro.
  intros; linLogIntro.
  linLogElim H.
  eapply sconjImpl.
  apply H.
  auto.
  intros; linLogIntro.
Qed.
Implicit Arguments objHpElimField [h s1 s2 x].


Lemma objHpIntroE : forall s h,
  h ||- unit ->
  h ||- objHp s empty_set.
Proof.
  intros.
  unfold objHp.
  apply fasIntroE.
  assumption.
Qed.


Lemma objHpIntro : forall h s1 s2 x,
  h ||- okObjAt s1 x ** objHp s1 s2 ->
    h ||- objHp s1 (set_add x s2).
Proof.
  intros h s1 s2 x H.
  unfold objHp.
  apply fasIntro.
  assumption.
  intro I.
  hypReplace H (sconjImplR H _ (fun h => fasElim (h:=h) I)).
  unfold okObjAt in H; unfold okFieldAt in H.
  linLogElim H.
  clear H4 H7 H10 H13.
  hypReplace H (sconjAssocR H).
  hypReplace H (sconjAssocR H).
  assert (h ||- ((x |-> x0 ** x |-> x2) ** top) ** top).
  apply (sconjImpl H); clear h H; intros h H.
  hypReplace H (sconjAssocL H).
  apply sconjAssocR.
  apply (sconjImpl H); clear h H; intros h H.
  assumption.
  apply sconjComm.
  apply (sconjImpl H); clear h H; intros h H.
  linLogIntro.
  assumption.
  linLogIntro.
  clear H.
  hypReplace H0 (sconjAssocL H0).
  refine (sconjifierL (sconjDisj (x:=x) (y:=x0) (z:=x2)) H0).
Qed.


Lemma objHpSplit : forall h s s1 s2,
  h ||- objHp s (set_union s1 s2) ->
    set_disj s1 s2 ->
    h ||- objHp s s1 ** objHp s s2.
Proof.
  assert (F : forall s0 s1 s2, set_canonP s1 -> 
    forall h,
      h ||- objHp s0 (set_union s1 s2) ->
	set_disj s1 s2 ->
	h ||- objHp s0 s1 ** objHp s0 s2).
  induction 1; intros h Hp D.

  hypReplace Hp (unitIL Hp).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  apply objHpIntroE; assumption.
  assumption.
  
  assert (IA : set_In a (set_union (set_add a s) s2)).
  apply set_union_intro1; apply set_add_intro1; reflexivity.
  hypReplace Hp (objHpElim IA Hp); clear IA.
  assert (Hp' : h ||- okObjAt s0 a ** objHp s0 (set_union s s2)).
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  replace (set_union (set_add a s) s2) with 
    (set_add a (set_union s s2)) in Hp.
  simpl in Hp.
  destruct (POnat.POnat.eqDec a a).
  rewrite not_In_removeEq in Hp.
  assumption.
  intro I.
  destruct (set_union_elim I).
  assert (a < a).
  apply H0; assumption.
  omega.
  apply (set_disj_elim D a).
  apply set_add_intro1; reflexivity.
  assumption.
  elimtype False; apply n; reflexivity.
  reflexivity.
  clear Hp.
  assert (Hp : h ||- okObjAt s0 a ** objHp s0 s ** objHp s0 s2).
  apply (sconjImplR Hp'); clear h Hp'; intros h Hp.
  apply IHset_canonP.
  assumption.
  apply set_disj_intro; intros k I1 I2.
  apply (set_disj_elim D k).
  apply set_add_intro2; assumption.
  assumption.

  clear Hp'.
  hypReplace Hp (sconjAssocR Hp).
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply objHpIntro; assumption.

  intros h s s1 s2 Hp D.
  assert (R := set_canonEqv s1).
  assert (Q := F s (set_canon s1) s2 (set_canonCanonP s1) h); clear F.
  assert (Hp' : h ||- objHp s (set_canon s1) ** objHp s s2).
  apply Q.
  unfold hpredTrue; setoid_rewrite <- R; assumption.
  setoid_rewrite <- R; assumption.
  apply (sconjImplL Hp'); clear Hp; intros h2 Hp.
  unfold hpredTrue; setoid_rewrite R; assumption.

Qed.
Implicit Arguments objHpSplit [h s s1 s2].


Theorem objHpSub : forall h objs0 objs A,
  h ||- objHp objs0 objs ** A ->
    forall objs', set_subset objs' objs ->
      h ||- objHp objs0 objs' ** top.
Proof.
  intros h objs0 objs A Hp objs' Sub.
  destruct (set_subsetSplit Sub) as [sb [Seq Sdisj]].
  assert (h ||- (objHp objs0 objs' ** top) ** top).
  sconjMatch.
  assert (Hp' : h ||- objHp objs0 (set_union objs' sb)).
  unfold hpredTrue; setoid_rewrite Seq.
  assumption.
  hypReplace Hp' (objHpSplit Hp' Sdisj).
  sconjMatch.
  linLogElim H.
  assumption.
Qed.


Add Morphism okFieldVal with signature set_eqv ==> eq ==> iff as morphOkFieldVal.
Proof.
  unfold okFieldVal.
  intros s s' E n; constructor; intros H; (destruct H; [left | right]); auto.
  setoid_rewrite <- E; auto.
  setoid_rewrite E; auto.
Qed.

Add Morphism okFieldAt with signature set_eqv ==> eq ==> eq ==> iff 
  as morphOkFieldAt.
Proof.
  unfold okFieldAt.
  intros s s' E n h; constructor; intros H; destruct H as [x H].
  exists x; destruct H as [ha [hb [J [HA HB]]]].
  exists ha; exists hb; intuition.
  llei HA; auto; setoid_rewrite <- E; auto.
  exists x; destruct H as [ha [hb [J [HA HB]]]].
  exists ha; exists hb; intuition.
  llei HA; auto; setoid_rewrite E; auto.
Qed.


Add Morphism okObjAt with signature set_eqv ==> eq ==> eq ==> iff
  as okObjAtMorphism.
Proof.
  unfold okObjAt.
  intros s s' E n h; constructor; intros Hp;
    destruct Hp as [ha [hb [J [HA HB]]]]; exists ha; exists hb; constructor; auto.
  constructor; unfold hpredTrue; rewrite <- E; assumption.
  constructor; unfold hpredTrue; rewrite E; assumption.
Qed.


Lemma objHpElimE : forall h s,
  h ||- objHp s empty_set ->
    h ||- unit.
Proof.
  unfold objHp.
  intros h s H.
  exact (fasElimE H).
Qed.


Lemma objHpMerge : forall h s s1 s1',
  h ||- objHp s s1 ** objHp s s1' ->
  h ||- objHp s (set_union s1 s1').
Proof.
  intros h s s1 s1' H.
  unfold hpredTrue; setoid_rewrite (set_canonEqv s1).
  assert (C := set_canonCanonP s1).
  
  assert (h ||- objHp s (set_canon s1) ** objHp s s1').
  apply (sconjImplL H); clear h H; intros h H.
  unfold hpredTrue; setoid_rewrite <- (set_canonEqv s1).
  exact H.

  clear H; rename H0 into H.
  generalize dependent s1'.
  generalize dependent h.
  generalize dependent (set_canon s1); clear s1.
  induction 1; simpl; intros h s1' H'.
  apply unitEL; simpl.
  apply (sconjImpl H'); clear h H'; intros h H.
  apply (objHpElimE H).
  assumption.

  apply objHpIntro.
  assert (I : set_In a (set_add a s0)).
  left; reflexivity.

  assert (h ||- (okObjAt s a ** objHp s s0) ** objHp s s1').
  apply (sconjImplL H'); clear H'; intros h' H'.
  assert (H2 := objHpElim I H'); clear H'.
  apply (sconjImplR H2); clear h' H2; intros h' H'.
  simpl in H'.
  destruct (POnat.POnat.eqDec a a).
  assert (~set_In a s0).
  intro X.
  assert (Q := H _ X).
  unfold POnat.POnat.LT in Q; omega.
  rewrite <- (set_removeEq H0).
  assumption.
  elimtype False; apply n; reflexivity.
  clear H' I.
  linLogElim H0.

  apply (sconjImplR H0).
  intros.
  unfold hpredTrue; apply IHC.
  assumption.

Qed.


(** mapHp lemmas *)

Lemma mapHpMorphHalf : forall s s' m m' h,
  set_eqv s s' ->
  mapEqvnn m m' ->
  mapHp s m h ->
  mapHp s' m' h.
Proof.
  intros s s' m m' h Es Em Hp.
  destruct Hp as [ha [hb [J [HA HB]]]].
  lle HA; exists ha; exists hb; intuition.
  lli.
  auto.
  intros x y Meq; setoid_rewrite <- Es.
  eapply H.
  rewrite (Em x); eauto.
  unfold hpredTrue; setoid_rewrite <- Es.
  rewrite <- hpredTrueEq; eapply fasImpl.
  eauto.
  simpl.
  intros x h' Hp.
  llei Hp; eauto.
  unfold mapEqvnn in Em; rewrite <- Em; auto.
Qed.


Add Morphism mapHp with signature set_eqv ==> mapEqvnn ==> eq ==> iff 
  as set_eqvMapHp.
  intros s s' Es m m' Em h; constructor; intro H; eapply mapHpMorphHalf; eauto.
  apply set_eqvSymm; auto.
  unfold mapEqvnn; apply mapEqvSymm; auto.
Qed.


Lemma mapHpIntroE : forall h,
  h ||- unit ->
    h ||- mapHp empty_set (fun f => None).
Proof.
  unfold mapHp.
  intros h Hp.
  lli; auto.
  discriminate 1.
Qed.


Lemma mapHpElim1 : forall x M,
  set_In x M -> forall phi h,
  h ||- mapHp M phi ->
    h ||- lexists v, !(phi x = Some v) ** x |-> v ** top.
Proof.
  unfold mapHp.
  intros x M I phi h Hp.
  linLogElim Hp.
  hypReplace Hp (fasElim I Hp); simpl in Hp.
  linLogElim Hp.
  linLogIntro; eauto.
  sconjMatch.
Qed.
Implicit Arguments mapHpElim1 [x M phi h].


Lemma mapHpElim2 : forall x M,
  ~set_In x M -> forall phi h,
  h ||- mapHp M phi -> 
    phi x = None.
Proof.
  unfold mapHp.
  intros x M I phi h H.
  linLogElim H.
  assert (X := H2 x).
  destruct (phi x).
  elimtype False; apply I.
  apply (X n); reflexivity.
  reflexivity.
Qed.
Implicit Arguments mapHpElim2 [x M phi h].


Lemma mapHpIntro : forall h M phi x y z,
  h ||- mapHp M phi ** x |->, (y::z::nil) ->
    h ||- mapHp (set_add x M) (map_add phi x y).
Proof.
  intros h M phi x y z Hp.
  assert (~set_In x M).
  intro I.

  assert (Hp' : h ||- (lexists v, x |-> v ** top) ** x |-> y ** top).
  sconjMatch.
  assert (Q := mapHpElim1 I Hp).
  linLogElim Q.
  linLogIntro; sconjMatch; eauto.
  simpl in Hp; sconjMatch.
  linLogElim Hp'.
  unfold mapHp.
  linLogIntro.
  intros x0 y0 Peq.
  compare x0 x.
  intro; subst.
  left; reflexivity.
  intro Ne.
  unfold map_addnn in Peq; unfold FnMap.map_add in Peq; rewrite beq_false in Peq.
  2:auto.
  right.
  destruct (set_In_dec x0 M).
  assumption.
  rewrite (sconjifierL (mapHpElim2 n (phi:=phi)) Hp) in Peq.
  discriminate Peq.
  apply fasIntro.
  2:assumption.
  hypReplace Hp (sconjComm Hp).
  simpl in Hp; linLogElim Hp.
  linLogIntro.
  unfold map_addnn; unfold FnMap.map_add.
  rewrite beq_true; reflexivity.
  (* for some reason, using scm without doing instantiate causes
     Windows to shut down Coq *)
  instantiate (1 := z); scm.
  unfold mapHp in Hp.
  lle Hp.
  apply (fasImpl Hp).
  clear h Hp; intros w h Hp.
  llei Hp.
  unfold map_addnn; unfold FnMap.map_add.
  compare x w.
  intros E; subst w; elimtype False; eauto.
  intro Ne; rewrite (beq_false Ne); eauto.
  eauto.
Qed.


Lemma mapHpPhiEqv : forall h M phi,
  h ||- mapHp M phi -> 
    forall phi', mapEqv phi phi' ->
    h ||- mapHp M phi'.
Proof.
  unfold mapHp.
  intros h M phi H phi' E.
  linLogElim H.
  linLogIntro.
  intros.
  eapply H2.
  rewrite (E x).
  eauto.
  apply (fasImpl H).
  clear h H2 H; intros x h H.
  linLogElim H; apply lexI with (x := x0); linLogIntro.
  rewrite <- (E x); assumption.
  eauto.
Qed.


(** buffer lemmas *)

Lemma bufferIntroE : forall h x,
  h ||- unit ->
    h ||- buffer x x.
Proof.
  unfold buffer.
  intros h x Hp.
  unfold hpredTrue; setoid_rewrite <- (set_eqvEmptyAddrRange x).
  apply fasIntroE; assumption.
Qed.


Lemma bufferElimE : forall h x,
  h ||- buffer x x ->
    h ||- unit.
Proof.
  unfold buffer.
  intros h x Hp.
  unfold hpredTrue in Hp; rewrite <- (set_eqvEmptyAddrRange x) in Hp.
  apply (fasElimE Hp).
Qed.


Lemma bufferElim : forall h x y,
  h ||- buffer x y ->
    aligned4 x y -> x <> y ->
    h ||- buffer x (y - 4) ** (y - 4 |->_).
Proof.
  unfold buffer.
  intros h x y Hp [k R] Ne.
  destruct k.
  elimtype False; omega.
  apply sconjComm.
  assert (I : set_In (y-4) (addrRange x y)).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := k).
  omega.
  hypReplace Hp (fasElim I Hp).
  simpl in Hp.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  unfold hpredTrue.
  setoid_replace (addrRange x (y - 4)) with
    (set_remove (y - 4) (addrRange x y)).
  apply Hp.

  clear Hp; intros z; constructor; intros I'.

  destruct (addrRangeElim I') as [I2 [n I3]]; clear I'.
  apply set_remove_intro.
  2:omega.
  subst.
  eapply addrRangeIntro.
  omega.
  reflexivity.

  destruct (set_remove_elim _ _ _ I') as [I1 I2]; clear I'.
  destruct (addrRangeElim I2) as [I3 [k' I4]]; clear I2.
  eapply addrRangeIntro.
  subst.
  replace (x + 4 * S k - 4) with (x + 4 * k) in I1.
  replace (4 + (x + 4 * k')) with (x + 4 * S k') in I3.
  assert (LT : k' < k).
  assert (LTE : k' <= k).
  omega.
  compare k' k; intro R.
  subst.
  elimtype False; apply I1; reflexivity.
  omega.
  omega.
  omega.
  omega.
  apply I4.

Qed.
Implicit Arguments bufferElim [h x y].


Lemma bufferMerge : forall h x y z,
  h ||- buffer x y ** buffer y z ->
    aligned4 x y -> aligned4 y z ->
    h ||- buffer x z.
Proof.
  unfold buffer.
  intros h x y z Hp [k R] [k' R'].
  generalize dependent h.
  generalize dependent x.
  induction k; intros x R h.

  subst.
  replace (x + 4 * 0) with x.
  intros.
  apply unitEL.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  unfold hpredTrue in Hp; setoid_rewrite <- (set_eqvEmptyAddrRange x) in Hp.
  apply (fasElimE Hp).
  auto.
  auto.

  hypReplace IHk (IHk (4 + x)).
  intro Hp.
  assert (I : set_In x (addrRange x z)).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0).
  omega.

  unfold hpredTrue.
  setoid_rewrite <- (addRemoveEqv I).
  clear I.
  setoid_rewrite (addrRangeRemoveMin x z).
  replace (x + 4) with (4 + x).
  apply fasIntro.

  assert (I : set_In x (addrRange x y)).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 0).
  omega.
  
  assert (Hp' : h ||- (x |->_ ** 
    (foralls z in addrRange (4 + x) y, z |->_)) **
    (foralls z in addrRange y z, z |->_)).
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  hypReplace Hp (fasElim I Hp).
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  unfold hpredTrue.
  replace (4 + x) with (x + 4).
  setoid_rewrite <- (addrRangeRemoveMin x y).
  apply Hp.
  intuition.
  hypReplace Hp' (sconjAssocL Hp').
  apply (sconjImplR Hp'); clear h Hp Hp'; intros h Hp.
  apply IHk.
  omega.
  assumption.
  intro Z; destruct (addrRangeElim Z) as [HA [q HQ]].
  omega.
  intuition.

Qed.


Lemma bufferIsObjs : forall h x y,
  aligned8 x y ->
  h ||- buffer x y ->
    h ||- foralls x in rangeObjs x y, x |->__.
Proof.
  intros h x y [k E]; subst.
  generalize dependent h.
  induction k; intro h.

  simpl; replace (x + 0) with x.
  2:auto.
  intro H.
  unfold hpredTrue.
  setoid_rewrite <- (set_eqvEmptyRangeObjs x).
  apply fasIntroE.
  eapply bufferElimE; eauto.

  intros.
  unfold hpredTrue.
  assert (Al : aligned8 x (x + 8 * S k)).
  exists (S k); reflexivity.
  assert (Ne : x <> x + 8 * S k).
  omega.
  setoid_rewrite (rangeObjsEndSplit Al Ne).
  replace (x + 8 * S k - 8) with (x + 8 * k).
  2:omega.
  apply fasIntro.
  clear Ne Al.
  replace (x + 8 * k |->__) with (x + 8 * k |->_ ** 4 + x + 8 * k |->_).
  2:reflexivity.
  apply sconjAssocR.
  assert (aligned4 x (x + 8 * S k)).
  exists (2 * S k); omega.
  assert (x <> x + 8 * S k).
  omega.
  hypReplace H (bufferElim H H0 H1); clear H0 H1.
  sconjGLift2.
  sconjLift2of2 H.
  apply (sconjImpl H); clear h H; intros h H.
  replace (4 + x + 8 * k) with (x + 8 * S k - 4).
  assumption.
  omega.
  assert (aligned4 x (x + 8 * S k - 4)).
  exists (2 * S k - 1); omega.
  assert (x <> x + 8 * S k - 4).
  omega.
  hypReplace H (bufferElim H H0 H1); clear H0 H1.
  apply sconjComm.
  sconjLift2of2 H.
  replace (x + 8 * S k - 4 - 4) with (x + 8 * k) in H.
  apply (sconjImpl H); clear h H; intros h H.
  assumption.
  apply IHk; assumption.
  omega.
  clear H Al IHk; intro I.
  destruct (rangeObjsElim I) as [I1 [n I2]]; clear I.
  omega.
Qed.


Lemma objsIsBuffer : forall h x y,
  h ||- (foralls z in (rangeObjs x y), z |->__) ->
    aligned8 x y ->
    h ||- buffer x y.
Proof.
  unfold buffer.
  intros h x y Hp [k R].

  subst.
  generalize dependent h.
  induction k.

  replace (x + 8 * 0) with x.
  intros h Hp.
  unfold hpredTrue; 
    setoid_rewrite <- (set_eqvEmptyAddrRange x).
  apply fasIntroE.
  unfold hpredTrue in Hp; setoid_rewrite <- (set_eqvEmptyRangeObjs x) in Hp.
  exact (fasElimE Hp).
  auto.

  intros h Hp.
  unfold hpredTrue.
  assert (I : set_In (x + 8 * S k - 4)  (addrRange x (x + 8 * S k))).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 2 * k + 1).
  omega.
  setoid_rewrite <- (addRemoveEqv I).
  clear I.
  assert (Al : aligned4 x (x + 8 * S k)).
  exists (2 * k + 2); omega.
  setoid_rewrite (addrRangeRemoveMax Al).
  clear Al.
  assert (I : set_In (x + 8 * S k - 4 - 4)  (addrRange x (x + 8 * S k -4))).
  eapply addrRangeIntro.
  omega.
  instantiate (1 := 2 * k).
  omega.
  setoid_rewrite <- (addRemoveEqv I).
  clear I.
  assert (Al : aligned4 x (x + 8 * S k - 4)).
  exists (2 * k + 1); omega.
  setoid_rewrite (addrRangeRemoveMax Al).
  clear Al.
  replace (x + 8 * S k - 4 - 4) with (x + 8 * k).
  replace (x + 8 * S k - 4) with (x + 8 * k + 4).

  apply fasIntro.
  
  assert (I : set_In (x + 8 * k) (rangeObjs x (x + 8 * S k))).
  eapply rangeObjsIntro.
  omega.
  reflexivity.

  hypReplace Hp (fasElim I Hp).
  hypReplace Hp (sconjAssocL Hp).
  sconjLift2 Hp.
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  replace (x + 8 * k + 4) with (4 + (x + 8 * k)).
  assumption.
  omega.
  
  apply fasIntro.
  apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply (IHk h).
  assert (Al : aligned8 x (x + 8 * S k)).
  exists (S k); reflexivity.
  assert (Ne : x <> (x + 8 * S k)).
  omega.
  replace (x + 8 * k) with (x + 8 * S k - 8).
  unfold hpredTrue.
  setoid_rewrite <- (rangeObjsEndRemove Al Ne).
  replace (x + 8 * S k - 8) with (x + 8 * k).
  apply Hp.
  omega.
  omega.

  intro I2; destruct (addrRangeElim I2) as [I2a [k' I2b]].
  omega.

  intro I2; destruct (set_add_elim I2) as [HA | HB]; clear I2.
  omega.
  destruct (addrRangeElim HB) as [I2a [k' I2b]].
  omega.
  omega.
  omega.

Qed.


Lemma objHpIsBuffer : forall h s0 x y,
  h ||- objHp s0 (rangeObjs x y) ->
    aligned8 x y ->
    h ||- buffer x y.
Proof.
  unfold objHp.
  intros h s0 x y Hp Al.
  apply objsIsBuffer.
  eapply fasImpl.
  apply Hp.
  simpl; unfold okObjAt; unfold okFieldAt; intros z h' Hp'.
  linLogElim Hp'.
  linLogIntro; eauto.
  assumption.
Qed.


Lemma bufferObjElim : forall h x y,
  h ||- buffer x y ->
  aligned8 x y -> x <> y ->
  h ||- x |->__ ** buffer (x + 8) y.
Proof.
  intros h x y Hp Al Ne.
  hypReplace Hp (bufferIsObjs _ Al Hp).
  assert (I : set_In x (rangeObjs x y)).
  destruct Al; eapply rangeObjsIntro.
  omega.
  instantiate (1 := 0); auto.
  hypReplace Hp (fasElim I Hp); clear I; cbv beta in Hp.
  sconjMatch.
  apply objsIsBuffer.
  replace (x + 8) with (8 + x); [idtac | omega].
  unfold hpredTrue; rewrite <- (rangeRemoveEqv x y); assumption.
  destruct Al as [k E]; subst y; destruct k.
  elimtype False; apply Ne; auto.
  exists k; omega.
Qed.



Lemma mapXSub : forall phi x y,
  mapX phi x = Some y ->
  forall phi', mapSub phi phi' ->
  mapX phi' x = Some y.
Proof.
  unfold mapX.
  intros phi x y H phi' S.
  destruct (isOdd x).
  assumption.
  eapply (mapSubSound (A := nat)); eauto.
Qed.


Lemma okFieldValWeaken : forall S x,
  okFieldVal S x ->
    forall S', set_subset S S' ->
      okFieldVal S' x.
Proof.
  intros S x [H | H] S' Sub; [left | right].
  assumption.
  apply Sub; assumption.
Qed.


Lemma okFieldAtWeaken : forall h S x,
  h ||- okFieldAt S x ->
    forall S', set_subset S S' ->
      h ||- okFieldAt S' x.
Proof.
  unfold okFieldAt; intros h S x Hp S' Sub.
  linLogElim Hp.
  linLogIntro; eauto.
  eapply okFieldValWeaken; eauto.
Qed.


Lemma okObjAtWeaken : forall h S x,
  h ||- okObjAt S x ->
    forall S', set_subset S S' ->
      h ||- okObjAt S' x.
Proof.
  unfold okObjAt.
  intros h S x Hp S' Sub.
  sconjMatch; eapply okFieldAtWeaken; eauto.
Qed.


Lemma objHpWeaken : forall h S X,
  h ||- objHp S X ->
    forall S', set_subset S S' ->
      h ||- objHp S' X.
Proof.
  unfold objHp.
  intros h S X Hp S' Sub.
  eapply fasImpl.
  apply Hp.
  clear h Hp; simpl.
  intros x h Hp.
  eapply okObjAtWeaken; eauto.
Qed.
Implicit Arguments objHpWeaken [h S X].


(* new attempt to describe GC guarantees *)
(* seems to work decently *)

Require Import Heap.
Import HeapMap.


Definition okObjHp rootVals objs :=
  !(forall v, set_In v rootVals -> okFieldVal objs v) **
  objHp objs objs.


Inductive reachable (h:heap) : nat -> nat -> Prop :=
  | reachRefl : forall x, reachable h x x
  | reachStep1 : forall x x' y, 
    Heap.get h x = Some x' ->
    reachable h x' y ->
    reachable h x y
  | reachStep2 : forall x x' y, 
    Heap.get h (4 + x) = Some x' ->
    reachable h x' y ->
    reachable h x y.

Definition rootReachable (h:heap) vv (x:nat) :=
  exists v, set_In v vv /\ reachable h v x.
  
Definition minObjHp rootVals objs :=
  (fun h => forall x:nat, set_In x objs -> rootReachable h rootVals x) &&
  okObjHp rootVals objs.


Add Morphism rootReachable with signature eq ==> set_eqv ==> eq ==> iff
  as rootReachMorph.
  unfold rootReachable.
  intros h s s' E n; constructor; intros [v [I R]]; exists v; constructor; auto.
  rewrite <- E; assumption.
  rewrite E; assumption.
Qed.


Add Morphism okObjHp with signature set_eqv ==> set_eqv ==> eq ==> iff
  as okObjHpMorph.
Proof.
  unfold okObjHp.
  intros s1 s1' E1 s2 s2' E2 h; constructor; intros [ha [hb [J [HA HB]]]]; lle HA;
    exists ha; exists hb; intuition.
  lli; auto.
  intros; rewrite <- E2; apply H; rewrite E1; assumption.
  unfold hpredTrue; rewrite <- E2; auto.
  lli; auto.
  intros; rewrite E2; apply H; rewrite <- E1; assumption.
  unfold hpredTrue; rewrite E2; auto.
Qed.


Add Morphism minObjHp with signature set_eqv ==> set_eqv ==> eq ==> iff 
  as minObjHpMorph.
  unfold minObjHp; intros s1 s1' E1 s2 s2' E2 h; constructor; intros [Hp Hp'];
    constructor.
  intros x I; rewrite <- E1; apply Hp; destruct (E2 x); auto.
  rewrite <- E1; rewrite <- E2; assumption.
  intros x I; rewrite E1; apply Hp; destruct (E2 x); auto.
  rewrite E1; rewrite E2; assumption.
Qed.


Lemma reachableTrans : forall h x y,
  reachable h x y -> forall z,
  reachable h y z ->
  reachable h x z.
Proof.
  intros h x y; induction 1; intros z R.
  assumption.
  hypReplace IHreachable (IHreachable _ R); clear R.
  econstructor 2; eauto.
  hypReplace IHreachable (IHreachable _ R); clear R.
  econstructor 3; eauto.
Qed.


Lemma minObjHpElim1 : forall h vv objs,
  h ||- minObjHp vv objs ->
  h ||- objHp objs objs.
Proof.
  unfold minObjHp; unfold okObjHp.
  intros h vv objs Hp.
  hypReplace Hp (aconjER Hp); lle Hp; assumption.
Qed.





Section Heapy.

  Require Import Heap.
  Import HeapMap.

  Lemma reachableJoinLeft : forall ha hb h,
    heap_join ha hb h -> forall y x,
      reachable ha y x ->
      reachable h y x.
  Proof.
    intros ha hb h J y x; induction 1.
    constructor.
    
    econstructor.
    instantiate (1 := x').
    eapply heapJoinSoundLeft; eauto.
    assumption.

    econstructor 3.
    instantiate (1 := x').
    eapply heapJoinSoundLeft; eauto.
    assumption.

  Qed.


  Lemma reachableJoinRight : forall ha hb h,
    heap_join ha hb h -> forall y x,
      reachable hb y x ->
      reachable h y x.
  Proof.
    intros ha hb h J y x; induction 1.
    constructor.
    
    econstructor.
    instantiate (1 := x').
    eapply heapJoinSoundRight; eauto.
    assumption.

    econstructor 3.
    instantiate (1 := x').
    eapply heapJoinSoundRight; eauto.
    assumption.

  Qed.


End Heapy.


Lemma minObjHpMerge : forall h rr rr' objs objs',
  h ||- minObjHp rr objs ** minObjHp rr' objs' ->
    h ||- minObjHp (rr |_| rr') (objs |_| objs').
Proof.
  unfold minObjHp.
  intros h rr rr' objs objs' Hp.
  hypReplace Hp (sconjDistrib1 Hp).
  destruct (aconjElim Hp) as [Hp1 Hp2]; clear Hp.
  hypReplace Hp1 (sconjDistrib2 Hp1).
  destruct (aconjElim Hp1) as [Hp1' _]; clear Hp1.
  hypReplace Hp2 (sconjDistrib2 Hp2).
  destruct (aconjElim Hp2) as [_ Hp2']; clear Hp2.
  apply aconjIntro.

  clear Hp2'; rename Hp1' into Hp.
  intros x I.
  destruct Hp as [ha [hb [J [HA HB]]]].
  destruct (set_union_elim I) as [I' | I']; clear I.
  clear HB.
  hypReplace HA (HA x I').
  destruct HA as [y [X Y]].
  exists y; constructor.
  apply set_union_intro1; assumption.
  eapply reachableJoinLeft; eauto.
  clear HA.
  hypReplace HB (HB x I').
  destruct HB as [y [X Y]].
  exists y; constructor.
  apply set_union_intro2; assumption.
  eapply reachableJoinRight; eauto.
  
  clear Hp1'; rename Hp2' into Hp.
  unfold okObjHp in Hp |- * ; linLogElim Hp.
  linLogIntro.
  
  intros v I; destruct (set_union_elim I) as [I' | I']; clear I.
  destruct (H1 _ I'); [left | right].
  assumption.
  apply set_union_intro1; assumption.
  destruct (H6 _ I'); [left | right].
  assumption.
  apply set_union_intro2; assumption.
  
  clear rr rr' H1 H6.
  apply objHpMerge; sconjMatch;
    eapply objHpWeaken; eauto; 
      intros x I; [apply set_union_intro1 | apply set_union_intro2]; assumption.

Qed.

Lemma minObjHpRootOk : forall h vv objs,
  h ||- minObjHp vv objs -> forall v,
    set_In v vv ->
    okFieldVal objs v.
Proof.
  intros h vv objs Hp v I.
  destruct (aconjElim Hp) as [_ X].
  unfold okObjHp in X; linLogElim X.
  auto.
Qed.


Definition objIsFwdedH phi h x' : hpred :=
  lexists v0', lexists v1',
  !(exists x, exists v0, exists v1,
    phi x = Some x' /\ 
    h ||- x |->, (v0::v1::nil) ** top /\ 
    mapX phi v0 = Some v0' /\ mapX phi v1 = Some v1') **
  x' |->, (v0'::v1'::nil).


Lemma minObjHpNilEmpty : forall h objs,
  h ||- minObjHp nil objs ->
    set_eqv objs empty_set.
Proof.
  unfold minObjHp.
  intros h objs Hp0.
  destruct (aconjElim Hp0) as [Hp _]; clear Hp0.
  destruct objs.
  apply set_eqvRefl.
  elimtype False.
  destruct (Hp a) as [v [I _]]; clear Hp.
  simpl; auto.
  simpl in I; assumption.
Qed.


Lemma noRootsEmptyMinObjHp : forall h objs h',
  h ||- (minObjHp nil objs && Heap.heap_eqv h') ->
  h' ||- unit.
Proof.
  intros h objs h' Hp.
  destruct (aconjElim Hp) as [Hp1 Hp2]; clear Hp.
  assert (X := minObjHpNilEmpty Hp1).
  destruct (aconjElim Hp1) as [_ Hp]; clear Hp1.
  eapply heapEqvTrue.
  intro x; apply eqvPredEqv; eauto.
  eapply objHpElimE; instantiate (1 := objs).
  unfold okObjHp in Hp; linLogElim Hp.
  unfold hpredTrue; rewrite <- X; assumption.
Qed.



Lemma subHeapReachable : forall h h',
    Heap.subHeap h h' -> forall x y,
    reachable h x y ->
    reachable h' x y.
Proof.
  intros h h' S x y R.
  induction R.
  constructor.

  econstructor 2.
  instantiate (1 := x').
  apply S; assumption.
  assumption.

  econstructor 3.
  instantiate (1 := x').
  apply S; assumption.
  assumption.

Qed.



Lemma objHpReachOkFieldL : forall h objs A,
  h ||- objHp objs objs ** A -> forall x y,
  reachable h x y ->
  okFieldVal objs x ->
  okFieldVal objs y.
Proof.
  intros h objs A Hp x y.
  induction 1; intros V.

  assumption.

  apply IHreachable.
  destruct V.
  unfold Heap.get in H; assert (X := isPtrOk x).
  destruct (isPtr x).
  elimtype False; eapply not_even_and_odd.
  2:apply H1.
  apply ptrPeven; assumption.
  discriminate H.

  assert (Hp' : h ||- (okObjAt objs x ** objHp objs (set_remove x objs)) ** A).
  sconjMatch.
  hypReplace Hp (objHpElim H1 Hp); sconjMatch.

  lle Hp'; clear Hp; rename Hp' into Hp.
  clear H1.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElim Hp.
  assert (X := contIntro2 _ _ H).
  assert (Y := contContEq Hp X); subst.
  assumption.

  apply IHreachable.
  destruct V.
  unfold Heap.get in H; assert (X := isPtrOk (4 + x)).
  destruct (isPtr (4 + x)).
  elimtype False; eapply not_even_and_odd.
  instantiate (1 := 4 + x).
  apply ptrPeven; assumption.
  simpl; constructor; constructor; constructor; constructor; assumption.
  discriminate H.

  assert (Hp' : h ||- (okObjAt objs x ** objHp objs (set_remove x objs)) ** A).
  sconjMatch; hypReplace Hp (objHpElim H1 Hp); sconjMatch.

  lle Hp'; clear Hp; rename Hp' into Hp.
  clear H1.
  unfold okObjAt in Hp; unfold okFieldAt in Hp; linLogElim Hp.
  assert (X := contIntro2 _ _ H).
  sconjLift2 Hp.
  assert (Y := contContEq Hp X); subst.
  assumption.

Qed.



Lemma objHpReachOkField : forall h objs,
  h ||- objHp objs objs -> forall x y,
  reachable h x y ->
  okFieldVal objs x ->
  okFieldVal objs y.
Proof.
  intros h objs Hp x y R Ok.
  eapply objHpReachOkFieldL; eauto.
  instantiate (1 := unit).
  lli; assumption.
Qed.


Lemma minObjHpSub : forall h objs A,
  h ||- objHp objs objs ** A -> forall vv objs',
  h ||- minObjHp vv objs' ** top ->
  (forall v, set_In v vv -> okFieldVal objs v) ->
    set_subset objs' objs.
Proof.
  intros h objs A Hp vv objs' Hp' V y I.
  unfold minObjHp in Hp'; cbv beta in Hp'.
  destruct (sconjExtract Hp') as [h' [Hp'' Hp2]]; clear Hp'.
  destruct (aconjElim Hp2) as [Hp3 Hp4]; clear Hp2.
  hypReplace Hp3 (Hp3 _ I); destruct Hp3 as [x [I' R]].
  hypReplace V (V _ I'); clear I'.

  assert (P : ptrP y).
  unfold okObjHp in Hp4.
  linLogElim Hp4.
  hypReplace Hp4 (objHpElim I Hp4); unfold okObjAt in Hp4; 
    unfold okFieldAt in Hp4; linLogElim Hp4.
  assert (X : h' ||- (!(ptrP y) ** top) ** top).
  sconjMatch.
  linLogIntro; eapply contElim1; eauto.
  linLogElim X; assumption.

  assert (reachable h x y).
  eapply subHeapReachable.
  2:eauto.
  eapply sconjEqSubHeap.
  instantiate (1 := top).
  sconjMatch.
  intro z.
  rewrite <- (Hp0 z).
  rewrite <- Heap.HeapMap.map_canonEqv; reflexivity.
  clear R Hp'' h' Hp4 I.
  assert (X := objHpReachOkFieldL Hp H V).
  destruct X.
  elimtype False; eapply not_even_and_odd.
  2:apply H0.
  apply ptrPeven; assumption.
  assumption.
Qed.


Section GCLemma.
  Require Import NatListSet.


  Require Import Heap.
  Require Import HeapDom.
  Import HeapMap.

  (* this should go somewhere else *)
  Lemma heapGetEqvMapEqv : forall h h',
    (forall x, get h x = get h' x) ->
    map_eqv h h'.
  Proof.
    intros h h' E x.
    hypReplace E (E (4 * x)).
    unfold get in E.
    assert (X := isPtrOk (4 * x)).
    destruct (isPtr (4 * x)).
    rewrite div44eq in E.
    unfold MapDef.get; assumption.
    elimtype False; apply X; apply (ptrP_intro x).
    reflexivity.
  Qed.


  Lemma fasHpSubDom : forall h objs,
    (h ||- foralls x in objs, x |->__) ->
      forall objs', h ||- (foralls x in objs', x |->__) ->
	set_subset objs objs'.
  Proof.
    unfold fas.
    intros h objs Hp objs' Hp' x I0.
    rewrite (set_canonEqv objs').
    assert (I : set_In x (set_canon objs)).
    rewrite <- (set_canonEqv objs); assumption.
    clear I0.
    assert (C := set_canonCanonP objs).
    assert (C' := set_canonCanonP objs').
    generalize dependent h.
    generalize dependent (set_canon objs').
    generalize dependent (set_canon objs).
    clear objs objs'.
    intros s I C; induction C; intros s' C' h Hp Hp'.

    elimtype False; assumption.

    destruct s'.
    simpl in Hp; simpl in Hp'.
    simpl; linLogElim Hp; eapply getUnitFalse; eauto.

    inversion_clear C'; rename H0 into C'; rename H1 into L'.
    rename H into L.
    simpl; simpl in I; simpl in Hp; simpl in Hp'; linLogElim Hp; linLogElim Hp'.

    assert (E : a = a0).
    clear x I IHC.
    assert (X := heap_get_In_dom a h).
    replace (get h a) with (Some x0) in X.
    assert (Deq : set_eqv (heap_dom h) ((set_add a0 empty_set) |_|
      (set_add (4 + a0) empty_set) |_|
	(set_mmap (fun x => (set_add x empty_set) |_| (set_add (4+x) empty_set)))
	  s')).
    clear Hp L L'.
    eapply sconjDom; eauto.
    intros; eapply contDom; eauto.
    intros; eapply sconjDom; eauto.
    intros; eapply contDom; eauto.
    intros; eapply fasDom.
    unfold fas.
    rewrite (set_canonCanonPEq C'); eauto.
    intros x h'1 HH; simpl in HH.
    linLogElim HH.
    eapply sconjDom; eauto.
    intros; eapply contDom; eauto.
    intros; eapply contDom; eauto.
    simpl in Deq.
    destruct (Deq a) as [Y _]; assert (I := Y X);
      clear X Y Deq.
    simpl in I.
    destruct I.
    auto.

    assert (X := heap_get_In_dom a0 h).
    replace (get h a0) with (Some x2) in X.
    assert (Deq : set_eqv (heap_dom h) ((set_add a empty_set) |_|
      (set_add (4 + a) empty_set) |_|
	(set_mmap (fun x => (set_add x empty_set) |_| (set_add (4+x) empty_set)))
	  s)).    
    clear Hp' L L'.
    eapply sconjDom; eauto.
    intros; eapply contDom; eauto.
    intros; eapply sconjDom; eauto.
    intros; eapply contDom; eauto.
    intros; eapply fasDom.
    unfold fas.
    rewrite (set_canonCanonPEq C); eauto.
    intros x h'1 HH; simpl in HH.
    linLogElim HH.
    eapply sconjDom; eauto;
      intros; eapply contDom; eauto.
    simpl in Deq.
    destruct (Deq a0) as [Y _]; assert (I := Y X);
      clear X Y Deq.
    simpl in I.
    destruct I.
    auto.
    elimtype False.
    destruct H.
    destruct H0.
    omega.
    destruct (set_mmap_elim _ _ _ H0).
    simpl in H1.
    destruct H1 as [II IB].
    assert (X := L _ II); clear L II.
    unfold POnat.POnat.LT in X.
    destruct IB.
    omega.
    destruct H1.
    omega.
    assumption.

    destruct H0.
    subst a0.
    destruct (set_mmap_elim _ _ _ H).
    simpl in H0.
    destruct H0 as [II IB].
    assert (X := L' _ II); clear L L' II; unfold POnat.POnat.LT in X.
    destruct IB.
    omega.
    destruct H0.
    omega.
    assumption.

    destruct (set_mmap_elim _ _ _ H); clear H.
    simpl in H1.
    destruct H1 as [II IB].
    assert (X := L' _ II); clear L' II; unfold POnat.POnat.LT in X.
    destruct (set_mmap_elim _ _ _ H0); clear H0 Hp Hp'.
    simpl in H.
    destruct H as [II' IB'].
    assert (X' := L _ II'); clear L II'; unfold POnat.POnat.LT in X'.
    intuition.

    symmetry; eapply contGetEqL; eauto.

    symmetry; eapply contGetEqL; eauto.

    subst a0.
    destruct I as [I | I]; [left | right].

    assumption.

    assert (X := contContEq Hp Hp').
    subst x2.
    assert (x1 = x3).
    sconjLift2 Hp; sconjLift2 Hp'; exact (contContEq Hp Hp').
    subst x3.

    hypReplace Hp (sconjAssocR Hp).
    hypReplace Hp' (sconjAssocR Hp').

    destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
    destruct (sconjElim Hp') as [ha' [hb' [J' [HA' HB']]]]; clear Hp'.
    
    hypReplace IHC (IHC I s' C' hb); clear I C C'.
    apply IHC.
    assumption.
    clear IHC L L'.
    unfold hpredTrue.
    cut (map_eqv hb hb').
    unfold heap_canon.
    intro Z; rewrite (map_canonEqvEq Z).
    assumption.

    apply heapGetEqvMapEqv.
    intro y.

    assert (get ha y = get ha' y).
    assert (D : set_eqv (heap_dom ha) ((set_add a empty_set) |_| (set_add (4+a) empty_set))).
    intros; eapply sconjDom; eauto;
      intros; eapply contDom; eauto.
    assert (D' : set_eqv (heap_dom ha') ((set_add a empty_set) |_| (set_add (4+a) empty_set))).
    intros; eapply sconjDom; eauto;
      intros; eapply contDom; eauto.
    compare y a; intro R.
    subst y.
    assert (X := contGetEqL HA).
    assert (Y := contGetEqL HA').
    rewrite X; rewrite Y; reflexivity.
    compare y (4 + a); intro R'.
    subst y.
    sconjLift2of2 HA; assert (X := contGetEqL HA).
    sconjLift2of2 HA'; assert (Y := contGetEqL HA').
    simpl; rewrite X; rewrite Y; reflexivity.
    replace (get ha y) with (None (A := nat)).
    replace (get ha' y) with (None (A := nat)).
    reflexivity.

    symmetry; eapply notInHeapDomNone.
    intro I.
    destruct (D' y) as [X _].
    assert (Y := X I); clear X I.
    simpl in Y.
    intuition.

    symmetry; eapply notInHeapDomNone.
    intro I.
    destruct (D y) as [X _].
    assert (Y := X I); clear X I.
    simpl in Y.
    intuition.

    assert (H1 := mapJoinSoundLeft J (w := div4 y)).
    assert (H1' := mapJoinSoundLeft J' (w := div4 y)).
    assert (H2 := mapJoinSoundRight J (w := div4 y)).
    assert (H2' := mapJoinSoundRight J' (w := div4 y)).
    assert (N := mapJoinSoundNone J (w := div4 y)).
    assert (N' := mapJoinSoundNone J' (w := div4 y)).
    assert (Disj1 := joinedMapsDisjointL J (w := div4 y)).
    assert (Disj1' := joinedMapsDisjointL J' (w := div4 y)).
    assert (Disj2 := joinedMapsDisjointR J (w := div4 y)).
    assert (Disj2' := joinedMapsDisjointR J' (w := div4 y)).
    unfold inDom in * .
    unfold get in H |- * .
    destruct (isPtr y).
    2:reflexivity.
    unfold MapDef.get in * .
    destruct (map_get ha (div4 y)).
    clear N.
    assert (map_get hb (div4 y) = None).
    destruct (map_get hb (div4 y)).
    elimtype False; apply Disj1; econstructor; reflexivity.
    reflexivity.
    assert (map_get hb' (div4 y) = None).
    destruct (map_get hb' (div4 y)).
    rewrite <- H in Disj1'.
    elimtype False; apply Disj1'; econstructor; reflexivity.
    reflexivity.
    rewrite H0; rewrite H3; reflexivity.
    clear H1 H1' Disj2 Disj2'.
    rewrite <- H in N'; rewrite <- H in Disj1'; clear H.
    destruct (map_get hb (div4 y)).
    clear N Disj1 Disj1'.
    assert (E : map_get h (div4 y) = Some b).
    auto.
    rewrite E in H2'; rewrite E in N'; clear E H2.
    destruct (map_get hb' (div4 y)); eauto.
    assert (E : map_get h (div4 y) = None).
    eauto.
    rewrite E in H2'; rewrite E in N'.
    clear H2 N Disj1 Disj1' E.
    destruct (map_get hb' (div4 y)); eauto.
  Qed.


  Lemma fasHpEqDom : forall h objs,
    (h ||- foralls x in objs, x |->__) ->
      forall objs', h ||- (foralls x in objs', x |->__) ->
	set_eqv objs objs'.
  Proof.
    intros h objs Hp objs' Hp'.
    intros y; constructor; intro.
    apply (fasHpSubDom _ _ Hp _ Hp' _ H).
    apply (fasHpSubDom _ _ Hp' _ Hp _ H).
  Qed.


  Lemma objHpEqObjs : forall h objs objs0 objs' objs0',
    h ||- objHp objs0 objs -> h ||- objHp objs0' objs' ->
      set_eqv objs objs'.
  Proof.
    unfold objHp; unfold okObjAt; unfold okFieldAt; intros; eapply fasHpEqDom.
    instantiate (1 := h); eapply fasImpl.
    apply H.
    simpl; intros x hh Hp; linLogElim Hp; linLogIntro; sconjMatch.
    eapply fasImpl.
    apply H0.
    simpl; intros x hh Hp; linLogElim Hp; linLogIntro; sconjMatch.
  Qed.


End GCLemma.


Definition fieldCopied x s y :=
  lexists v, !(heapOf s ||- x |-> v ** top) ** y |-> v.

Definition objCopied x s y :=
  fieldCopied x s y ** fieldCopied (4+x) s (4+y).


Lemma objCopiedRefl : forall s A B C, s |||- objHp A B ** C ->
  s |||- (objHp A B && foralls x in B, objCopied x s x) ** C.
Proof.
  intros s A B C Hp.
  destruct (sconjElim Hp) as [h1 [h2 [J [H1 H2]]]].
  eapply sconjIntro.
  3:eauto.
  2:assumption.

  constructor.
  assumption.
  clear H2.
  unfold objHp in H1.
  apply (fasImplIS H1); clear H1.
  intros x h' IB Sub0.
  assert (Sub : Heap.subHeap h' (heapOf s)).
  eapply Heap.subHeapTrans.
  eauto.
  eapply Heap.heap_joinSub; eauto.
  clear Sub0 J h2.
  rename Hp into Hp0.
  unfold okObjAt; unfold okFieldAt; intro Hp.
  linLogElim Hp.
  assert (Hp' := subHeapImpl Hp Sub); linLogElim Hp'.
  unfold objCopied; unfold fieldCopied; linLogIntro.
  sconjMatch.
  eauto.
  sconjLift2 Hp'; sconjMatch.

Qed.
Implicit Arguments objCopiedRefl [s A B C].


