Require Export NatListSet.
Require Import Heap.
Require Import List.
Require Import OptUtil.
Require Import NatUtil.


(** hypReplace, but with an explicit new name *)
Ltac hypReplaceN H y H' :=
  assert (H' := y); try (clear H; rename H' into H).

(** replace H with y in the hypotheses, keeping the name if possible *)
Ltac hypReplace H y :=
  let H' := fresh in
    (hypReplaceN H y H').

Ltac hyr := hypReplace.



Definition nset := NatListSetM.set.

Import HeapMap.

(** Here we define heap predicates that are supposed to implement linear and
   separation logic connectives, and their properties.  The only linear assumption
   allowed are heap bindings.
*)

Set Implicit Arguments.

Notation hpred := (heap -> Prop).

Definition hpredTrue (h:heap) (A:hpred) := A (heap_canon h).

Notation "h ||- A" := (hpredTrue h A) (at level 80).


Lemma heapEqvTrue : forall h h',
  Heap.heap_eqv h h' -> forall A,
  h' ||- A ->
  h ||- A.
Proof.
  unfold hpredTrue.
  intros h h' E Hp A.
  unfold heap_canon.
  rewrite (Heap.HeapMap.map_canonEqvEq E).
  assumption.
Qed.


Lemma hpredTrueEq : forall h A, (h ||- A) = A (heap_canon h).
Proof.
  reflexivity.
Qed.


(** *** Standard linear logic connectives *)


(** simultaneous conjuction *)
Definition sconj (A B:hpred) : hpred :=
  fun h => exists ha, exists hb,
    heap_join (heap_canon ha) (heap_canon hb) h /\
    ha ||- A /\ hb ||- B.

Infix "**" := sconj (at level 60, right associativity).


(** alternative conjuction *)
Definition aconj (A B:hpred) : hpred :=
  fun h => A h /\ B h.

Infix "&&" := aconj (at level 60, right associativity).


(** disjunction *)
Definition ldisj (A B:hpred) : hpred :=
  fun h => A h \/ B h.

Infix "+++" := ldisj (at level 60, right associativity).


(** unit *)
Definition unit : hpred := heap_eqv empty_heap.


(** top *)
Definition top : hpred :=
  fun s => True.


(** exists *)
Definition lex (A:Set) (f:A -> hpred) : hpred :=
  fun s => exists x, f x s.

Notation "'lexists' x , A" := (lex (fun x => A))
  (at level 200, x ident).

Definition lexT (A:Type) (f:A -> hpred) : hpred :=
  fun s => exists x, f x s.

Notation "'lexistsT' x , A" := (lexT (fun x => A))
  (at level 200, x ident).


(** intuitionistic truth *)
Definition bang (A:Prop) : hpred :=
  fun s => unit s /\ A.

Notation "! A" := (bang A) (at level 55).



(** *** Separation logic connectives *)

(** cont l x holds if heap is just the binding from l to x *)
Definition cont (l:nat) (x:nat) : hpred :=
  fun h => ptrP l /\ heap_eqv h (singleton_heap l x).
Infix "|->" := cont (at level 59, right associativity).


(** fas implements the iterated separating conjunction,
   written \forall_\ast x \in s. p, from "Local Reasoning 
   about a Copying Garbage Collector"by Lars Birkedal, Noah 
   Torp-Smith and John C. Reynolds *)

Fixpoint fas' (p:nat -> hpred) (s:nset) {struct s} : hpred :=
  match s with
    | List.nil => unit
    | List.cons x s' => (p x) ** (fas' p s')
  end.

Definition fas p s := fas' p (set_canon s).

Notation "'foralls' x 'in' s , A" := (fas (fun x => A) s) 
  (at level 200, x ident).

(** The set cannot have any duplicates to ensure that equivalent sets
   don't have different behavior.  We go further and canonicalize the
   set to make reasoning about fas easier.
*)


(** some standard separation logic abbreviations *)
Notation addrOffset := 4.

(** l |-> x1, ..., xn *)
(**  The special case for lists of length 1 avoids having to
     worry about units floating around *)
Fixpoint contL (l:nat) (xx:list nat) {struct xx} : hpred :=
  match xx with
    | nil => unit
    | x::nil => l |-> x
    | x::xx' => l |-> x ** contL (addrOffset+l) xx'
  end.


Infix "|->," := contL (at level 59, right associativity).


Definition pairCont x v0 v1 := x |->, (v0::v1::nil).
Infix "|->>" := pairCont (at level 59, right associativity).


(** l |-> - *)
Notation "l |->_" := (lexists xyzyzy, l |-> xyzyzy) (at level 59).
(** l |-> -, - *)
Notation "l |->__" := (((l) |->_) ** ((addrOffset+l) |->_)) (at level 59).
(** l |-> -, -, - *)
Notation "l |->___" := (((l) |->_) ** ((addrOffset+l) |->__)) (at level 59).
(** l |-> -, -, -, - *)
Notation "l |->____" := (((l) |->_) ** ((addrOffset+l) |->___)) (at level 59).


(** *** Basic introductions and eliminations *)

Lemma sconjIntro : forall ha hb h A B,
  ha ||- A -> hb ||- B -> heap_join ha hb h ->
    h ||- A ** B.
Proof.
  intros ha hb h A B PA PB J.
  exists ha; exists hb.
  constructor.
  apply mapJoinCanon.
  assumption.
  auto.
Qed.
Implicit Arguments sconjIntro [ha hb h A B].


Lemma sconjElim : forall h A B,
  h ||- A ** B ->
    exists h1, exists h2, heap_join h1 h2 h /\ h1 ||- A /\ h2 ||- B.
Proof.
  destruct 1 as [h1 [h2 H]].
  exists h1; exists h2.
  intuition.
  eapply mapJoinEqv; eauto; apply map_eqvSymm; apply map_canonEqv.
Qed.


Lemma aconjIntro : forall h A B,
  h ||- A -> h ||- B ->
    h ||- A && B.
Proof.
  unfold aconj; constructor; assumption.
Qed.
Implicit Arguments aconjIntro [h A B].


Lemma aconjEL : forall h A B,
  h ||- A && B ->
    h ||- A.
Proof.
  unfold aconj; destruct 1; assumption.
Qed.

Lemma aconjER : forall h A B,
  h ||- A && B ->
    h ||- B.
Proof.
  unfold aconj; destruct 1; assumption.
Qed.


Lemma aconjElim : forall h A B,
  h ||- A && B -> 
    h ||- A /\ h ||- B.
Proof.
  intros h A B Hp.
  constructor.
  exact (aconjEL Hp).
  exact (aconjER Hp).
Qed.


Lemma unitIntro : empty_heap ||- unit.
Proof.
  unfold unit.
  unfold hpredTrue; apply map_eqvRefl.
Qed.

Lemma unitElim : forall h,
  h ||- unit ->
    heap_eqv h empty_heap.
Proof.
  unfold unit.
  intros h H.
  apply map_eqvSymm.
  eapply map_eqvTrans.
  apply H.
  apply map_eqvSymm.
  apply map_canonEqv.
Qed.


Lemma topIntro : forall h,
  h ||- top.
Proof.
  unfold top; unfold hpredTrue; trivial.
Qed.


Lemma lexI : forall h (A:Set) (f:A->hpred) x,
  h ||- f x ->
    h ||- lex f.
Proof.
  intros.
  exists x.
  assumption.
Qed.
Implicit Arguments lexI [h A f].


Lemma lexI' : forall h (A:Set) (f:A->hpred),
  (exists x, h ||- f x) ->
    h ||- lex f.
Proof.
  intros h A f [x Hp].
  apply (lexI x Hp).
Qed.


Lemma lexE : forall h (A:Set) (f:A->hpred),
  h ||- lex f -> 
    exists x, h ||- f x.
Proof.
  unfold lex; auto.
Qed.


Lemma bangIntro : forall h (A:Prop),
  h ||- unit -> A ->
    h ||- ! A.
Proof.
  unfold bang.
  intros h A Hp Aok.
  constructor; assumption.
Qed.


Lemma bangElim1 : forall h A,
  h ||- ! A -> A.
Proof.
  unfold bang.
  unfold hpredTrue.
  intuition.
Qed.


Lemma bangElim2 : forall h A,
  h ||- ! A -> h ||- unit.
Proof.
  unfold bang; unfold hpredTrue.
  intuition.
Qed.


Lemma bangElim : forall h A,
  h ||- ! A ->
    h ||- unit /\ A.
Proof.
  intros h A Hp.
  constructor.
  exact (bangElim2 Hp).
  exact (bangElim1 Hp).
Qed.


Lemma contIntro : forall l w,
  ptrP l ->
  singleton_heap l w ||- l |-> w.
Proof.
  intros.
  constructor.
  assumption.
  apply map_eqvSymm; apply map_canonEqv.
Qed.


Lemma contElim : forall h l w,
  h ||- l |-> w ->
    ptrP l /\ heap_eqv h (singleton_heap l w).
Proof.
  unfold cont.
  intros h l w [H1 H2].
  constructor.
  assumption.
  eapply map_eqvTrans.
  apply map_canonEqv.
  assumption.
Qed.


Lemma contElim1 : forall l w h,
  h ||- l |-> w ->
    ptrP l.
Proof.
  unfold cont; unfold hpredTrue.
  intuition.
Qed.


(** implications between sconjs *)

Notation "A ==> B" :=
  (forall h, h ||- A -> h ||- B) (at level 13).


Lemma sconjImplL : forall h A B,
  h ||- (A ** B) -> 
    forall A', (A ==> A') -> 
    h ||- (A' ** B).
Proof.
  destruct 1 as [ha [hb [J [HA HB]]]].
  exists ha; exists hb.
  intuition.
Qed.


Lemma sconjImplR : forall h A B,
  h ||- (A ** B) -> 
    forall B', (B ==> B') -> 
    h ||- (A ** B').
Proof.
  destruct 1 as [ha [hb [J [HA HB]]]].
  exists ha; exists hb.
  intuition.
Qed.


Lemma sconjImpl : forall h A B,
  h ||- A ** B -> forall A' B',
    (A ==> A') -> (B ==> B') -> 
    h ||- A' ** B'.
Proof.
  intros.
  assert (h ||- (A' ** B)).
  eapply sconjImplL; [eauto | assumption].
  eapply sconjImplR; [eauto | assumption].
Qed.


Lemma sconjImplS : forall h A B,
  h ||- A ** B -> forall A' B',
    (forall h', subHeap h' h -> h' ||- A -> h' ||- A') ->
    (forall h', subHeap h' h -> h' ||- B -> h' ||- B') ->
    h ||- A' ** B'.
Proof.
  intros h A B Hp A' B' Aimp Bimp.
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]].
  assert (JA := heap_joinSub J).
  assert (JB := heap_joinSub (Heap.HeapMap.ABJoin.mapJoinComm J)).
  eapply sconjIntro; eauto.
Qed.


Lemma sconjifierL : forall A (P:Prop),
  (forall h, h ||- A -> P) ->
  forall h B,
    h ||- A ** B -> P.
Proof.
  intros A P0 AP h B H.
  destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]].
  eapply AP; eauto.
Qed.


Lemma sconjifierR : forall A (P:Prop),
  (forall h, h ||- A -> P) ->
  forall h B,
    h ||- B ** A -> P.
Proof.
  intros A P0 AP h B H.
  destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]].
  eapply AP; eauto.
Qed.



(** properties of sconj *)

Lemma sconjComm : forall A B s,
  s ||- A ** B -> s ||- B ** A.
Proof.
  intros.
  destruct (sconjElim H) as [s1 [s2 [J [HA HB]]]].
  eapply sconjIntro; eauto.
  apply mapJoinComm; assumption.
Qed.


Lemma sconjAssocL : forall A B C s,
  s ||- (A ** B) ** C -> s ||- A ** (B ** C).
Proof.
  intros.
  destruct (sconjElim H) as [sab [sc [Jabc [HAB HC]]]]; clear H.
  destruct (sconjElim HAB) as [sa [sb [Jab [HA HB]]]]; clear HAB.
  destruct (mapJoinAssoc Jab Jabc) as [sbc [Jabc' Jbc]].
  eapply sconjIntro; eauto.
  eapply sconjIntro; eauto.
Qed.


Lemma sconjAssocR : forall A B C s,
  s ||- A ** (B ** C) -> s ||- (A ** B) ** C.
Proof.
  intros.
  destruct (sconjElim H) as [sa [sbc [Jabc [HA HBC]]]].
  destruct (sconjElim HBC) as [sb [sc [Jbc [HB HC]]]].
  destruct (mapJoinAssoc (mapJoinComm Jbc) (mapJoinComm Jabc)) as [sab [Jcab Jba]].
  assert (Jabc' := mapJoinComm Jcab).
  assert (Jab := mapJoinComm Jba).
  eapply sconjIntro; eauto.
  eapply sconjIntro; eauto.
Qed.


Lemma sconjExtract : forall h A B,
  h ||- A ** B ->
    exists h0, h ||- heap_eqv h0 ** B /\ h0 ||- A.
Proof.
  intros h A B H.
  destruct (sconjElim H) as [ha [hb [J [HA HB]]]].
  exists ha.
  constructor.
  assert (heap_eqv ha (Heap.heap_canon ha)).
  apply map_canonEqv.
  exact (sconjIntro H0 HB J).
  assumption.
Qed.


Lemma sconjInsert : forall h h' B,
  h ||- Heap.heap_eqv h' ** B ->
    forall A, h' ||- A ->
      h ||- A ** B.
Proof.
  intros h ha' B H A HA.
  destruct (sconjElim H) as [ha [hb [J [HA' HB]]]].

  assert (J' : heap_join ha' hb h).
  eapply mapJoinEqv.
  apply J.
  apply map_eqvSymm; eapply map_eqvTrans.
  apply HA'.
  apply map_eqvSymm; apply map_canonEqv.
  apply map_eqvRefl.
  apply map_eqvRefl.

  exact (sconjIntro HA HB J').

Qed.
Implicit Arguments sconjInsert [h h' B A].


(** sconj and friends *)

Lemma lexEL  : forall (T:Set) h (A:T -> hpred) B,
  h ||- (lex A) ** B -> h ||- lexists x, (A x) ** B.
Proof.
  intros T h A B H.
  destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]].
  destruct (lexE HA) as [x HA']; clear HA.
  eapply lexI; eapply sconjIntro; eauto.
Qed.


Lemma lexER : forall (T:Set) h A (B:T -> hpred),
  h ||- A ** (lex B) -> h ||- lexists x, A ** (B x).
Proof.
  intros T h A B H.
  destruct (lexEL (sconjComm H)) as [x H'].
  exists x.
  apply sconjComm; assumption.
Qed.


Lemma lexIL : forall (T:Set) h A (B:T -> hpred),
  h ||- (lexists x, (B x) ** A) -> h ||- (lex B) ** A.
Proof.
  intros T h A B Hp.
  destruct (lexE Hp) as [x Hp'].
  eapply sconjImplL; eauto.
  intros; eapply lexI; eauto.
Qed.


Lemma lexIR : forall (T:Set) h A (B:T -> hpred),
  h ||- (lexists x, A ** (B x)) -> h ||- A ** (lex B).
Proof.
  intros T h A B Hp.
  destruct (lexE Hp) as [x Hp'].
  eapply sconjImplR; eauto.
  intros; eapply lexI; eauto.
Qed.


Lemma unitIL : forall A h,
  h ||- A -> h ||- unit ** A.
Proof.
  intros A h H.
  eapply sconjIntro with (ha := empty_heap) (hb := h).
  apply unitIntro.
  assumption.
  apply joinUnitIntro.
Qed.
Implicit Arguments unitIL [A h].


Lemma unitIR : forall A h,
  h ||- A -> h ||- A ** unit.
Proof.
  intros A h H.
  exact (sconjComm (unitIL H)).
Qed.
Implicit Arguments unitIR [A h].


Lemma unitEL : forall A h,
  h ||- unit ** A -> h ||- A.
Proof.
  intros A h H.
  destruct H as [ha [hb [J [HA HB]]]].
  assert (map_eqv hb (heap_canon h)).
  apply joinUnitLeft.
  eapply mapJoinEqv.
  apply J.
  unfold unit in HA; apply map_eqvSymm; assumption.
  apply map_eqvSymm; apply map_canonEqv.
  apply map_eqvRefl.
  assert (R := map_canonEqvEq H); clear H.
  unfold hpredTrue.
  unfold heap_canon in * |- * .
  rewrite <- (map_canonCanonEq h).
  rewrite <- R.
  assumption.
Qed.


Lemma unitER : forall A h,
  h ||- A ** unit -> h ||- A.
Proof.
  intros A h H.
  apply unitEL.
  apply sconjComm.
  assumption.
Qed.


Theorem sconjDistrib1 : forall h A B C,
  h ||- A ** (B && C) ->
    h ||- (A ** B) && (A ** C).
Proof.
  intros.
  destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]].
  apply aconjIntro.
  assert (HB' := aconjEL HB); eapply sconjIntro; eauto.
  assert (HB' := aconjER HB); eapply sconjIntro; eauto.
Qed.


Theorem sconjDistrib2 : forall h A B C,
  h ||- (A && B) ** C ->
    h ||- (A ** C) && (B ** C).
Proof.
  intros h A B C Hp.
  hypReplace Hp (sconjComm Hp).
  destruct (sconjDistrib1 Hp) as [Hp1 Hp2].
  constructor; apply sconjComm; assumption.
Qed.


Lemma bangEL : forall h A B,
  h ||- ! A ** B ->
    A /\ h ||- B.
Proof.
  intros h A B H.
  assert (A /\ h ||- unit ** B).
  destruct (sconjElim H) as [ha [hb [J [HA HB]]]].
  constructor.
  apply (bangElim1 HA).
  assert (HA' := bangElim2 HA).
  eapply sconjIntro; eauto.
  intuition.
  exact (unitEL H2).
Qed.


Lemma bangEL1 : forall h A B,
  h ||- ! A ** B ->
  A.
Proof.
  intros h A B H.
  destruct (bangEL H).
  assumption.
Qed.


Lemma bangEL2 : forall h A B,
  h ||- ! A ** B ->
  h ||- B.
Proof.
  intros h A B H.
  destruct (bangEL H).
  assumption.
Qed.


Lemma bangIL : forall h (A:Prop) B,
  A -> h ||- B ->
  h ||- ! A ** B.
Proof.
  intros h A B HA HB.
  unfold bang.
  assert (Q := unitIL HB).
  eapply sconjImpl; unfold hpredTrue; eauto.
  assumption.
Qed.


Lemma bangIR : forall h (A:Prop) B,
  A -> h ||- B ->
  h ||- B ** ! A.
Proof.
  intros.
  apply sconjComm.
  apply bangIL; auto.
Qed.


Theorem fasIntroE : forall p h,
  h ||- unit ->
    h ||- fas p empty_set.
Proof.
  intros.
  unfold fas.
  simpl.
  assumption.
Qed.


Lemma fasEIntroL : forall h A,
  h ||- A -> forall P,
    h ||- fas P empty_set ** A.
Proof.
  intros h A Hp P.
  hypReplace Hp (unitIL Hp).
  apply (sconjImplL Hp); clear h Hp; intros h Hp.
  apply fasIntroE; assumption.
Qed.


Lemma fasEIntroR : forall h A,
  h ||- A -> forall P,
    h ||- A ** fas P empty_set.
Proof.
  intros h A Hp P.
  apply sconjComm.
  apply fasEIntroL; assumption.
Qed.


(** basic properties of cont *)


Lemma contGetEq : forall h l w,
  h ||- l |-> w -> 
    get h l = Some w.
Proof.
  intros h l w H.
  destruct (contElim H) as [Q R].
  assert (R2 := R (div4 l)).
  replace MapDef.get with map_get in R2; clear R.
  2:reflexivity.
  unfold get.
  rewrite (isPtrEq Q).
  rewrite R2.
  unfold singleton_heap; apply singletonGetEq.
Qed.


Theorem sconjDisj : forall x y z h,
  h ||- x |-> y ** x |-> z ->
    False.
Proof.
  intros.
  destruct (sconjElim H) as [h1 [h2 [J [H1 H2]]]]; clear H.

  assert (X1 : inDom (div4 x) h1).
  assert (X := contGetEq H1).
  assert (Y := contElim1 H1).
  unfold Heap.get in X.
  rewrite (isPtrEq Y) in X.
  exists y; assumption.

  assert (X2 : inDom (div4 x) h2).
  assert (X := contGetEq H2).
  assert (Y := contElim1 H2).
  unfold Heap.get in X.
  rewrite (isPtrEq Y) in X.
  exists z; assumption.

  exact (joinedMapsDisjointL J X1 X2).

Qed.


(** various lemmas, mostly for the tactics *)

Lemma topTopIntro : forall h,
  h ||- top ->  h ||- top ** top.
Proof.
  intros h _.
  unfold top.
  eapply sconjIntro; unfold hpredTrue; trivial.
  apply joinUnitIntro.
Qed.


Lemma topTopIL : forall h A,
  h ||- top ** A ->
    h ||- top ** top ** A.
Proof.
  intros h A Hp.
  apply sconjAssocL.
  apply (sconjImplL Hp); intros h' Hp'.
  exact (topTopIntro Hp').
Qed.


Lemma topTopEL : forall h A,
  h ||- top ** top ** A ->
    h ||- top ** A.
Proof.
  intros h A Hp.
  hypReplace Hp (sconjAssocR Hp).
  apply (sconjImplL Hp); intros h' Hp'.
  apply topIntro.
Qed.


Lemma sconjDisjL : forall x y z h A,
  h ||- x |-> y ** x |-> z ** A ->
    False.
Proof.
  intros x y z h A Hp.
  assert (Hp' : h ||- !False ** top).
  hypReplace Hp (sconjAssocR Hp).
  apply (sconjImpl Hp); clear h Hp; intros h Hp.
  elimtype False; apply (sconjDisj Hp).
  apply topIntro.
  exact (bangEL1 Hp').
Qed.


Lemma contLNilIL : forall h A,
  h ||- A -> forall x,
  h ||- x |->, nil ** A.
Proof.
  intros; simpl; apply unitIL; assumption.
Qed.
Implicit Arguments contLNilIL [h A].


Lemma contLNilEL : forall h x A,
  h ||- x |->, nil ** A ->
  h ||- A.
Proof.
  intros; simpl; apply unitEL; assumption.
Qed.


Lemma contLConsIL : forall h x y l A,
  h ||- x |-> y ** 4 + x |->, l ** A ->
  h ||- x |->, (y::l) ** A.
Proof.
  destruct l; simpl; intros A Hp.
  apply (sconjImplR Hp); intros; apply unitEL; assumption.
  apply sconjAssocR; assumption.
Qed.
Implicit Arguments contLConsIL [h x y l A].


Lemma contLConsEL : forall h x y l A,
  h ||- x |->, (y::l) ** A ->
  h ||- x |-> y ** 4 + x |->, l ** A.
Proof.
  destruct l; simpl; intros A Hp.
  apply (sconjImplR Hp); intros; apply unitIL; assumption.
  apply sconjAssocL; assumption.
Qed.
Implicit Arguments contLConsEL [h x y l A].



Lemma sconjLift2Ok : forall h A B C,
  h ||- A ** B ** C -> 
    h ||- B ** A ** C.
Proof.
  intros h A B C H.
  apply sconjComm; apply sconjAssocR.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjComm; assumption.
Qed.

Ltac sconjLift2 H :=
  hypReplace H (sconjLift2Ok H).


(** This tactic attempts to apply all the introductions to the goal.
Intuitionistic subgoals are extracted, existentials are
econstructored, units are eliminated, tops are "moved" all the way to
the right and combined, sconjs are converted into right normal form,
fas over empty sets are eliminated, and so on.

This tactic fails if no changes can be made.  This prevents large
subtrees of sconjs from being needlessly taken apart and reassembled. *)

Ltac linLogIntro' :=
  match goal with
    | |- _ ||- top => fail
    | |- _ ||- unit => assumption
    | |- _ ||- bang _ => apply bangIntro
    | |- _ ||- lex _ => apply lexI'; econstructor; try linLogIntro'
    | |- _ ||- sconj (fas _ empty_set) _ => apply fasEIntroL; try linLogIntro'
    | |- _ ||- sconj _ (fas _ empty_set) => apply fasEIntroR; try linLogIntro'
    | |- _ ||- sconj (contL _ nil) _ => apply contLNilIL; try linLogIntro'
    | |- _ ||- sconj (contL _ _) _ => apply contLConsIL; try linLogIntro'
    | |- _ ||- sconj _ unit => apply unitIR; try linLogIntro'
    | |- _ ||- sconj unit _ => apply unitIL; try linLogIntro'
    | |- _ ||- sconj (bang _) _ => apply bangIL; [auto | try linLogIntro']
    | |- _ ||- sconj _ (bang _) => apply bangIR; [auto | try linLogIntro']
    | |- _ ||- sconj (lex _) _ => apply lexIL; try linLogIntro'
    | |- _ ||- sconj _ (lex _) => apply lexIR; try linLogIntro'
    | |- _ ||- sconj (sconj _ _) _ => apply sconjAssocR; try linLogIntro'
      (* add case for just bare contL? *)
    | |- _ ||- sconj top ?A => 
      match A with
	| top => apply topTopIntro; try linLogIntro'
	| sconj ?B _ =>
	  match B with
	    | top => apply topTopIL; try linLogIntro'
	    | _ => apply sconjLift2Ok; try linLogIntro'
	  end
	| _ => apply sconjComm; try linLogIntro'
      end
    | |- _ ||- sconj _ _ => sconjSubIntro
  end
with sconjSubIntro :=
  let h := fresh "h" in 
  let H := fresh in
    first
    [eapply sconjImpl; 
      [idtac |
	intros h H; linLogIntro'; try (apply H) |
	intros h H; try linLogIntro'; try (apply H)] |
      eapply sconjImpl;
	[idtac |
	  intros h H; try (apply H) |
	    intros h H; linLogIntro'; try (apply H)]].

Ltac linLogIntro :=
  first [linLogIntro'; try apply topIntro |
    apply topIntro].

Ltac lli := linLogIntro.


(** This tactic attempts to apply elimination rules to the specified
hypotheses.  Some of this is similar to linLogIntro: sconjs are
converted to right normal form, all tops are moved to the right and
combined, units are removed, existentials are unpacked.  In addition, in cases like 
"x |-> y ** x |-> z", the goal is immediately solved. *)

Ltac linLogElim Hp :=
  let Hp' := fresh "Hp" in
    (match type of Hp with
       | _ ||- sconj (sconj _ _) _ => assert (Hp' := sconjAssocL Hp)
       | _ ||- sconj unit _ => assert (Hp' := unitEL Hp)
       | _ ||- sconj _ unit => assert (Hp' := unitER Hp)
       | _ ||- sconj (contL _ nil) _ => assert (Hp' := contLNilEL Hp)
       | _ ||- sconj (contL _ _) _ => assert (Hp' := contLConsEL Hp)
       | ?h ||- sconj top ?A => 
	 match A with 
	   | top => assert (Hp' := topIntro h)
	   | sconj ?B _ =>
	     match B with
 	       | top => assert (Hp' := topTopEL Hp)
	       | _ => assert (Hp' := sconjLift2Ok Hp)
	     end
	   | _ => assert (Hp' := sconjComm Hp)
	 end
       | _ ||- lex _ => 
	 let x := fresh "x" in
           destruct (lexE Hp) as [x Hp']
       | _ ||- bang _ =>
	 let HA := fresh in
	   (assert (HA := bangElim1 Hp);
	    assert (Hp' := bangElim2 Hp))
       | _ ||- sconj (cont ?x ?y) (cont ?x ?z) =>
	 elimtype False; exact (sconjDisj Hp)
       | _ ||- sconj (cont ?x ?y) (sconj (cont ?x ?z) _) =>
	 elimtype False; exact (sconjDisjL Hp)
       | _ ||- sconj _ _ => sconjSubSimpl Hp Hp'
       | _ ||- aconj _ _ => (* i think this case is broken *)
	 let HA := fresh in
         let HB := fresh in
	     (assert (HA := aconjEL Hp);
	      assert (HB := aconjER Hp);
	      ((linLogElim HA; try linLogElim HB) ||
		(linLogElim HB));
	      assert (Hp' := aconjIntro HA HB);
	      clear HA HB)
     end;
    first [(clear Hp; rename Hp' into Hp; try (linLogElim Hp)) |
      try (linLogElim Hp')])
with sconjSubSimpl Hp Hp' :=
  let h1 := fresh "h" in
  let h2 := fresh "h" in
  let J := fresh "J" in
  let Hp1 := fresh "H" in
  let Hp2 := fresh "H" in
    (destruct (sconjElim Hp) as [h1 [h2 [J [Hp1 Hp2]]]];
      ((linLogElim Hp1; try linLogElim Hp2) ||
	(linLogElim Hp2));
      assert (Hp' := sconjIntro Hp1 Hp2 J);
      clear h1 h2 J Hp1 Hp2).

Ltac lle := linLogElim.

Ltac llei Hp := linLogElim Hp; lli.


(** Relating |-> to heap contents *)

Lemma contGetEqL : forall h l w A,
  h ||- l |-> w ** A -> 
    get h l = Some w.
Proof.
  intros s l w A Q.
  destruct (sconjElim Q) as [ha [hb [J [HA HB]]]].
  unfold get.
  destruct (contElim HA) as [X _].
  rewrite (isPtrEq X).
  eapply mapJoinSoundLeft.
  eauto.
  assert (Y := contGetEq HA).
  unfold get in Y.
  rewrite (isPtrEq X) in Y.
  assumption.
Qed.


Lemma contContEq : forall h x y1 A,
    h ||- x |-> y1 ** A -> forall y2 B,
      h ||- x |-> y2 ** B ->
	y1 = y2.
Proof.
  intros h x y1 A H1 y2 B H2.
  assert (Some y1 = Some y2).
  rewrite <- (contGetEqL H1).
  rewrite <- (contGetEqL H2).
  reflexivity.
  injection H; auto.
Qed.


Lemma contSetEqL : forall h x y A,
  h ||- x |-> y ** A -> forall z,
    set h x z = Some (uset h x z).
Proof.
  intros.
  unfold set.
  rewrite (contGetEqL H).
  reflexivity.
Qed.


Lemma contUpdate : forall a v1 v2 h,
  h ||- a |-> v1 ->
    uset h a v2 ||- a |-> v2.
Proof.
  unfold cont.
  unfold singleton_heap; unfold uset.
  intros a v1 v2 h [P H].
  constructor.
  assumption.
  clear P.
  assert (E : heap_eqv (singleton_map (div4 a) v1) h).
  eapply map_eqvTrans.
  apply map_eqvSymm; eauto.
  apply map_eqvSymm; apply map_canonEqv.
  clear H.
  assert (R : heap_eqv (singleton_map (div4 a) v2) (map_add h (div4 a) v2)).
  assert (Q := map_eqvAdd (div4 a) v2 E); clear E.
  assert (Z := map_eqvSymm (singletonAddEqv (div4 a) v1 v2)).
  exact (map_eqvTrans Z Q).
  eapply map_eqvTrans.
  apply map_eqvSymm; apply map_canonEqv.
  apply map_eqvSymm; assumption.
Qed.


Lemma contUpdateSt : forall h a v1 A,
  h ||- (a |-> v1 ** A) -> forall v2,
    uset h a v2 ||- (a |-> v2 ** A).
Proof.
  intros h a v1 A [h1 [h2 [J [H1 H2]]]] v2.
  exists (uset h1 a v2).
  exists h2.
  intuition.
  eapply mapJoinEqv
    with (ma:=uset h1 a v2) (mb:=h2) (m:=uset h a v2).
  unfold uset; eapply map_addJoin.
  assert (R := contGetEq H1).
  unfold get in R.
  destruct (isPtr a).
  apply R.
  discriminate R.
  eapply mapJoinEqv.
  apply J.
  apply map_eqvSymm; apply map_canonEqv.
  apply map_eqvSymm; apply map_canonEqv.
  apply map_eqvSymm; apply map_canonEqv.
  apply map_canonEqv.
  apply map_canonEqv.
  apply map_canonEqv.
  eapply contUpdate; eauto.
Qed.


Lemma contUpdateNeqTop : forall h x v1,
  h ||- (x |-> v1 ** top) -> forall y v2,
    x <> y -> ptrP y ->
    Heap.uset h y v2 ||- (x |-> v1 ** top).
Proof.
  intros h x v1 Hp y v2 Ne Py.
  destruct (sconjElim Hp) as [h1 [h2 [J [H1 _]]]]; clear Hp.
  apply (sconjIntro (ha := h1) (hb := Heap.uset h2 y v2)).
  assumption.
  linLogIntro.
  assert (G1 : Heap.HeapMap.MapDef.get h1 (div4 y) = None).
  destruct H1 as [Px h1eq].
  rewrite map_canonEqv.
  rewrite h1eq.
  simpl.
  destruct (POnat.POnat.eqDec (div4 y) (div4 x)).
  elimtype False; apply Ne.  
  destruct (ptrP_elim Px) as [kx Hx].
  destruct (ptrP_elim Py) as [ky Hy].
  subst.
  rewrite div44eq in e.
  rewrite div44eq in e.
  rewrite e; reflexivity.
  reflexivity.

  constructor; intro k.
  compare k (div4 y); intro E.
  subst k.
  assert (Q := isPtrOk y).
  destruct (isPtr y).
  rewrite G1.
  simpl.
  destruct (POnat.POnat.eqDec (div4 y) (div4 y)).
  constructor.
  elimtype False; auto.
  elimtype False; auto.
  destruct J as [J]; hypReplace J (J k).
  replace (Heap.HeapMap.MapDef.get (Heap.uset h2 y v2) k)
    with (Heap.HeapMap.MapDef.get h2 k).
  replace (Heap.HeapMap.MapDef.get (Heap.uset h y v2) k)
    with (Heap.HeapMap.MapDef.get h k).
  assumption.
  simpl.
  destruct (POnat.POnat.eqDec k (div4 y)).
  elimtype False; auto.
  reflexivity.
  simpl.
  destruct (POnat.POnat.eqDec k (div4 y)).
  elimtype False; auto.
  reflexivity.
Qed.


Lemma contUnUpdateNeqTop : forall h x y v1 v2,
  Heap.uset h y v2 ||- (x |-> v1 ** top) ->
    x <> y -> ptrP y ->
    h ||- (x |-> v1 ** top).
Proof.
  intros h x y v1 v2 Hp Ne Py.
  destruct (sconjElim Hp) as [h1 [h2 [[J] [Hp1 _]]]]; clear Hp.
  eapply sconjIntro.
  apply Hp1.
  linLogIntro.
  instantiate (1 := 
    (fun m => match map_get h (div4 y) with
	       | Some v => map_add m (div4 y) v
	       | None => m
	      end)
    (map_diff h2 (singleton_map (div4 y) v2))).
  destruct (contElim Hp1) as [Px H1eq]; clear Hp1.
  constructor; intro k.
  unfold MapDef.get in * |- * .
  hypReplace J (J k).

  assert (Ne' : div4 x <> div4 y).
  destruct (ptrP_elim Px) as [z1 Ex].
  destruct (ptrP_elim Py) as [z2 Ey].
  subst x; subst y.
  rewrite div44eq; rewrite div44eq.
  intro; subst z1; apply Ne; reflexivity.

  compare k (div4 y); intro E.
  
  subst k.
  rewrite H1eq; simpl.
  destruct (POnat.POnat.eqDec (div4 y) (div4 x)) as [Eq | _].
  elimtype False; auto.
  destruct (map_get h (div4 y)).
  simpl; destruct (POnat.POnat.eqDec (div4 y) (div4 y)).
  constructor.
  elimtype False; apply n; reflexivity.

  assert (I : ASet.set_In (div4 y) (map_dom (singleton_map (div4 y) v2))).
  left; reflexivity.

  rewrite (map_get_diffInEq _ _ I).
  constructor.

  assert (NI : ~ASet.set_In k (map_dom (singleton_map (div4 y) v2))).
  simpl; intuition.

  simpl in J.
  assert (X := map_mem_In_dom (div4 y) h).
  destruct (map_get h (div4 y)).
  simpl.
  destruct (POnat.POnat.eqDec k (div4 y)).
  elimtype False; auto.

  rewrite <- (map_get_diffNotInEq _ _ NI).
  assumption.

  rewrite <- (map_get_diffNotInEq _ _ NI).
  destruct (POnat.POnat.eqDec k (div4 y)).
  elimtype False; auto.
  assumption.

Qed.


(** properties of forall_\ast *)


Add Morphism fas with signature set_eqv ==> eq ==> iff as fasMorph.
Proof.
  unfold fas.
  intros p s s' E h; constructor; intros H.
  rewrite <- (set_canonEqvEq E); auto.
  rewrite (set_canonEqvEq E); auto.
Qed.


Lemma fasEqvSet : forall S S' h p,
  h ||- fas p S -> 
    set_eqv S S' -> 
    h ||- fas p S'.
Proof.
  unfold fas.
  intros.
  rewrite <- (set_canonEqvEq H0).
  auto.
Qed.


Lemma fasImpl : forall P s h,
  h ||- (fas P s) -> forall Q,
  (forall x h, h ||- P x -> h ||- Q x) ->
  h ||- (fas Q s).
Proof.
  unfold fas.
  intros P s.
  induction (set_canon s).

  auto.

  simpl.
  intros.
  apply (sconjImpl H).
  auto.
  auto.

Qed.
Implicit Arguments fasImpl [P s h].


Lemma fasImpl0 : forall P Q,
  (forall x h, h ||- P x -> h ||- Q x) -> forall s h,
  h ||- (fas P s) -> 
  h ||- (fas Q s).
Proof.
  intros; eapply fasImpl; eauto.
Qed.


Lemma fasImplIS : forall P s h,
  h ||- (fas P s) -> forall Q,
    (forall x h', set_In x s -> subHeap h' h -> h' ||- P x -> h' ||- Q x) ->
    h ||- (fas Q s).
Proof.
  intros P s h Hp Q Imp.

  assert (PQ' : forall (x : POnat.POnat.A) (h' : Heap.heap),
    set_In x (set_canon s) -> subHeap h' h ->
    h' ||- P x -> 
      h' ||- Q x).
  intros x h' I S HP.
  apply Imp.
  setoid_rewrite (set_canonEqv s).
  assumption.
  assumption.
  assumption.
  
  clear Imp.
  generalize dependent h.
  unfold fas.
  induction (set_canon s).
  auto.

  simpl.
  intros h Hp PQ.
  destruct (sconjElim Hp) as [h1 [h2 [J [H1 H2]]]].
  eapply sconjIntro.
  3:apply J.
  apply PQ.
  left; reflexivity.
  eapply heap_joinSub; eauto.
  assumption.
  apply IHs0.
  assumption.

  intros; apply PQ; auto.
  eapply subHeapTrans; eauto.
  eapply heap_joinSub.
  apply Heap.HeapMap.ABJoin.mapJoinComm.
  eauto.

Qed.
Implicit Arguments fasImplIS [P s h].


Theorem fasIntro : forall a s p h,
  h ||- p a ** fas p s -> ~set_In a s ->
    h ||- fas p (set_add a s).
Proof.
unfold fas; simpl.
  intros a s p h H NI0.
  assert (NI : ~set_In a (set_canon s)).
  intro Z.
  apply NI0.
  apply set_canon_elim; assumption.
  clear NI0.
  generalize dependent h; generalize dependent s.
  intro s.
  induction (set_canon s).
  
  auto.
  
  simpl.
  intros Ni h H.
  
  destruct (POnat.POnat.compareDec a a0).
  destruct s1.
  assumption.
  elimtype False; intuition.
  simpl.

  assert (h ||- (p a0 ** p a ** fas' p s0)).
  apply sconjComm.
  apply sconjAssocR.
  eapply sconjImpl.
  apply H.
  auto.
  intros; apply sconjComm; assumption.
  clear H.
  eapply sconjImpl.
  apply H0.
  auto.
  intros.
  apply IHs0.
  auto.
  assumption.
Qed.


Theorem fasElimE : forall h P,
  h ||- fas P empty_set ->
    h ||- unit.
Proof.
  unfold fas.
  simpl.
  auto.
Qed.


Lemma fasElim' : forall a p s h,
  set_In a s -> h ||- fas' p s ->
  exists s1, exists s2,
    h ||- fas' p s1 ** p a ** fas' p s2 /\ s1 ++ (a::s2) = s.
Proof.
  induction s.
  
  simpl; intros; contradiction.

  simpl.
  intros h I ST.
  destruct I.
  subst.
  clear IHs.
  exists empty_set; exists s.
  simpl.
  constructor.
  2:reflexivity.
  apply unitIL.
  assumption.
  destruct (sconjElim ST) as [ha [hb [J [HA HB]]]]; clear ST.
  destruct (IHs _ H HB) as [s1 [s2 [FA' Seq']]]; clear H HB IHs.
  exists (a0::s1); exists s2.
  simpl.
  constructor.
  apply sconjAssocR.
  eapply sconjIntro; eauto.
  rewrite Seq'.
  reflexivity.
Qed.


Lemma fas'Combine : forall p s1 s2 h,
  h ||- fas' p s1 ** fas' p s2 ->
    h ||- fas' p (s1++s2).
Proof.
  induction s1.
  simpl.
  intros.
  apply unitEL.
  assumption.
  
  simpl.
  intros s2 h H.
  assert (Z := sconjAssocL H); clear H.
  eapply sconjImpl.
  apply Z.
  auto.
  intros s H.
  apply IHs1.
  assumption.
Qed.


Theorem fasElim : forall a s,
  set_In a s -> forall p h,
    h ||- fas p s ->
    h ||- p a ** fas p (set_remove a s).
Proof.
  unfold fas.
  intros a s I p h F.
  destruct (fasElim' _ _ _ _ (set_canon_intro _ _ I) F)
    as [s1 [s2 [F' Eq]]].
  assert (F'' : h ||- p a ** fas' p s1 ** fas' p s2).
  apply sconjAssocL.
  apply sconjComm.
  apply sconjAssocL.
  apply sconjComm.
  eapply sconjImpl.
  apply F'.
  auto.
  intros.
  apply sconjComm.
  assumption.
  clear F'.
  eapply sconjImpl.
  apply F''.
  auto.
  clear I F F'' h.
  replace (s1 ++ a :: s2) with (set_union s1 (set_add a s2)) in Eq.
  assert (C : set_canonP (set_union s1 (set_add a s2))).
  rewrite Eq.
  apply set_canonCanonP.
  assert (~ set_In a s1).
  assert (Z1 := canonP_set_union_In_elim2 a s1 (set_add a s2) C).
  apply Z1; left; reflexivity.
  assert (~ set_In a s2).
  apply canonP_set_add_elim1.
  eapply canonP_set_union_elim2; eauto.

  intros h Z.
  assert (Z' := fas'Combine _ _ _ Z); clear Z.
  assert (set_eqv (set_canon (set_remove a s)) (set_union s1 s2)).
  clear Z'.
  intro a'.
  destruct (NatUtil.natEqDec a' a).
  subst.
  constructor.
  intro Q.
  absurd (a <> a).
  auto.
  eapply set_remove_elim2.
  apply set_canon_elim.
  eauto.
  intro Q.
  elimtype False.
  destruct (set_union_elim Q); auto.

  replace (set_canon (set_remove a s))
    with (set_remove a (set_canon s)).
  rewrite <- Eq.
  constructor.
  intro.
  assert (Z := set_remove_elim1 _ _ _ H1); clear H1.
  apply set_union_intro.
  destruct (set_union_elim Z).
  left; assumption.
  right; destruct (set_add_elim H1).
  intuition.
  assumption.

  intro.
  apply set_remove_intro.
  apply set_union_intro.
  destruct (set_union_elim H1).
  left; assumption.
  right; apply set_add_intro2.
  assumption.
  assumption.

  symmetry; apply set_canonRemoveSwap.
  replace (set_canon (set_remove a s))
    with (set_canon (set_canon (set_remove a s))).
  assert (Z := set_canonEqv (set_union s1 s2)).
  assert (Q := set_eqvTrans H1 Z); clear H1 Z.
  rewrite (set_canonEqvEq Q); clear Q.
  replace (s1 ++ s2) with (set_canon (set_canon (set_union s1 s2))) in Z'.
  assumption.
  rewrite set_canonCanonPEq.
  rewrite set_canonCanonPEq.
  reflexivity.
  clear Eq H H0 h Z' p s.
  generalize dependent C.
  induction s1.
  simpl.
  apply canonP_set_add_elim.
  simpl.
  inversion_clear 1.
  constructor.
  apply IHs1.
  assumption.
  intros.
  apply H0.
  destruct (set_union_elim H1).
  apply set_union_intro1; assumption.
  apply set_union_intro2; apply set_add_intro2; assumption.
  apply set_canonCanonP.
  apply set_canonCanonPEq.
  apply set_canonCanonP.
  reflexivity.
Qed.

Implicit Arguments fasElim [a s p h].


(** Tactics for manipulating sconjs *)

(** should probably move these up a bit *)

Ltac sconjLift2of2 H :=
  hypReplace H (sconjComm H).



Lemma sconjLift3of3Ok : forall h A B C,
  h ||- A ** B ** C ->
    h ||- C ** A ** B.
Proof.
  intros h A B C H.
  apply sconjComm; apply sconjAssocR; assumption.
Qed.


Ltac sconjLift3of3 H :=
  hypReplace H (sconjLift3of3Ok H).


Lemma sconjLift3Ok : forall h A B C D,
  h ||- A ** B ** C ** D ->
    h ||- C ** A ** B ** D.
Proof.
  intros h A B C D H.
  apply sconjComm; apply sconjAssocR.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjComm; assumption.
Qed.


Ltac sconjLift3 H := 
  hypReplace H (sconjLift3Ok H).


Lemma sconjLift4of4Ok : forall h A B C D,
  h ||- A ** B ** C ** D ->
  h ||- D ** A ** B ** C.
Proof.
  intros h A B C D H.
  apply sconjComm; apply sconjAssocR.
  apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; assumption.
Qed.


Ltac sconjLift4of4 H :=
  hypReplace H (sconjLift4of4Ok H).


Lemma sconjLift4Ok : forall h A B C D E,
  h ||- A ** B ** C ** D ** E ->
  h ||- D ** A ** B ** C ** E.
Proof.
  intros h A B C D E H.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjComm.
  assumption.
Qed.


Ltac sconjLift4 H :=
  hypReplace H (sconjLift4Ok H).



Lemma sconjLift5of5Ok : forall h A B C D E,
  h ||- A ** B ** C ** D ** E ->
  h ||- E ** A ** B ** C ** D.
Proof.
  intros h A B C D E H.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  assumption.
Qed.


Ltac sconjLift5of5 H :=
  hypReplace H (sconjLift5of5Ok H).


Lemma sconjLift5Ok : forall h A B C D E F,
  h ||- A ** B ** C ** D ** E ** F ->
  h ||- E ** A ** B ** C ** D ** F.
Proof.
  intros h A B C D E F H.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjComm.
  assumption.
Qed.


Ltac sconjLift5 H :=
  hypReplace H (sconjLift5Ok H).


Lemma sconjLift6of6Ok : forall h A B C D E F,
  h ||- A ** B ** C ** D ** E ** F ->
  h ||- F ** A ** B ** C ** D ** E.
Proof.
  intros h A B C D E F H.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  assumption.
Qed.

Ltac sconjLift6of6 H :=
  hypReplace H (sconjLift6of6Ok H).


Lemma sconjLift6Ok : forall h A B C D E F G,
  h ||- A ** B ** C ** D ** E ** F ** G ->
  h ||- F ** A ** B ** C ** D ** E ** G .
Proof.
  intros h A B C D E F G Hp.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjComm.
  assumption.
Qed.


Ltac sconjLift6 H :=
  hypReplace H (sconjLift6Ok H).


Lemma sconjLift7of7Ok : forall h A B C D E F G,
  h ||- A ** B ** C ** D ** E ** F ** G ->
  h ||- G ** A ** B ** C ** D ** E ** F.
Proof.
  intros h A B C D E F G H.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  apply sconjAssocR; apply (sconjImplR H); clear h H; intros h H.
  assumption.
Qed.

Ltac sconjLift7of7 H :=
  hypReplace H (sconjLift7of7Ok H).


Lemma sconjLift7Ok : forall h A B C D E F G H,
  h ||- A ** B ** C ** D ** E ** F ** G ** H ->
  h ||- G ** A ** B ** C ** D ** E ** F ** H .
Proof.
  intros h A B C D E F G H Hp.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjComm.
  assumption.
Qed.


Ltac sconjLift7 H :=
  hypReplace H (sconjLift7Ok H).


Lemma sconjLift8Ok : forall h A B C D E F G H I,
  h ||- A ** B ** C ** D ** E ** F ** G ** H ** I ->
  h ||- H ** A ** B ** C ** D ** E ** F ** G ** I .
Proof.
  intros h A B C D E F G H I Hp.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjComm.
  assumption.
Qed.


Ltac sconjLift8 H :=
  hypReplace H (sconjLift8Ok H).


Lemma sconjLift8of8Ok : forall h A B C D E F G H,
  h ||- A ** B ** C ** D ** E ** F ** G ** H ->
  h ||- H ** A ** B ** C ** D ** E ** F ** G.
Proof.
  intros h A B C D E F G H Hp.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  assumption.
Qed.

Ltac sconjLift8of8 H :=
  hypReplace H (sconjLift8of8Ok H).


Lemma sconjLift9of9Ok : forall h A B C D E F G H I,
  h ||- A ** B ** C ** D ** E ** F ** G ** H ** I ->
  h ||- I ** A ** B ** C ** D ** E ** F ** G ** H.
Proof.
  intros h A B C D E F G H I Hp.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  assumption.
Qed.

Ltac sconjLift9of9 H :=
  hypReplace H (sconjLift9of9Ok H).


Lemma sconjLift9Ok : forall h A B C D E F G H I J,
  h ||- A ** B ** C ** D ** E ** F ** G ** H ** I ** J ->
  h ||- I ** A ** B ** C ** D ** E ** F ** G ** H ** J .
Proof.
  intros h A B C D E F G H I J Hp.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjComm.
  assumption.
Qed.


Ltac sconjLift9 H :=
  hypReplace H (sconjLift9Ok H).


Lemma sconjLift10Ok : forall h A B C D E F G H I J K,
  h ||- A ** B ** C ** D ** E ** F ** G ** H ** I ** J ** K ->
  h ||- J ** A ** B ** C ** D ** E ** F ** G ** H ** I ** K.
Proof.
  intros h A B C D E F G H I J K Hp.
  apply sconjComm.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjAssocR; apply (sconjImplR Hp); clear h Hp; intros h Hp.
  apply sconjComm.
  assumption.
Qed.


Ltac sconjLift10 H :=
  hypReplace H (sconjLift10Ok H).




Ltac sconjGLift2 := apply sconjComm; apply sconjAssocR.

Ltac sconjGLift3 :=
  apply sconjComm; apply sconjAssocR;
  apply sconjComm; apply sconjAssocR.


Lemma sconjGLift3' : forall h A B C D,
  h ||- C ** A ** B ** D ->
    h ||- A ** B ** C ** D.
Proof.
  intros h A B C D Hp.
  sconjLift3 Hp; sconjLift3 Hp.
  assumption.
Qed.


Lemma sconjGLift2' : forall h A B C,
  h ||- B ** A ** C ->
    h ||- A ** B ** C.
Proof.
  intros h A B C Hp.
  sconjLift2 Hp.
  assumption.
Qed.


(** More complex fas lemmas *)

Lemma fasFasDisjF : forall F h P1 P2 s1 s2 A,
  h ||- fas P1 s1 ** fas P2 s2 ** A ->
    (forall h x, h ||- P1 x -> h ||- F x |->_ ** top) ->
    (forall h x, h ||- P2 x -> h ||- F x |->_ ** top) ->
    set_disj s1 s2.
Proof.
  intros F h P1 P2 s1 s2 A H Imp1 Imp2.
  apply set_disj_intro.
  intros x I1 I2.
  assert (h ||- ((F x |->_ ** top)** top) ** ((F x |->_ ** top) ** top) ** top).
  apply (sconjImpl H); clear h H; intros h H.
  hypReplace H (fasElim I1 H).
  apply (sconjImpl H); clear h H; intros h H.
  eauto.
  linLogIntro.
  apply (sconjImpl H); clear h H; intros h H.
  hypReplace H (fasElim I2 H).
  apply (sconjImpl H); clear h H; intros h H.
  eauto.
  linLogIntro.
  linLogIntro.
  linLogElim H0.
Qed.


Lemma fasFasDisj : forall h P1 P2 s1 s2 A,
  h ||- fas P1 s1 ** fas P2 s2 ** A ->
    (forall h x, h ||- P1 x -> h ||- x |->_ ** top) ->
    (forall h x, h ||- P2 x -> h ||- x |->_ ** top) ->
    set_disj s1 s2.
Proof.
  exact (fasFasDisjF (fun x => x)).
Qed.
Implicit Arguments fasFasDisj [h P1 P2 s1 s2 A].


Lemma fasFasDisj' : forall h P1 P2 s1 s2,
  h ||- fas P1 s1 ** fas P2 s2 ->
    (forall h x, h ||- P1 x -> h ||- x |->_ ** top) ->
    (forall h x, h ||- P2 x -> h ||- x |->_ ** top) ->
    set_disj s1 s2.
Proof.
  intros h P1 P2 s1 s2 Hp Imp1 Imp2.
  eapply fasFasDisj.
  instantiate (4 := h); instantiate (1 := unit).
  lli; eauto.
  assumption.
  assumption.
Qed.
Implicit Arguments fasFasDisj' [h P1 P2 s1 s2].


Lemma fasSplit : forall h A s s',
  h ||- fas A (set_union s s') ->
    set_disj s s' ->
    h ||- (fas A s) ** (fas A s').
Proof.
  intros h A s s' H0 D0.
  assert (Eq := set_canonEqv s).
  assert (D : set_disj (set_canon s) s').
  setoid_rewrite <- Eq; assumption.
  clear D0.
  assert (H : h ||- fas A (set_union (set_canon s) s')).
  rewrite hpredTrueEq;  setoid_rewrite <- Eq; assumption.
  clear H0.

  assert (Z : h ||- fas A (set_canon s) ** fas A s').
  clear Eq.
  assert (C := set_canonCanonP s).
  generalize dependent h.
  generalize dependent (set_canon s).
  clear s.
  induction 2.

  simpl; intros h H.
  hypReplace H (unitIL H).
  apply (sconjImplL H); clear h H; intros h H.
  exact (fasIntroE _ H).

  rename H into LT; intros h H.
  assert (I : set_In a (set_union (set_add a s) s')).
  apply set_union_intro1; apply set_add_intro1; reflexivity.
  hypReplace H (fasElim I H); clear I.
  assert (H' : h ||- A a ** fas A s ** fas A s').
  apply (sconjImplR H); clear h H; intros h H.
  apply IHC.
  apply set_disj_intro; intros x I I'.
  assert (I2 : set_In x (set_add a s)).
  apply set_add_intro2; assumption.
  exact (set_disj_elim D _ I2 I').
  rewrite hpredTrueEq;
    setoid_replace (set_union s s') with (set_remove a (set_union (set_add a s) s')).
  auto.
  clear H IHC C.
  setoid_rewrite (set_removeUnionDistrib a (set_add a s) s').
  assert (NI : ~set_In a s').
  intro NI.
  eapply (set_disj_elim D).
  apply set_add_intro1; reflexivity.
  assumption.
  rewrite (set_removeEq NI).
  clear NI.
  replace (set_remove a (set_add a s)) with (set_remove a s).
  destruct (set_In_dec a s).
  elimtype False; assert (Q := LT _ s0).
  unfold POnat.POnat.LT in Q; omega.
  clear LT.
  rewrite (set_removeEq n).
  apply set_eqvRefl.
  simpl.
  destruct (POnat.POnat.eqDec a a).
  reflexivity.
  elimtype False; apply n; reflexivity.
  hypReplace H' (sconjAssocR H').
  apply (sconjImplL H'); clear h H H'; intros h H.
  apply fasIntro.
  assumption.
  intro Z; assert (AA := LT _ Z).
  unfold POnat.POnat.LT in AA; omega.
  apply (sconjImplL Z).
  intros.
  unfold hpredTrue; setoid_rewrite Eq; assumption.
Qed.


Lemma fasImplI : forall h P s,
  h ||- (fas P s) ->
    forall Q,
      (forall x h, set_In x s -> h ||- P x -> h ||- Q x) ->
      h ||- (fas Q s).
Proof.
  intros h P s H Q PQ.

  assert (PQ' : forall (x : POnat.POnat.A) (h : Heap.heap),
    set_In x (set_canon s) -> h ||- P x -> h ||- Q x).
  intros x h' I HP.
  apply PQ.
  setoid_rewrite (set_canonEqv s).
  assumption.
  assumption.
  
  clear PQ.
  generalize dependent h.
  generalize dependent PQ'.
  unfold fas.
  induction (set_canon s).
  auto.

  simpl.
  intros PQ h H.
  apply (sconjImpl H); clear h H; intros h H.
  apply PQ; auto.
  apply IHs0; auto.
Qed.
Implicit Arguments fasImplI [h P s].


Lemma fasMerge : forall P h s s',
  h ||- fas P s ** fas P s' -> set_disj s s' ->
  h ||- fas P (set_union s s').
Proof.
  intros P h s s' H0 D0.
  unfold hpredTrue; setoid_rewrite (set_canonEqv s).
  assert (C := set_canonCanonP s).

  assert (H : h ||- fas P (set_canon s) ** fas P s').
  apply (sconjImplL H0); intros.
  unfold hpredTrue; setoid_rewrite <- (set_canonEqv s).
  assumption.
  clear H0.

  assert (D : set_disj (set_canon s) s').
  setoid_rewrite <- (set_canonEqv s).
  assumption.
  clear D0.

  generalize dependent s'.
  generalize dependent h.
  generalize dependent (set_canon s); clear s.
  induction 1; simpl; intros h s' H' D.

  apply unitEL; simpl.
  apply (sconjImpl H'); clear h H'; intros h H.
  assumption.
  assumption.

  assert (NI : ~set_In a s).
  intro X.
  assert (Q := H _ X).
  unfold POnat.POnat.LT in Q; omega.

  apply fasIntro.
  assert (I : set_In a (set_add a s)).
  left; reflexivity.

  assert (h ||- P a ** fas P s ** fas P s').
  apply sconjAssocL.
  apply (sconjImplL H'); clear H'; intros h' H'.
  hypReplace H' (fasElim I H').
  apply (sconjImplR H'); clear h' H'; intros h' H'.  

  simpl in H'.
  destruct (POnat.POnat.eqDec a a).
  clear e.
  rewrite <- (set_removeEq NI).
  assumption.
  elimtype False; apply n; reflexivity.

  clear H' I.
  apply (sconjImplR H0).
  intros.
  unfold hpredTrue; apply IHC.
  assumption.

  apply set_disj_intro.
  intros x I I'.
  eapply (set_disj_elim D).
  2:apply I'.
  apply set_add_intro2; assumption.

  intro I.
  destruct (set_union_elim I) as [I' | I']; clear I.
  auto.
  eapply (set_disj_elim D).
  2:apply I'.
  apply set_add_intro1; reflexivity.

Qed.


Lemma subHeapImpl : forall h' A,
  h' ||- A ->
    forall h, Heap.subHeap h' h ->
      h ||- A ** top.
Proof.
  intros h' A Hp h S.
  destruct (Heap.subHeapSplit S) as [h'' J].
  eapply sconjIntro.
  apply Hp.
  linLogIntro.
  apply J.
Qed.
Implicit Arguments subHeapImpl [h' A h].


(* map_mapRange related lemmas *)

Lemma contMapRangeImpl : forall h x y,
  h ||- x |-> y -> forall f : nat -> nat,
  map_mapRange f h ||- x |-> f y.
Proof.
  intros h x y Hp f.
  destruct (contElim Hp) as [ptrX Heq]; clear Hp.
  eapply heapEqvTrue.
  instantiate (1 := singleton_heap x (f y)).
  unfold heap_eqv in Heq |- * ;
    unfold singleton_heap in Heq |- * .
  eapply heap_eqvTrans.
  2:unfold singleton_heap; apply map_mapRangeSingleton.
  apply map_mapRangeEqv; assumption.
  eapply contIntro; assumption.
Qed.


Lemma sconjMapRangeImplS : forall h A X,
  h ||- A ** X -> forall f A' B',
  (forall h', Heap.subHeap h' h -> h' ||- A -> map_mapRange f h' ||- A') ->
  (forall h', Heap.subHeap h' h -> h' ||- X -> map_mapRange f h' ||- B') ->
  map_mapRange f h ||- A' ** B'.
Proof.
  intros h A B Hp f A' B' Aimp Bimp.
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]];
    eapply sconjIntro.
  apply Aimp.
  eapply Heap.heap_joinSub; eauto.
  assumption.
  apply Bimp.
  eapply heap_joinSub.
  apply mapJoinComm; eauto.
  assumption.
  apply map_mapRangeJoin; assumption.
Qed.
Implicit Arguments sconjMapRangeImplS [h A X].


Lemma sconjMapRangeImpl : forall h A X,
  h ||- A ** X -> forall f A' B',
  (forall h, h ||- A -> map_mapRange f h ||- A') ->
  (forall h, h ||- X -> map_mapRange f h ||- B') ->
  map_mapRange f h ||- A' ** B'.
Proof.
  intros; eapply sconjMapRangeImplS; eauto.
Qed.
Implicit Arguments sconjMapRangeImpl [h A X].


Lemma unitMapRangeImpl : forall h,
  h ||- unit -> forall f,
  map_mapRange f h ||- unit.
Proof.
  intros h Hp f.
  assert (E := unitElim Hp); clear Hp.
  eapply heapEqvTrue.
  2:apply unitIntro.
  apply heap_eqvTrans with (map_mapRange f empty_heap).
  apply map_mapRangeEqv; assumption.
  simpl; apply heap_eqvRefl.
Qed.


Lemma fasMapRangeImplIS : forall P S h,
  h ||- fas P S -> forall (f : nat -> nat) (Q : nat -> hpred),
  (forall h' x, Heap.subHeap h' h -> set_In x S -> h' ||- P x -> 
    map_mapRange f h' ||- Q x) ->
  map_mapRange f h ||- fas Q S.
Proof.
  unfold fas.
  intros P s.
  assert (C : set_canonP (set_canon s)).
  apply set_canonCanonP.
  intros h Hp f Q Pimp.
  assert (Pimp' : forall h' x, subHeap h' h -> set_In x (set_canon s) ->
    h' ||- P x -> map_mapRange f h' ||- Q x).
  intros; apply Pimp; auto; rewrite (set_canonEqv s); assumption.
  clear Pimp; rename Pimp' into Pimp.
  rename s into s0; set (s := set_canon s0) in * ; 
    generalize dependent h; generalize dependent s; clear s0.
  induction 1; simpl; intros.

  apply unitMapRangeImpl; assumption.

  rename H into Lt.
  apply (sconjMapRangeImplS Hp); auto.
  intros; apply IHC; auto.
  intros; apply Pimp.
  eapply subHeapTrans; eauto.
  right; assumption.
  assumption.
Qed.
Implicit Arguments fasMapRangeImplIS [h P S].


Lemma fasMapRangeImplI : forall h P S,
  h ||- fas P S -> forall f Q,
  (forall h x, set_In x S -> h ||- P x -> map_mapRange f h ||- Q x) ->
  map_mapRange f h ||- fas Q S.
Proof.
  intros; eapply fasMapRangeImplIS; eauto.
Qed.
Implicit Arguments fasMapRangeImplI [h P S].


Lemma fasMapRangeImpl : forall h P S,
  h ||- fas P S -> forall f Q,
  (forall h x, h ||- P x -> map_mapRange f h ||- Q x) ->
  map_mapRange f h ||- fas Q S.
Proof.
  intros; eapply fasMapRangeImplIS; eauto.
Qed.
Implicit Arguments fasMapRangeImpl [h P S].


(* map_mapDom related lemmas *)

Definition heapToMapMap f x := div4 (f (4 * x)).

Definition heap_mapDom f : heap -> heap := map_mapDom (heapToMapMap f).

Definition heapMapOk f (h:heap) :=
  (forall x, ptrP x -> ptrP (f x)) /\ 
  (forall x y, set_In x (heap_dom h) -> set_In y (heap_dom h) ->
    x <> y -> f x <> f y).


Lemma ptrFMultNe : forall f h,
  heapMapOk f h ->
  domMapOk (heapToMapMap f) h.
Proof.
  intros f h [P F] x y Ix Iy NE;
    unfold POnat.POnat.A in *; unfold ASet.set_In in Ix, Iy.
  assert (Ne' : 4 * x <> 4 * y).
  omega.
  assert (Ix' : set_In (4 * x) (heap_dom h)).
  unfold heap_dom; apply set_map_intro with x; auto.
  assert (Iy' : set_In (4 * y) (heap_dom h)).
  unfold heap_dom; apply set_map_intro with y; auto.
  assert (Ne3 := F _ _ Ix' Iy' Ne').
  unfold heapToMapMap.
  destruct (ptrP_elim (P _ (ptrP_intro' x))) as [k Ex]; rewrite Ex.
  destruct (ptrP_elim (P _ (ptrP_intro' y))) as [k' Ey]; rewrite Ey.
  rewrite div44eq; rewrite div44eq.
  intro; subst k'; rewrite <- Ey in Ex; clear Ey; apply (F (4 * x) (4 * y)); assumption.
Qed.


Lemma contMapDomImpl : forall h x y,
  h ||- x |-> y -> forall f,
    heapMapOk f h ->
    heap_mapDom f h ||- f x |-> y.
Proof.
  intros h x y Hp f [fPtr Pne].
  destruct (contElim Hp) as [xPtr Heq]; clear Hp.
  assert (fxPtr := fPtr _ xPtr).
  eapply heapEqvTrue.
  instantiate (1 := singleton_heap (f x) y).
  2:apply contIntro; assumption.
  unfold heap_eqv in Heq |- * ; unfold singleton_heap in Heq |- * .
  eapply heap_eqvTrans.
  instantiate (1 := heap_mapDom f (singleton_map (div4 x) y)).
  unfold heap_mapDom; apply map_mapDomEqv.
  2:assumption.
  apply ptrFMultNe; constructor; assumption.
  unfold heap_mapDom.
  replace (map_mapDom (fun x0 : POnat.POnat.A => div4 (f (4 * x0)))
    (singleton_map (div4 x) y)) with (singleton_map (div4 (f (4 * (div4 x)))) y);
    [idtac | reflexivity].
  destruct (ptrP_elim xPtr) as [k' E']; rewrite E'; rewrite div44eq.
  apply heap_eqvRefl.

Qed.


Lemma sconjMapDomImplS0 : forall h A X,
  h ||- A ** X -> forall f A' B',
  (forall h', Heap.subHeap h' h -> h' ||- A -> heap_mapDom f h' ||- A') ->
  (forall h', Heap.subHeap h' h -> h' ||- X -> heap_mapDom f h' ||- B') ->
  heapMapOk f h ->
  heap_mapDom f h ||- A' ** B'.
Proof.
  intros h A B Hp f A' B' Aimp Bimp fOk.
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]];
    eapply sconjIntro.
  apply Aimp.
  eapply Heap.heap_joinSub; eauto.
  assumption.
  apply Bimp.
  eapply heap_joinSub.
  apply mapJoinComm; eauto.
  assumption.
  unfold heap_mapDom; apply map_mapDomJoin.
  assumption.
  intros x y Ne.
  apply ptrFMultNe; assumption.
Qed.
Implicit Arguments sconjMapDomImplS0 [h X].


Lemma subHeapDomSubset : forall h h',
  subHeap h h' ->
  set_subset (heap_dom h) (heap_dom h').
Proof.
  intros h h' Sub x Ix.
  unfold subHeap in Sub.
  hyr Sub (Sub x).
  destruct (set_map_elim Ix) as [y [xeq Iy]]; clear Ix.
  unfold heap_dom; apply set_map_intro with y.
  assumption.
  assert (X := map_mem_In_dom y h).
  assert (X' := map_mem_In_dom y h').
  unfold get in Sub.
  rewrite isPtrEq in Sub.
  2:subst x; apply ptrP_intro'.
  replace (div4 x) with y in Sub.
  2:subst x; rewrite div44eq; reflexivity.
  destruct (map_get h y).
  destruct (map_get h' y).
  assumption.
  assert (D : None = Some b).
  auto.
  discriminate D.
  elf; auto.
Qed.


Lemma heapMapOkSub : forall f h,
  heapMapOk f h -> forall h',
    subHeap h' h ->
    heapMapOk f h'.
Proof.
  intros f h [P F] h' Sub; constructor.
  assumption.
  intros x y Ix Iy NE; apply (F x y).
  apply (subHeapDomSubset Sub); assumption.
  apply (subHeapDomSubset Sub); assumption.
  assumption.
Qed.
  

Lemma sconjMapDomImplS : forall h A X,
  h ||- A ** X -> forall f A' B',
  (forall h', Heap.subHeap h' h -> heapMapOk f h' -> h' ||- A -> heap_mapDom f h' ||- A') ->
  (forall h', Heap.subHeap h' h -> heapMapOk f h' -> h' ||- X -> heap_mapDom f h' ||- B') ->
  heapMapOk f h ->
  heap_mapDom f h ||- A' ** B'.
Proof.
  intros h A X Hp f A' B' MA MB Hok;
    eapply sconjMapDomImplS0.
  apply Hp.
  intros; apply MA; auto; apply heapMapOkSub with h; assumption.
  intros; apply MB; auto; apply heapMapOkSub with h; assumption.
  assumption.
Qed.
Implicit Arguments sconjMapDomImplS [h X].

Lemma sconjMapDomImpl : forall h A X,
  h ||- A ** X -> forall f A' B',
  (forall h, h ||- A -> heapMapOk f h -> heap_mapDom f h ||- A') ->
  (forall h, h ||- X -> heapMapOk f h -> heap_mapDom f h ||- B') ->
  heapMapOk f h ->
  heap_mapDom f h ||- A' ** B'.
Proof.
  intros; eapply sconjMapDomImplS; eauto.
Qed.
Implicit Arguments sconjMapDomImpl [h X].


Lemma unitMapDomImpl : forall h,
  h ||- unit -> forall f,
    heapMapOk f h ->
    heap_mapDom f h ||- unit.
Proof.
  intros h Hp f Fok.
  assert (E := unitElim Hp); clear Hp.
  eapply heapEqvTrue.
  2:apply unitIntro.
  apply heap_eqvTrans with (heap_mapDom f empty_heap).
  unfold heap_mapDom.
  apply map_mapDomEqv.
  apply ptrFMultNe; assumption.
  assumption.
  simpl; apply heap_eqvRefl.
Qed.


Lemma fasMapDomImplIS0 : forall P S h,
  h ||- fas P S -> forall (f : nat -> nat) (Q : nat -> hpred),
  (forall h' x, Heap.subHeap h' h -> set_In x S -> h' ||- P x -> 
    heap_mapDom f h' ||- Q x) ->
  heapMapOk f h ->
  heap_mapDom f h ||- fas Q S.
Proof.
  unfold fas.
  intros P s.
  assert (C : set_canonP (set_canon s)).
  apply set_canonCanonP.
  intros h Hp f Q Pimp fOk.
  assert (Pimp' : forall h' x, subHeap h' h -> set_In x (set_canon s) ->
    h' ||- P x -> heap_mapDom f h' ||- Q x).
  intros; apply Pimp; auto; rewrite (set_canonEqv s); assumption.
  clear Pimp; rename Pimp' into Pimp.
  rename s into s0; set (s := set_canon s0) in * ; 
    generalize dependent h; generalize dependent s; clear s0.
  induction 1; simpl; intros.

  apply unitMapDomImpl; assumption.

  rename H into Lt.
  apply (sconjMapDomImplS _ Hp); auto.
  intros; apply IHC; auto.
  intros; apply Pimp.
  eapply subHeapTrans; eauto.
  right; assumption.
  assumption.
Qed.
Implicit Arguments fasMapDomImplIS0 [h P S].


Lemma fasMapDomImplIS : forall P S h,
  h ||- fas P S -> forall (f : nat -> nat) (Q : nat -> hpred),
  (forall h' x, Heap.subHeap h' h -> heapMapOk f h' -> set_In x S -> h' ||- P x -> 
    heap_mapDom f h' ||- Q x) ->
  heapMapOk f h ->
  heap_mapDom f h ||- fas Q S.
Proof.
  intros; eapply fasMapDomImplIS0; eauto.
  intros h' x Sub Ix Hp; apply H0; auto.
  apply heapMapOkSub with h; assumption.
Qed.
Implicit Arguments fasMapDomImplIS [h P S].


Lemma fasMapDomImplI : forall h P S,
  h ||- fas P S -> forall f Q,
  (forall h x, set_In x S -> heapMapOk f h -> h ||- P x -> heap_mapDom f h ||- Q x) ->
  heapMapOk f h ->
  heap_mapDom f h ||- fas Q S.
Proof.
  intros; eapply fasMapDomImplIS; eauto.
Qed.
Implicit Arguments fasMapDomImplI [h P S].


Lemma fasMapDomImpl : forall h P S,
  h ||- fas P S -> forall f Q,
  (forall h x, heapMapOk f h -> h ||- P x -> heap_mapDom f h ||- Q x) ->
  heapMapOk f h ->
  heap_mapDom f h ||- fas Q S.
Proof.
  intros; eapply fasMapDomImplIS; eauto.
Qed.
Implicit Arguments fasMapDomImpl [h P S].


(** Simple tactic for matching a separated conjunction hypothesis to
the goal.  The leftmost in the goal is set up with an implication with
the leftmost in the goal, and so on.  Use the sconjLift tactics above
to get things in the right order, then apply this.  Some attempt is
made to solve trivial cases of the subimplications.

*)

Ltac sconjMatch :=
  eauto;
  match goal with
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?B =>
      exact Hp
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?C =>
      apply (sconjImplR Hp); try (clear Hp); try (clear h);
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; sconjMatch)
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?B =>
      apply (sconjImplL Hp); try (clear Hp); try (clear h); 
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; sconjMatch)
    | Hp : ?h ||- sconj _ _ |- ?h ||- sconj _ _ =>
      apply (sconjImpl Hp); try (clear Hp); try (clear h); 
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; sconjMatch)
    | |- _ ||- top => apply topIntro
    | |- _ => idtac
  end.

Ltac scm := sconjMatch.

Ltac lleim := 
  match goal with
    | Hp : ?h ||- _ |- ?h ||- _ => llei Hp; sconjMatch
  end.

Ltac illeim := intros; lleim.


(** tactic for instantiating lexists *)
Ltac existsl y := apply lexI with (x := y).



(**
 A tactic dealing with the heap.
                                                                                
 It removes all the unchanged parts, the changed parts are left for
 manual proof.  It is guaranteed to terminate.
                                                                                
 ig: for the number of shifts left to be used on goal.
 ih: for the number of shifts left to be used on hyp.
 size: for the number of separation conjuction parts.
                                                                                
 Should be used like "matchHeap 4 4 4" if the goal is in the form of
 "h ||- A ** B ** C ** D"
                                                                                
 First, it shifts the heap in goal to match the heap in hyp, with "size" tries.
 Then, it shift the hyp once if it can, and retries the tactic.

 Underlying idea from Ch. 10 of Coq manual.

*)
Ltac matchHeap ig ih size :=
  match goal with
    | [|- ?h ||- top] => apply topIntro (* true *)
    | [Hp : ?h ||- ?A |- ?h ||- ?A ] => exact Hp (* assumption *)
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?C] => (* match left *)
      let h1 := fresh "h" in
	(apply (sconjImplR Hp);
	 (clear h Hp; intros h Hp) || (clear Hp; intros h1 Hp);
	 repeat (hypReplace Hp (sconjAssocL Hp)); repeat (apply sconjAssocR);
         matchHeap (pred size) (pred size) (pred size))
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?B] => (* match right *)
      let h' := fresh "h" in
	(apply (sconjImplL Hp);
	 (clear h Hp; intros h Hp) || (clear Hp; intros h' Hp);
	 repeat (hypReplace Hp (sconjAssocL Hp)); repeat (apply sconjAssocR);
         matchHeap (pred size) (pred size) (pred size))
    | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?D] => (* shift *)
      match eval compute in ig with
	| O => (* no more shift on goal *)
	  match eval compute in ih with
	    | O => idtac (* finished *)
	    | _ => (* shift hyp, continue with matching *)
              hypReplace Hp (sconjComm Hp); repeat (hypReplace Hp (sconjAssocL Hp));
	      matchHeap size (pred ih) size
	  end
	| _ => (* shift goal *)
	  apply sconjComm; repeat (apply sconjAssocR); matchHeap (pred ig) ih size
      end
    | _ => idtac
  end.
                                                                                
(** an similar tactic dealing with "rget s rgcInfo |->, ..." *)
                                                                                
Ltac smartAssocL H :=
 match goal with
   | [ H : _ ||- contL _ _ ** _ |- _ ] => idtac
   | [ H : _ ||- (_ ** _) ** _ |- _ ] => hypReplace H (sconjAssocL H);
     smartAssocL H
   | _ => idtac
 end.

Ltac smartAssocR :=
 match goal with
   | [ |- _ ||- contL _ _ ** _ ] => idtac
   | [ |- _ ||- (_ ** _) ** _ ] => apply sconjAssocR; smartAssocR
   | _ => idtac
 end.

Ltac matchHeapContL ig ih size :=
 match goal with
   | [|- ?h ||- top] => apply topIntro (* true *)
   | [Hp : ?h ||- ?A |- ?h ||- ?A ] => exact Hp (* assumption *)
   | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?C] => (* match left *)
     let h1 := fresh "h" in
       (apply (sconjImplR Hp);
	(clear h Hp; intros h Hp) || (clear Hp; intros h1 Hp);
	smartAssocL Hp; smartAssocR;
	matchHeapContL (pred size) (pred size) (pred size))
   | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?B] => (* match right *)
     let h' := fresh "h" in
       (apply (sconjImplL Hp);
	(clear h Hp; intros h Hp) || (clear Hp; intros h' Hp);
	smartAssocL Hp; smartAssocR;
	matchHeapContL (pred size) (pred size) (pred size))
   | [Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?D] => (* shift *)
     match eval compute in ig with
   | O => (* no more shift on goal *)
     match eval compute in ih with
       | O => idtac (* finished *)
       | _ => (* shift hyp, continue with matching *)
         hypReplace Hp (sconjComm Hp); smartAssocL Hp;
         matchHeapContL size (pred ih) size
     end
   | _ => (* shift goal *)
     apply sconjComm; smartAssocR; matchHeapContL (pred ig) ih size
     end
   | _ => idtac
 end.


Ltac sconjStep' size :=
 match goal with
 | [Hp : ?h ||- _ |- ?h ||- _ ] =>
   smartAssocL Hp; smartAssocR; matchHeapContL size size size
 end.



(** extensions of matchHeap to deal with various issues *)
Ltac matchHeap2 Hp ig ih hshifts :=
  try apply Hp;
  try apply topIntro;
  match type of Hp with
    | ?h ||- sconj ?A ?B =>
      match goal with 
	| |- _ ||- sconj ?C ?D =>
	  first [
	    match eval hnf in (A = C) with
	      | (?A' = ?A') => 
	        apply (sconjImplR Hp); try (clear Hp); try (clear h);
		let h' := fresh "h" in
		let Hp' := fresh "Hp" in
		  (intros h' Hp';
		   matchHeap2 Hp' (pred ig) (pred hshifts) (pred hshifts))
	      | ((cont ?x ?y) = (cont ?x ?z)) =>
		replace z with y;
		  [matchHeap2 Hp ig ih hshifts | try reflexivity]
	    end |
	    match eval hnf in (B = D) with
	      | (?B' = ?B') =>
		apply (sconjImplL Hp); try (clear Hp); try (clear h);
		let h' := fresh "h" in
		let Hp' := fresh "Hp" in
		  (intros h' Hp'; 
		   matchHeap2 Hp' (pred ig) (pred hshifts) (pred hshifts))
	    end |
	    match eval compute in ih with
	      | O =>
		match eval compute in ig with
		  | O => idtac (* finished *)
		  | _ => (* shift goal *)
		    apply sconjComm; 
		    try linLogIntro;
		    matchHeap2 Hp (pred ig) hshifts hshifts
		end
	      | _ => (* shift hyp *)
		let Hp' := fresh "Hp" in
		  (assert (Hp' := sconjComm Hp);
		   try linLogElim Hp';
		   try clear Hp;
		   matchHeap2 Hp' ig (pred ih) hshifts)
	    end]
      end
    | _ => idtac
  end.
  

Ltac numSconj k A :=
  match A with
    | sconj ?B ?C => 
      let k' := (fun n => k (S n)) in
	numSconj k' C
    | _ => k O
  end.


Ltac sconjStepX :=
  match goal with
    | [Hp : ?h ||- ?A |- ?h ||- ?B ] =>
      let k := 
        (fun hshifts => 
	  let k' := (fun gshifts => matchHeap2 Hp gshifts hshifts hshifts) in
	    numSconj k' B) in
	numSconj k A
 end.


(** Older version of the linLog introduction tactic *)

Ltac linLogIntroOld :=
  match goal with
    | |- _ ||- top => apply topIntro
    | |- _ ||- bang _ => apply bangIntro
    | |- _ ||- sconj (fas _ empty_set) _ => apply fasEIntroL; try linLogIntroOld
    | |- _ ||- sconj _ (fas _ empty_set) => apply fasEIntroR; try linLogIntroOld
    | |- _ ||- sconj _ unit => apply unitIR; try linLogIntroOld
    | |- _ ||- sconj unit _ => apply unitIL; try linLogIntroOld
    | |- _ ||- sconj (bang _) _ => apply bangIL; [auto | try linLogIntroOld]
    | |- _ ||- sconj _ (bang _) => apply bangIR; [auto | try linLogIntroOld]
    | |- _ ||- sconj (lex _) _ => apply lexIL
    | |- _ ||- sconj _ (lex _) => apply lexIR
    | |- _ ||- sconj (sconj _ _) _ => apply sconjAssocR; try linLogIntroOld
    | |- _ ||- sconj _ _ => sconjSubIntroOld
  end
with sconjSubIntroOld :=
  let h := fresh "h" in 
  let H := fresh in
    first
    [eapply sconjImpl; 
      [idtac |
	intros h H; linLogIntroOld; try (apply H) |
	intros h H; try linLogIntroOld; try (apply H)] |
      eapply sconjImpl;
	[idtac |
	  intros h H; try (apply H) |
	    intros h H; linLogIntroOld; try (apply H)]].


(** Older version of the linLog elimination tactic *)

Ltac linLogElimOld H :=
  (match type of H with
     | _ ||- sconj (sconj _ _) _ => hypReplace H (sconjAssocL H)
     |  _ ||- sconj unit _ => hypReplace H (unitEL H)
     |  _ ||- sconj _ unit => hypReplace H (unitER H)
     |  ?h ||- sconj top top => hypReplace H (topIntro h)
     |  _ ||- lex _ => 
       let x := fresh "x" in
       let H' := fresh in
         (destruct (lexE H) as [x H'];
	   clear H; rename H' into H)
     |  _ ||- bang _ =>
       let HA := fresh in
	 (assert (HA := bangElim1 H);
	   hypReplace H (bangElim2 H))
     |  _ ||- sconj _ _ => sconjSubSimplOld H
     |  _ ||- aconj _ _ =>
       let HA := fresh in
       let HB := fresh in
       let H' := fresh in
	 (assert (H' := H);
	   destruct H' as [HA HB];
	   ((linLogElimOld HA; try linLogElimOld HB) ||
	     (linLogElimOld HB));
	   hypReplace H (aconjIntro HA HB);
	   clear HA HB)
   end);
   try (linLogElimOld H)
with sconjSubSimplOld H :=
  let h1 := fresh "h" in
  let h2 := fresh "h" in
  let J := fresh "J" in
  let HA := fresh "H" in
  let HB := fresh "H" in
    (destruct (sconjElim H) as [h1 [h2 [J [HA HB]]]];
      ((linLogElimOld HA; try linLogElimOld HB) ||
	(linLogElimOld HB));
      hypReplace H (sconjIntro HA HB J);
      clear J HA HB h1 h2).


(** heap_eqv stuff *)

Lemma sconjHpSet : forall h h0 A,
  h ||- heap_eqv h0 ** A ->
  forall a v h0',
    set h0 a v = Some h0' ->
      uset h a v ||- heap_eqv h0' ** A.
Proof.
  intros h h0 A Hp a v h0' Heq.
  destruct (sconjElim Hp) as [ha [hb [J [HA HB]]]]; clear Hp.
  exists h0'; exists hb.
  intuition.
  eapply mapJoinEqv
    with (ma:=uset ha a v) (mb:=hb) (m:=uset h a v).
  
  assert (G : exists x, MapDef.get ha (div4 a) = Some x).
  unfold set in Heq.
  assert (Heq' : heap_eqv ha h0).
  intro x.
  rewrite map_canonEqv.
  rewrite HA.
  reflexivity.
  unfold get in Heq.
  destruct (isPtr a).
  rewrite <- Heq' in Heq.
  destruct (MapDef.get ha (div4 a)).
  injection Heq; econstructor; reflexivity.
  discriminate Heq.
  discriminate Heq.
  destruct G as [x Geq].

  unfold uset; eapply map_addJoin; eauto.
  
  unfold set in Heq.
  destruct (get h0 a).
  injection Heq; intro R; rewrite <- R; clear R Heq.
  intro x.
  rewrite <- map_canonEqv.
  simpl.
  destruct (POnat.POnat.eqDec x (div4 a)).
  reflexivity.
  rewrite HA.
  rewrite <- map_canonEqv; reflexivity.
  discriminate Heq.
  apply map_canonEqv.
  apply map_canonEqv.
  intro.
  apply map_canonEqv.
Qed.


Lemma heapEqvAdd : forall h x v h' A,
  h ||- x |-> v ** Heap.heap_eqv h' ** A ->
    h ||- Heap.heap_eqv (Heap.uset h' x v) ** A.
Proof.
  intros.
  sconjStep' 3.
  intro a.
  unfold uset.
  destruct (sconjElim H) as [h1 [h2 [J [Hp1 Hp2]]]].
  unfold hpredTrue in * .
  compare (div4 x) a; intro R; subst.
  simpl.
  destruct POnat.POnat.eqDec.
  hypReplace Hp1 (contGetEq Hp1).
  unfold get in Hp1.
  Lemma ptrPDec : forall n,
    ptrP n \/ ~ ptrP n.
  Proof.
    intro n.
    pattern n.
    apply ind_4.
    left; constructor.
    right; intro F; inversion F.
    right; intro F; inversion F.
    right; intro F; inversion F.
    intros.
    destruct H; [left |right].
    constructor; trivial.
    intro F; apply H; inversion_clear F; apply H0.
  Qed.
  destruct (ptrPDec x).
  rewrite (isPtrEq H0) in Hp1.
  rewrite <- (map_canonEqv h).
  rewrite (mapJoinSoundLeft J Hp1).
  reflexivity.
  Lemma isPtrEq' : forall n,
    ~ ptrP n -> isPtr n = false.
  Proof.
    intro n; pattern n; apply ind_4; intros.
    elimtype False; apply H; constructor.
    reflexivity.
    reflexivity.
    reflexivity.
    apply H.
    intro F; apply H0.
    constructor; exact F.
  Qed.
  rewrite (isPtrEq' H0) in Hp1.
  discriminate.
  tauto.

  simpl.
  destruct POnat.POnat.eqDec.
  subst a; tauto.
  rewrite Hp2.
  rewrite <- (map_canonEqv h).
  rewrite <- (map_canonEqv h2).
  assert (MapDef.get h2 a = None \/ exists v0, MapDef.get h2 a = Some v0).
  destruct (MapDef.get h2 a); [right; exists b |left]; reflexivity.
  destruct H0.
  destruct Hp1 as [_ Hp1].
  hypReplace Hp1 (Hp1 a).
  simpl in Hp1.
  destruct POnat.POnat.eqDec.
  subst; tauto.
  rewrite <- (map_canonEqv h1) in Hp1.
  rewrite (mapJoinSoundNone J Hp1 H0).
  exact H0.
  destruct H0.
  rewrite (mapJoinSoundRight J H0).
  exact H0.
Qed.


Lemma contIntro2 : forall h x v,
  Heap.get h x = Some v -> h ||- x |-> v ** top.
Proof.
  intros h x v E.

  assert (Px : ptrP x).
  unfold get in E.
  assert (X := isPtrOk x).
  destruct (isPtr x).
  assumption.
  discriminate.

  assert (subHeap (singleton_heap x v) h).
  unfold get in E.
  rewrite (isPtrEq Px) in E.
  intros addr val.
  unfold get.
  destruct (isPtr addr); intros E'; try discriminate.
  simpl in E'.
  destruct POnat.POnat.eqDec; try discriminate.
  rewrite <- e in E.
  rewrite <- E'.
  exact E.
  eapply subHeapImpl.
  2: apply H.
  unfold hpredTrue.
  unfold cont.
  split; trivial.
  apply map_eqvSymm; apply map_canonEqv.
Qed.

Lemma heapAddOk : forall h A,
  h ||- A -> forall x, 
    ptrP x ->
    ~ (h ||- x |->_ ** top) -> forall v,
      Heap.uset h x v ||- x |-> v ** A.
Proof.
  intros h A Hp x Px NHp v.
  apply sconjIntro with (ha := singleton_heap x v)
    (hb := h); trivial.
  unfold hpredTrue.
  unfold cont.
  split; trivial.
  apply map_eqvSymm; apply map_canonEqv.
  constructor; intros.
  compare k (div4 x); intro R; subst.
  simpl.
  destruct POnat.POnat.eqDec; try tauto.
  assert (get h x = None \/ exists v0, get h x = Some v0).
  destruct (get h x); [right; exists n | left]; reflexivity.
  destruct H.
  unfold get in H.
  rewrite (isPtrEq Px) in H.
  unfold MapDef.get.
  rewrite H.
  apply leftJoin.
  destruct H.
  elimtype False; apply NHp.
  linLogIntro.
  apply (contIntro2 _ _ H).
  simpl.
  destruct POnat.POnat.eqDec.
  tauto.
  destruct (MapDef.get h k);
    constructor.
Qed.


Lemma contElim1L : forall h x w A,
  h ||- x |-> w ** A ->
  ptrP x.
Proof.
  intros h x w A Hp.
  assert (Hp' : h ||- (!(ptrP x) ** top) ** top).
  sconjMatch.
  linLogIntro; eapply contElim1; eauto.
  linLogElim Hp'; assumption.
Qed.


(** an empty heap cannot contain anything *)
Lemma unitContElim : forall h,
  h ||- unit -> forall x y A,
  h ||- x |-> y ** A ->
  False.
Proof.
  intros h Hp x y A Hp'.
  assert (Some y = None).
  rewrite <- (contGetEqL Hp').
  assert (U := unitElim Hp).
  hypReplace U (U (div4 x)).
  unfold Heap.get.
  rewrite isPtrEq.
  simpl in U; assumption.
  eapply contElim1L; eauto.
  discriminate H.
Qed.



Lemma heapEqvImpl : forall h A h',
  h ||- A && Heap.heap_eqv h' ->
  h' ||- A.
Proof.
  intros h A h' Hp.
  destruct (aconjElim Hp) as [HX HY]; clear Hp.
  assert (Heq : Heap.heap_eqv h' h).
  intro.
  rewrite HY.
  symmetry; apply Heap.HeapMap.map_canonEqv.

  clear HY.
  assert (A (Heap.heap_canon h')).
  unfold Heap.heap_canon; rewrite (Heap.HeapMap.map_canonEqvEq Heq).
  assumption.
  assumption.
Qed.


Lemma eqvPredEqv : forall h h',
  h ||- Heap.heap_eqv h' ->
  Heap.heap_eqv h' h.
Proof.
  unfold Heap.heap_eqv.
  intros h h' Hp x.
  rewrite Hp.
  symmetry.
  apply Heap.HeapMap.map_canonEqv.
Qed.


Lemma eqvPredEqv' : forall h h',
  Heap.heap_eqv h' h ->
  h ||- Heap.heap_eqv h'.
Proof.
  intros h h' E.
  unfold hpredTrue; eapply Heap.heap_eqvTrans.
  eauto.
  apply Heap.HeapMap.map_canonEqv.  
Qed.


Lemma unitElim2 : forall h,
  h ||- unit ->
  h ||- Heap.heap_eqv Heap.empty_heap.
Proof.
  intros h Hp.
  unfold unit in Hp.
  assumption.
Qed.


Lemma sconjEqSubHeap : forall h h' A,
  h ||- Heap.heap_eqv h' ** A ->
  Heap.subHeap h' h.
Proof.
  intros h h' A Hp x v Heq'.

  assert (Hp' : h' ||- x |-> v ** top).
  apply contIntro2.
  assumption.
  hypReplace Hp (sconjInsert Hp Hp').
  linLogElim Hp; eapply contGetEqL; eauto.
Qed.


Lemma getUnitFalse : forall h x v A,
  h ||- x |-> v ** A ->
  h ||- unit ->
  False.
Proof.
  intros h x v A Hp Hp'.
  assert (Heap.get h x = Some v).
  eapply (contGetEqL Hp).
  assert (Heap.get h x = None).
  unfold Heap.get.    
  rewrite (Heap.HeapMap.map_canonEqv h).
  rewrite <- Hp'.
  destruct (isPtr x); reflexivity.
  rewrite H in H0; discriminate H0.
Qed.


Lemma subHeapContEq : forall h h',
  Heap.subHeap h h' -> forall x v A,
    h' ||- x |-> v  ** A -> forall v' B,
    h  ||- x |-> v' ** B ->
    v = v'.
Proof.
  intros h h' S x v A Hp v' B Hp'.
  assert (X := contGetEqL Hp).
  assert (Y := contGetEqL Hp').
  hypReplace S (S _ _ Y).
  rewrite X in S; injection S; auto.
Qed.


Lemma heapEqvImplL : forall h A h' B,
  h ||- (A && Heap.heap_eqv h') ** B ->
    h' ||- A.
Proof.
  intros h A h' B Hp.
  destruct (sconjElim Hp) as [ha [hb [_ [HA _]]]].
  destruct (aconjElim HA) as [HA1 HA2].
  eapply heapEqvTrue.
  2:eauto.
  intros x.
  rewrite HA2.
  rewrite <- Heap.HeapMap.map_canonEqv; reflexivity.
Qed.


Lemma heapEqvImplR : forall h A h' B,
  h ||- (Heap.heap_eqv h' && A) ** B ->
    h' ||- A.
Proof.
  intros h A h' B Hp.
  eapply heapEqvImplL.
  instantiate (1 := B).
  instantiate (1 := h).
  scm.
  destruct (aconjElim Hp); apply aconjIntro; assumption.
Qed.


Lemma intuitImpl : forall h A B,
  h ||- A ** B ->
  (forall h1 h2, Heap.subHeap h1 h2 -> h1 ||- A -> h2 ||- A) ->
  h ||- A.
Proof.
  intros h A B Hp Imp.
  destruct (sconjElim Hp) as [ha [hb [J [HA _]]]].
  eapply Imp; eauto.
  eapply Heap.heap_joinSub; eauto.
Qed.


Lemma fasElimL : forall a s p h A,
  set_In a s -> h ||- fas p s ** A->
    h ||- (p a ** fas p (set_remove a s)) ** A.
Proof.
  intros.
  eapply sconjImplL.
  apply H0.
  intros.
  apply fasElim;auto.
Qed.
Implicit Arguments fasElimL [a s p h A].


Lemma objsMerge : forall h X X',
  h ||- (foralls x in X, x |->__) ** (foralls x in X', x |->__) ->
  h ||- (foralls x in X |_| X', x |->__).
Proof.
  intros h S S' Hp; apply fasMerge.
  assumption.
  apply (fasFasDisj' Hp); illeim.
Qed.
Implicit Arguments objsMerge [h X X'].


Ltac llMatch :=
  eauto;
  match goal with
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?B =>
      exact Hp
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?A ?C =>
      apply (sconjImplR Hp); try (clear Hp); try (clear h);
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; llMatch)
    | Hp : ?h ||- sconj ?A ?B |- ?h ||- sconj ?C ?B =>
      apply (sconjImplL Hp); try (clear Hp); try (clear h); 
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; llMatch)
    | Hp : ?h ||- sconj _ _ |- ?h ||- sconj _ _ =>
      apply (sconjImpl Hp); try (clear Hp); try (clear h); 
	let h' := fresh "h" in
	let Hp' := fresh "Hp" in
	  (intros h' Hp'; llMatch)
    | |- _ ||- top => apply topIntro

    | Hp : ?h ||- fas ?A ?X |- ?h ||- fas ?B ?X =>
      apply (fasImplI Hp); try (clear Hp); try (clear h);
        let x := fresh "x" in
        let h' := fresh "h" in
        let I := fresh "I" in
        let Hp' := fresh "Hp" in
           (intros x h' I Hp')
    | |- _ => idtac
  end.


Ltac llm := llMatch.
