Require Export LinLog.
Require Export List.


Set Implicit Arguments.


(*

This file describes three types of singly linked lists:
  1) basic lists (ListAt) that describe the raw contents
  2) nat lists (NatListAt) that store their contents with a 31-bit encoding
  3) summed list, that only describe the sum of the nats in the list
*)


Definition NULL := 1.


(* basic linked list *)

Fixpoint ListAt (l:list nat) (x:nat) {struct l} : hpred :=
  match l with
    | nil => !(x = NULL)
    | y::l' => 
      lexists x', !(x <> NULL) ** (x |->, (y::x'::nil)) ** ListAt l' x'
  end.


Lemma ListAtEq : forall l1 l2 h x A B,
  h ||- (ListAt l1 x ** A) && (ListAt l2 x ** B) ->
      l1 = l2.
Proof.
  induction l1; destruct l2; simpl; intros s x A B Hp;
    destruct (aconjElim Hp) as [H1 H2]; linLogElim H1; linLogElim H2.
  reflexivity.
  elimtype False; auto.
  elimtype False; auto.

  clear H6.
  assert (a = n).
  assert (X := contContEq H1 H2); omega.
  subst n.
  assert (R2 : x0 = x1).
  sconjLift2 H1.
  sconjLift2 H2.
  eapply contContEq; eauto.

  subst.
  replace l2 with l1.
  reflexivity.
  sconjLift3 H1.
  sconjLift3 H2.
  eapply IHl1.
  apply aconjIntro; eauto.

Qed.



Lemma ListAtIntroNil : forall h,
  h ||- unit ->
  h ||- ListAt nil NULL.
Proof.
  intros h Hp.
  simpl; linLogIntro; auto.
Qed.


Lemma ListAtIntroCons : forall h x x' m l, 
  h ||- x |->, (m::x'::nil) ** ListAt l x' ->
  x <> NULL ->
  h ||- ListAt (m::l) x.
Proof.
  intros h x x' m l Hp Ne.
  linLogElim Hp.
  simpl; linLogIntro.
  sconjMatch.
Qed.


Lemma ListAtElimNil1 : forall h l,
  h ||- ListAt l NULL ->
    l = nil.
Proof.
  destruct l; intros Hp.
  reflexivity.
  simpl in Hp; linLogElim Hp; elimtype False; auto.
Qed.


Lemma ListAtElimNil2 : forall h l,
  h ||- ListAt l NULL ->
  h ||- unit.
Proof.
  destruct l; simpl; intros Hp.
  linLogElim Hp; assumption.
  linLogElim Hp; elimtype False; auto.
Qed.


Lemma ListAtElimCons : forall h l x,
  h ||- ListAt l x -> x <> NULL ->
    exists m, exists l', exists x', 
      l = m::l' /\ h ||- x |->, (m::x'::nil) ** ListAt l' x'.
Proof.
  intros h l x Hp Nz.
  destruct l; simpl in Hp; linLogElim Hp.
  elimtype False; auto.
  clear H1.
  exists n; exists l; exists x0.
  constructor.
  reflexivity.
  linLogIntro; sconjMatch.
Qed.



(* list of nats *)

Definition natEnc (n:nat) := S (2 * n).

Definition NatListAt (l:list nat) (x:nat) : hpred := ListAt (map natEnc l) x.


Lemma encNatListEq : forall l l',
  map natEnc l = map natEnc l' ->
  l = l'.
Proof.
  induction l; destruct l'; simpl.
  reflexivity.
  discriminate 1.
  discriminate 1.
  injection 1; clear H; intros E1 E2.
  replace a with n.
  rewrite (IHl l'); auto.
  omega.
Qed.


Lemma NatListAtEq : forall l1 l2 h x A B,
  h ||- (NatListAt l1 x ** A) && (NatListAt l2 x ** B) ->
      l1 = l2.
Proof.
  unfold NatListAt.
  intros l1 l2 h x A B Hp.
  apply encNatListEq.
  destruct (aconjElim Hp) as [Hp1 Hp2]; clear Hp.
  eapply ListAtEq.
  apply aconjIntro; eauto.
Qed.


Lemma NatListAtIntroNil : forall h,
  h ||- unit ->
  h ||- NatListAt nil NULL.
Proof.
  unfold NatListAt; intros; apply ListAtIntroNil; assumption.
Qed.


Lemma NatListAtIntroCons : forall h x x' m l, 
  h ||- x |->, (natEnc m::x'::nil) ** NatListAt l x' ->
  x <> NULL ->
  h ||- NatListAt (m::l) x.
Proof.
  unfold NatListAt.
  intros h x x' m l Hp Ne.
  replace (map natEnc (m :: l)) with (natEnc m :: map natEnc l).
  eapply ListAtIntroCons; eauto.
  reflexivity.
Qed.


Lemma NatListAtElimNil1 : forall h l,
  h ||- NatListAt l NULL ->
    l = nil.
Proof.
  unfold NatListAt; intros h l Hp.
  assert (map natEnc l = nil).
  eapply ListAtElimNil1; eauto.
  destruct l; [reflexivity | discriminate H].
Qed.


Lemma NatListAtElimNil2 : forall h l,
  h ||- NatListAt l NULL ->
  h ||- unit.
Proof.
  unfold NatListAt; intros h l Hp; eapply ListAtElimNil2; eauto.
Qed.


Lemma NatListAtElimCons : forall h l x,
  h ||- NatListAt l x -> x <> NULL ->
    exists m, exists l', exists x', 
      l = m::l' /\ h ||- x |->, (natEnc m::x'::nil) ** NatListAt l' x'.
Proof.
  unfold NatListAt.
  intros h l x Hp Nz.
  destruct (ListAtElimCons _ _ Hp Nz) as [m [l' [x' [Leq Hp']]]]; clear Hp Nz.
  destruct l.
  discriminate Leq.
  simpl in Leq; injection Leq; clear Leq; intros E1 E2.
  subst l'; subst m.
  exists n; exists l; exists x'.
  constructor; auto.
Qed.



(* list of nats where we only care about the sum *)


Fixpoint ListSum (l:list nat) : nat :=
  match l with
    | nil => 0
    | x::l' => x + (ListSum l')
  end.


Definition ListSumAt (n x:nat) :=
  lexists l, !(ListSum l = n) ** NatListAt l x.


Lemma ListSumAtEq : forall h x x' y A B,
  h ||- (ListSumAt x y ** A) && (ListSumAt x' y ** B) ->
    x = x'.
Proof.
  unfold ListSumAt.
  intros s x x' y A B Hp.
  destruct (aconjElim Hp) as [H1 H2]; clear Hp.
  linLogElim H1; linLogElim H2.
  subst.
  replace x1 with x0.
  reflexivity.
  eapply NatListAtEq.
  apply aconjIntro; eauto.
Qed.


Lemma ListSumAtIntroNil : forall h,
  h ||- unit ->
  h ||- ListSumAt 0 NULL.
Proof.
  intros h Hp.
  unfold ListSumAt.
  existsl (nil (A := nat)).
  simpl.
  linLogIntro.
  apply NatListAtIntroNil; assumption.
Qed.


Lemma ListSumAtIntroCons : forall h x x' m n, 
  h ||- x |->, (natEnc m::x'::nil) ** ListSumAt n x' ->
  x <> NULL ->
    h ||- ListSumAt (m + n) x.
Proof.
  unfold ListSumAt; simpl.
  intros h x x' m n H Ne.
  linLogElim H.
  subst n.
  existsl (m::x0).
  linLogIntro.
  eapply NatListAtIntroCons; try linLogIntro; eauto.
Qed.  



Lemma ListSumAtElimNil1 : forall h n,
  h ||- ListSumAt n NULL ->
    n = 0.
Proof.
  unfold ListSumAt.
  intros h n H.
  linLogElim H.
  assert (x = nil).
  eapply NatListAtElimNil1; eauto.
  subst x; simpl in H2; subst n; reflexivity.
Qed.


Lemma ListSumAtElimNil2 : forall h l,
  h ||- ListSumAt l NULL ->
  h ||- unit.
Proof.
  unfold ListSumAt.
  intros h l Hp.
  linLogElim Hp.
  eapply NatListAtElimNil2; eauto.
Qed.


Lemma ListSumAtElimCons : forall h n x,
  h ||- ListSumAt n x -> x <> NULL ->
    exists m, exists n', exists x', 
      n = m + n' /\ h ||- x |->, (natEnc m::x'::nil) ** ListSumAt n' x'.
Proof.
  unfold ListSumAt.
  intros h n x Hp Nz.
  linLogElim Hp.
  subst n.
  destruct (NatListAtElimCons _ _ Hp Nz) as [m [l' [x' [Leq Hp']]]]; clear Hp Nz.
  subst x0.
  exists m; exists (ListSum l'); exists x'.
  constructor.
  reflexivity.
  sconjMatch.
  linLogIntro; assumption.
Qed.


Lemma ListSumAtElimConsSt : forall h n x A,
  h ||- ListSumAt n x ** A -> x <> NULL ->
    exists m, exists n', exists x', 
      n = m + n' /\ h ||- x |->, (natEnc m::x'::nil) ** ListSumAt n' x' ** A.
Proof.
  intros h n x A Hp Ne.

  assert (Hp' : h ||-
    (lexists m, lexists n', lexists x', 
      !(n = m + n') ** x |->, (natEnc m::x'::nil) ** ListSumAt n' x') ** A).
  sconjMatch.
  destruct (ListSumAtElimCons Hp Ne) as [m [n' [x' [E Hp']]]]; clear Hp Ne.
  subst n.
  linLogElim Hp'.
  existsl m; existsl n'; existsl x'.
  linLogIntro; assumption.

  linLogElim Hp'.
  econstructor; econstructor; econstructor; constructor.
  eauto.
  linLogIntro; eauto.
Qed.
