Require Export State.
Require List.
Require Import LinLog.
Require Import Machine.


Set Implicit Arguments.


Notation "s |||- A" := (heapOf s ||- A) (at level 80).


(** basic reasoning about rnum equality *)
Lemma beq_rnum_refl : forall r,
  (beq_rnum r r) = true.
Proof.
  intro r; destruct r; auto.
Qed.

Lemma beq_rnum_neq : forall r r',
  r <> r' -> (beq_rnum r r') = false.
Proof.
  intro r; destruct r;
    intro r'; destruct r'; intuition.
Qed.

Hint Rewrite beq_rnum_neq beq_rnum_refl : RNum8.


(* using this is faster than using compare *)
Lemma rnumEqDec : forall (x y:rnum), {x = y} + {x <> y}.
Proof.
  intros x y.
  compare x y; intro E; [left | right]; auto.
Qed.


(** reasoning about branch tests *)

Lemma btneEq : forall x y,
  Machine.doBtest btne x y = negb (beq_nat x y).
Proof.
  reflexivity.
Qed.

Lemma bteqEq : forall x y,
  Machine.doBtest bteq x y = beq_nat x y.
Proof.
  reflexivity.
Qed.


(** it is convenient to be able describe heap updates that always succeed *)
Definition uhset (s:state) (a:nat) (v:nat) : state :=
  (Heap.uset (heapOf s) a v, rfileOf s).

Lemma rgetZeroEq : forall s,
  rget s r0 = 0.
Proof.
  unfold rget.
  unfold RFile.get.
  simpl.
  reflexivity.
Qed.

Lemma rgetRsetEq : forall r (s:state) v,
  r <> r0 -> rget (rset s r v) r = v.
Proof.
  unfold rget; unfold rset; unfold rfileOf.
  unfold RFile.set; unfold RFile.get.
  intros.
  rewrite beq_rnum_neq.
  simpl.
  rewrite beq_rnum_refl.
  auto.
  auto.
Qed.


(* compare on regs is slow, so we just do it here once and for all *)
Lemma regCompare : forall r r' : rnum, r = r' \/ r <> r'.
Proof.
  intros r r'.
  compare r r'; auto.
Qed.

Lemma r0compare : forall r, r = r0 \/ ~(r = r0).
Proof.
  intro r.
  compare r r0; auto.
Qed.

Lemma rgetRsetNeq : forall r r' (s:state) v,
  r <> r' -> rget (rset s r v) r' = (rget s r').
Proof.
  unfold rget; unfold rset; unfold rfileOf.
  unfold RFile.set.
  intros r r' s v E.
  simpl.
  unfold RFile.get.
  destruct (r0compare r').
  
  subst.
  rewrite beq_rnum_refl.
  reflexivity.
  
  rewrite beq_rnum_neq.
  rewrite (beq_rnum_neq E).
  reflexivity.
  auto.

Qed.

Lemma rgetUhsetEq : forall s x y r,
  rget (uhset s x y) r = rget s r.
Proof.
  reflexivity.
Qed.

Lemma hgetRsetEq : forall s x y z,
  hget (rset s x y) z = hget s z.
Proof.
  reflexivity.
Qed.

Lemma contHgetEqL : forall h l w A,
  h ||- l |-> w ** A -> forall s,
    heapOf s = h ->
    hget s l = Some w.
Proof.
  unfold hget.
  intros.
  eapply contGetEqL.
  subst.
  eauto.
Qed.

Lemma contHsetEqL : forall h x y A,
  h ||- x |-> y ** A -> forall s,
    heapOf s = h -> forall z,
    hset s x z = Some (uhset s x z).
Proof.
  unfold hset; unfold uhset.
  intros.
  rewrite H0.
  rewrite (contSetEqL H).
  reflexivity.
Qed.


Lemma heapOfRsetEq : forall s r x,
  heapOf (rset s r x) = heapOf s.
Proof.
  reflexivity.
Qed.


Lemma heapOfUhsetRsetEq : forall s r v x y,
  heapOf (uhset (rset s r v) x y) = heapOf (uhset s x y).
Proof.
  reflexivity.
Qed.

Lemma heapOfPairEq : forall h r,
  heapOf (h, r) = h.
Proof.
  reflexivity.
Qed.


Lemma rfileOfPairEq : forall h r,
  rfileOf (h, r) = r.
Proof.
  reflexivity.
Qed.


Lemma rfileGetEq : forall s,
  RFile.get (rfileOf s) = rget s.
Proof.
  reflexivity.
Qed.


Lemma rgetPairRfileOf : forall h s,
  rget (h, rfileOf s) = rget s.
Proof.
  reflexivity.
Qed.


Hint Rewrite rgetZeroEq rgetUhsetEq hgetRsetEq 
  heapOfRsetEq heapOfUhsetRsetEq heapOfPairEq rfileOfPairEq
  rgetPairRfileOf rfileGetEq : stSimplBase.
Hint Rewrite rgetRsetEq rgetRsetNeq
  using (solve [auto | discriminate]) : stSimplBase.
Ltac stSimpl := autorewrite with stSimplBase.


(* it would be nice if I could use autorewrite to do this *)
Ltac stSimplSideCond := solve [assumption | discriminate].

Ltac lupEqIn X := 
  rewrite rgetRsetEq in X; [idtac | stSimplSideCond].

Ltac lupNeqIn X := 
  rewrite rgetRsetNeq in X; [idtac | stSimplSideCond].

Ltac lupSimplIn X :=
  repeat (rewrite rgetZeroEq in X);
  repeat (lupEqIn X);
  repeat (lupNeqIn X);
  repeat (lupEqIn X).




(* properties of commands *)

Lemma adduStep : forall rd rs rt s,
  cstep (addu rd rs rt) s = Some (rset s rd (rget s rs + rget s rt)).
Proof.
  reflexivity.
Qed.


Lemma addiuStep : forall rd rs i s,
  cstep (addiu rd rs i) s = Some (rset s rd (i + rget s rs)).
Proof.
  reflexivity.
Qed.


Lemma addiuStep0 : forall rd rs s,
  cstep (addiu rd rs 0) s = Some (rset s rd (rget s rs)).
Proof.
  reflexivity.
Qed.


Lemma addiuStepZero : forall rd s k,
  cstep (addiu rd zero k) s = Some (rset s rd k).
Proof.
  intros rd s k; simpl; stSimpl.
  replace (k + 0) with k.
  reflexivity.
  auto.
Qed.


Lemma subuStep : forall rd rs rt s,
  cstep (subu rd rs rt) s = Some (rset s rd (rget s rs - rget s rt)).
Proof.
  reflexivity.
Qed.


Lemma andi1Step : forall rd rs s,
  cstep (andi1 rd rs) s = Some (rset s rd (if isOdd (rget s rs) then 1 else 0)).
Proof.
  reflexivity.
Qed.


Lemma sltuStep : forall rd rs rt s,
  cstep (sltu rd rs rt) s = Some (rset s rd (if blt_nat (rget s rs) (rget s rt) then 1 else 0)).
Proof.
  reflexivity.
Qed.


Lemma lwStep : forall h x y A,
  h ||- x |-> y ** A -> forall s rd i rs,
    heapOf s = h -> i + rget s rs = x ->
    cstep (lw rd i rs) s = Some (rset s rd y).
Proof.
  intros.
  simpl.
  subst.
  rewrite (contHgetEqL H); auto.
Qed.


Lemma swStep : forall h x y A,
  h ||- x |-> y ** A -> forall s rs i rd,
    heapOf s = h -> i + rget s rd = x ->
    cstep (sw rs i rd) s = Some (uhset s x (rget s rs)).
Proof.
  intros.
  simpl.
  subst.
  rewrite (contHsetEqL H); auto.
Qed.


Require Import Machine.
Require Import CAP.

Lemma doBtestSound : forall bt x y,
  if doBtest bt x y then 
    btPred bt x y 
  else 
    btPred (btNeg bt) x y.
Proof.
  destruct bt; intros x y; simpl.
  
  (* bteq *)
  assert (H : x = y \/ x <> y); [omega | destruct H].
  rewrite H; rewrite beq_true.
  reflexivity.
  rewrite beq_false; assumption.

  (* btne *)
  assert (H : x = y \/ x <> y); [omega | destruct H].
  rewrite H; rewrite beq_true.
  simpl; reflexivity.
  rewrite beq_false; assumption.

Qed.


Lemma bltNeqEqv : forall x y,
  (if blt_nat x y then 1 else 0) <> 0 ->
    x < y.
Proof.
  intros x y.
  assert (Z := blt_natSound x y).
  destruct (blt_nat x y); intuition.
Qed.


Lemma bltEqEqv : forall x y,
  (if blt_nat x y then 1 else 0) = 0 ->
    x >= y.
Proof.
  intros x y.
  assert (Z := blt_natSound x y).
  destruct (blt_nat x y); intuition.
Qed.


Ltac doBtestSimpl :=
  match goal with 
    |[|- context [(doBtest ?bt ?x ?y)]] =>
      assert (Bt := doBtestSound bt x y);
	destruct (doBtest bt x y); 
	  unfold btPred in Bt; 
	    [idtac | unfold btNeg in Bt];
	    try (hypReplace Bt (bltNeqEqv _ _ Bt));
	      try (hypReplace Bt (bltEqEqv _ _ Bt))
  end.


Lemma btPredNegOk : forall bt x y,
  btPred bt x y -> ~btPred (btNeg bt) x y.
Proof.
  destruct bt; intros x y; simpl; omega.
Qed.



Require Import Heap.
Import HeapMap.


Lemma heapJoinSoundLeft : forall ha hb h,
  heap_join ha hb h -> forall w v,
  get ha w = Some v ->
  get h w = Some v.
Proof.
  unfold get.
  intros ha hb h J w v Geq.
  destruct (isPtr w).
  apply (mapJoinSoundLeft J Geq).
  assumption.
Qed.


Lemma heapJoinSoundRight : forall ha hb h,
  heap_join ha hb h -> forall w v,
  get hb w = Some v ->
  get h w = Some v.
Proof.
  unfold get.
  intros ha hb h J w v Geq.
  destruct (isPtr w).
  apply (mapJoinSoundRight J Geq).
  assumption.
Qed.


Lemma heapJoinGetGetEq : forall h1 h2 h,
  heap_join h1 h2 h -> forall x v,
    get h x = Some v -> forall rf v',
      hget (h1, rf) x = Some v' ->
      v = v'.
  intros h1 h2 h J x v G1 rf v' G2.
  unfold hget in G2; unfold get in * |- .
  destruct (isPtr x).
  simpl in G2.
  assert (Q := mapJoinSoundLeft J G2).
  assert (Some v = Some v').
  rewrite <- G1.
  apply Q.
  injection H; auto.
  discriminate G1.
Qed.



(* setoid on states *)

Definition rfile_eqv rf rf' :=
  forall r, RFile.get rf r = RFile.get rf' r.

Definition state_eqv (s s':state) :=
  let (h, r) := s in
  let (h', r') := s' in
    Heap.heap_eqv h h' /\ rfile_eqv r r'.

Require Import Setoid.

Lemma state_eqvRefl : forall s,
  state_eqv s s.
Proof.
  intros [h r].
  constructor; intro; reflexivity.
Qed.


Lemma state_eqvSymm : forall s s',
  state_eqv s s' ->
  state_eqv s' s.
Proof.
  intros [h r] [h' r'] [Heq Req].
  constructor; intro; auto.
Qed.
Implicit Arguments state_eqvSymm [s s'].


Lemma state_eqvTrans : forall s s' s'',
  state_eqv s s' ->
  state_eqv s' s'' ->
  state_eqv s s''.
Proof.
  intros [h r] [h' r'] [h'' r''] [Heq Req] [Heq' Req'].
  constructor; intro.
  rewrite Heq; rewrite Heq'; reflexivity.
  rewrite Req; rewrite Req'; reflexivity.
Qed.
Implicit Arguments state_eqvTrans [s s' s''].


Lemma StateSetoid : Setoid_Theory state state_eqv.
Proof.
  constructor.
  apply state_eqvRefl.
  apply state_eqvSymm.
  apply state_eqvTrans.
Qed.

Add Setoid state state_eqv StateSetoid as setoidState.


Lemma rgetEqv : forall s s',
  state_eqv s s' ->
  forall r, rget s r = rget s' r.
Proof.
  intros [h rf] [h' rf'] [_ Req] r.
  unfold rget; simpl.
  rewrite Req; reflexivity.
Qed.

Implicit Arguments rgetEqv [s s'].


Lemma heap_eqvIntro : forall h h',
  (forall l, Heap.get h l = Heap.get h' l) ->
  Heap.heap_eqv h h'.
Proof.
  unfold Heap.get.
  intros h h' Heq x.
  hypReplace Heq (Heq (4 * x)); rewrite div44eq in Heq; rewrite isPtrEq in Heq.
  assumption.
  eapply ptrP_intro; reflexivity.
Qed.


Lemma heap_eqvElim : forall h h' l,
  heap_eqv h h' -> get h l = get h' l.
Proof.
  intros.
  unfold get.
  destruct (isPtr l); auto.
Qed.

    
Lemma hgetEqv : forall s s',
  state_eqv s s' ->
  forall x, hget s x = hget s' x.
Proof.
  intros [h rf] [h' rf'] [Heq _] x.
  unfold hget; simpl.
  apply heap_eqvElim; assumption.
Qed.
Implicit Arguments hgetEqv [s s'].


Lemma stateHeapEqv : forall s s',
  state_eqv s s' ->
  heap_eqv (heapOf s) (heapOf s').
Proof.
  intros [h r] [h' r'] [R _] x.
  stSimpl; apply R.
Qed.
Implicit Arguments stateHeapEqv [s s'].



Lemma heap_eqvSubHeapL : forall h h' A,
  h ||- Heap.heap_eqv h' ** A ->
  Heap.subHeap h' h.
Proof.
  intros h h' A Hp.
  destruct (sconjElim Hp) as [ha [hb [J [HA _]]]].
  unfold hpredTrue in HA.
  intros x y E.
  eapply heapJoinSoundLeft; eauto.
  rewrite Heap.heap_canonEqv.
  unfold Heap.get; rewrite <- HA; assumption.
Qed.


Lemma uhsetRsetEq : forall s r x y z,
  uhset (rset s r x) y z = rset (uhset s y z) r x.
Proof.
  reflexivity.
Qed.


Lemma hgetEq : forall l,
  ptrP l -> forall s v,
  hget (uhset s l v) l = Some v.
Proof.
  intros l Pl [h rf] v.
  unfold hget; unfold uhset; unfold Heap.get; unfold Heap.uset.
  simpl.
  destruct (POnat.POnat.eqDec (div4 l) (div4 l)).
  rewrite (isPtrEq Pl); reflexivity.
  elf; auto.
Qed.
Implicit Arguments hgetEq [l].


Lemma hgetNeq : forall l,
  ptrP l -> forall l',
  ptrP l' -> l <> l' -> forall s v,
  hget (uhset s l' v) l = hget s l.
Proof.
  intros l Pl l' Pl' NE [h rf] v.
  unfold hget; unfold uhset; unfold Heap.get; unfold Heap.uset; simpl.
  destruct (POnat.POnat.eqDec (div4 l) (div4 l')).

  elf.
  destruct (ptrP_elim Pl) as [k Pll]; clear Pl;
    destruct (ptrP_elim Pl') as [k' Pll']; clear Pl';
      subst l l'.
  rewrite div44eq in e; rewrite div44eq in e.
  auto.

  reflexivity.

Qed.
Implicit Arguments hgetNeq [l].



(* tactics to make reasoning about the state a little nicer *)

Ltac ssteps :=
  repeat (first [rewrite addiuStepZero | rewrite addiuStep0 | rewrite addiuStep
    | rewrite adduStep | rewrite subuStep | rewrite andi1Step | rewrite sltuStep];
  stSimpl).

(* a fancier version of swStep would also update the heap predicate used *)

Ltac msteps Hp :=
  ssteps;
  first [rewrite (lwStep Hp); [idtac | auto | auto] |
         rewrite (swStep Hp); [idtac | auto | auto]]; 
  ssteps.


(* msteps that updates the heap predicate when you do a store *)
Ltac swStepTac Hp :=
  match goal with 
    | |- match cstep (sw ?rs ?k ?rd) ?s with Some _ => _ | None => _ end =>
      rewrite (swStep Hp); 
	[hyr Hp (contUpdateSt Hp (rget s rs));
	  generalize dependent Hp; stSimpl; intro | auto | auto]
  end.

Ltac msteps' Hp :=
  ssteps;
  first [rewrite (lwStep Hp); [stSimpl | auto | auto] |
         swStepTac Hp];
  ssteps.

(* slightly modified version of doBtestSimpl *)
Ltac bsteps :=
  ssteps;
  match goal with 
    |[|- context [(doBtest ?bt ?x ?y)]] =>
      let Bt := fresh "Bt" in
	(assert (Bt := doBtestSound bt x y);
	  destruct (doBtest bt x y); 
	    unfold CAP.btPred in Bt; 
	      [idtac | unfold CAP.btNeg in Bt];
	      try (hypReplace Bt (bltNeqEqv _ _ Bt));
		try (hypReplace Bt (bltEqEqv _ _ Bt));
		  try (discriminate Bt); try (elf; auto; fail))
  end;
  ssteps.

(* 
use "replace X with Y by T" instead
Tactic Notation "replace" constr(x) "with" constr(y) "using" tactic(z) :=
  replace x with y; [idtac | z].

use "replace x with y in H by z" instead
Tactic Notation "replace" constr(x) "with" constr(y) "in" hyp(H) "using" tactic(z) :=
  replace x with y in H; [idtac | z].
*)

Tactic Notation "rreplace" constr(x) "with" constr(y) :=
  replace x with y by reflexivity.

Tactic Notation "rreplace" constr(x) "with" constr(y) "in" hyp(H) :=
  replace x with y in H by reflexivity .

Tactic Notation "omreplace" constr(x) "with" constr(y) :=
  replace x with y by omega.

Tactic Notation "omreplace" constr(x) "with" constr(y) "in" hyp(H) :=
  replace x with y in H by omega.
