Require Export SCAP.
Require Export LinLog.
Require Import StateTheory.
Require Import WeakSCAP.

Set Implicit Arguments.


(* assembly implementation of swap function *)

Definition ri := r1.
Definition rj := r2.
Definition ru := r3.
Definition rv := r4.

Definition swap :=
  lw ru 0 ri;  (* ru := [ri] *)
  lw rv 0 rj;  (* rv := [rj] *)
  sw rv 0 ri;  (* [ri] := rv *)
  sw ru 0 rj;  (* [rj] := ru *)
  jr ra.       (* return *)


(* swap function specification *)

Definition swapPre : spred := fun s =>
  s |||- rget s ri |->_ ** rget s rj |->_ ** top.

Definition swapGuar : guar := fun s s' =>
  (forall x y A,
    s |||- rget s ri |-> x ** rget s rj |-> y ** A ->
    s' |||- rget s' ri |-> y ** rget s' rj |-> x ** A) /\
  (forall r, r <> ru -> r <> rv -> rget s r = rget s' r).


(* verification that the implementation matches the spec *)

Theorem swapOk : iseqOkS (fun _ => NoneT _) swapPre swapGuar swap.
Proof.
  (* generate the VC *)
  unfold swap, swapPre.
  iseqWeakSolver idtac.

  (* process the precondition *)
  intros s Hp; lle Hp; rename x0 into y.

  (* step through the procedure, one instruction at a time *)
  msteps Hp.
  sconjLift2 Hp; msteps Hp.
  sconjLift2 Hp; msteps Hp.
  assert (Hp' := contUpdateSt Hp y).
  sconjLift2 Hp'; msteps Hp'.
  clear Hp'.
  stSimpl.

  (* now we've related the initial state to the final state,
     and shown it is safe to execute the block. *)
  constructor.

  (* show that it is safe to leave this block *)
  trivial.

  (* show the the block has the proper behavior *)
  intros s' E; subst s'.
  unfold swapGuar; stSimpl.
  constructor.

  (* show the final heap has the right shape *)
  intros x' y' A Hp'.

  (* first show that x' = x and y' = y *)
  assert (E1 := contContEq Hp Hp').
  sconjLift2 Hp; sconjLift2 Hp'; assert (E2 := contContEq Hp Hp').
  subst x' y'; clear Hp; rename Hp' into Hp.

  (* now step apply the writes to the heap predicate *)
  sconjLift2 Hp; hyr Hp (contUpdateSt Hp y).
  sconjLift2 Hp; hyr Hp (contUpdateSt Hp x).
  sconjLift2 Hp.
  assumption.

  (* show that only ru and rv change *)
  intros r NEu NEv.
  stSimpl; reflexivity.

Qed.
